/* editdistance.c: Edit distance calculation algorithm.
 * Copyright (C) 2003  Richard Boulton
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/** @file editdistance.c
 *  @brief Implementation of edit distance calculation algorithm.
 *
 *  Based on that described in:
 *
 *  "An extension of Ukkonen's enhanced dynamic programming ASM algorithm"
 *  by Hal Berghel, University of Arkansas
 *  and David Roach, Acxiom Corporation
 */

#include "editdistance.h"

typedef int (*edist_eqfn)(void *, void *);

struct edist_seq_ {
    char * ptr;
    int len;
};
typedef struct edist_seq_ edist_seq;


enum edit_type {
    ETYPE_NOP,
    ETYPE_INSERT,
    ETYPE_DELETE,
    ETYPE_TRANSPOSE,
    ETYPE_SUBSTITUTE
};

struct edist_edit_ {
    enum edit_type type;
    int position;
};
typedef struct edist_edit_ edist_edit;

struct edist_state_ {
    /* Function used to compare two items for equality. */
    edist_eqfn eqfn;
    /* Size of each item (for now, all items have to be same size). */
    size_t itemsize;

    edist_seq seq1;
    edist_seq seq2;

    /* Array of f(k,p) values, where f(k,p) = the largest index i such that
     * d(i,j) = p and d(i,j) is on diagonal k.
     * ie: f(k,p) = largest i s.t. d(i,k+i) = p
     * Where: d(i,j) = edit distance between substrings of length i and j.
     */
    int * fkp;
    int fkp_rows;
    int fkp_cols;

    /** Array of edits needed to get to fkp.
     */
    edist_edit * edits;

    /* Maximum possible edit distance (this is referred to as ZERO_K in
     * the algorithm description by Berghel and Roach). */
    int maxdist;
};
typedef struct edist_state_ edist_state;

static int get_f_kp(edist_state * state, int k, int p)
{
    return (state->fkp)[(k + state->maxdist) * state->fkp_cols + p + 1];
}

static void set_f_kp(edist_state * state, int k, int p, int val)
{
    (state->fkp)[(k + state->maxdist) * state->fkp_cols + p + 1] = val;
}

static edist_edit * get_edit(edist_state * state, int k, int p)
{
    return (state->edits) + (k + state->maxdist) * state->fkp_cols + p + 1;
}

static void set_edit(edist_state * state, int k, int p,
		     enum edit_type type, int position)
{
    edist_edit * edit = get_edit(state, k, p);

    edit->type = type;
    edit->position = position;
}

static int is_transposed(edist_state * state, int pos1, int pos2)
{
    if (pos1 <= 0 || pos1 >= state->seq1.len ||
	// pos2 <= 0 || - this never happens, I think
	pos2 >= state->seq2.len) return 0;
    return (state->eqfn(state->seq1.ptr + (pos1 - 1) * state->itemsize,
			state->seq2.ptr + pos2 * state->itemsize) &&
	    state->eqfn(state->seq1.ptr + pos1 * state->itemsize,
			state->seq2.ptr + (pos2 - 1) * state->itemsize));
}

static void edist_calc_f_kp(edist_state * state, int k, int p)
{
    int maxlen = get_f_kp(state, k, p - 1) + 1; /* dist if do substitute */
    int maxlen2 = get_f_kp(state, k - 1, p - 1); /* dist if do insert */
    int maxlen3 = get_f_kp(state, k + 1, p - 1) + 1; /* dist if delete */

    enum edit_type type = ETYPE_SUBSTITUTE;
    if (is_transposed(state, maxlen, maxlen + k)) {
	maxlen += 1;
	type = ETYPE_TRANSPOSE;
    }

    if (maxlen >= maxlen2) {
	if (maxlen >= maxlen3) {
	    set_edit(state, k, p, type, maxlen);
	} else {
	    maxlen = maxlen3;
	    set_edit(state, k, p, ETYPE_DELETE, maxlen);
	}
    } else {
	if (maxlen2 >= maxlen3) {
	    maxlen = maxlen2;
	    set_edit(state, k, p, ETYPE_INSERT, maxlen);
	} else {
	    maxlen = maxlen3;
	    set_edit(state, k, p, ETYPE_DELETE, maxlen);
	}
    }

    /* Check for exact matches, and increase the length until we don't have
     * one. */
    while (maxlen < state->seq1.len &&
	   maxlen + k < state->seq2.len &&
	   state->eqfn(state->seq1.ptr + maxlen * state->itemsize,
		       state->seq2.ptr + (maxlen + k) * state->itemsize)) {
	maxlen = maxlen + 1;
    }
    set_f_kp(state, k, p, maxlen);
}

#define INF 10000000
static int edist_init(edist_state * state, edist_eqfn eqfn, size_t itemsize,
		       edist_seq seq1, edist_seq seq2)
{
    int len1 = seq1.len;
    int len2 = seq2.len;
    int k;
    state->eqfn = eqfn;
    state->itemsize = itemsize;
    state->seq1 = seq1;
    state->seq2 = seq2;
    state->maxdist = max(len1, len2);
    /* Each row represents a value of k, from -maxdist to maxdist. */
    state->fkp_rows = (state->maxdist) * 2 + 1;
    /* Each column represents a value of p, from -1 to maxdist. */
    state->fkp_cols = (state->maxdist) + 2;
    /* fkp is stored as a rectangular array, row by row. */
    state->fkp = malloc(state->fkp_rows * state->fkp_cols * sizeof(int));
    if (state->fkp == NULL)
         return -1;
    state->edits = malloc(state->fkp_rows * state->fkp_cols * sizeof(edist_edit));
    if (state->edits == NULL) {
         free(state->fkp);
         return -1;
    }

    for (k = -(state->maxdist); k != state->maxdist; k++) {
	int p;
	for (p = -1; p != state->maxdist; p++) {
	    if (p == abs(k) - 1) {
		if (k < 0) {
		    set_f_kp(state, k, p, abs(k) - 1);
		} else {
		    set_f_kp(state, k, p, -1);
		}
	    } else if (p < abs(k)) {
		set_f_kp(state, k, p, -INF);
	    }
	    set_edit(state, k, p, ETYPE_NOP, 0);
	}
    }
    return 0;
}

int EXPORT edit_distance(seqcmp_eqfn eqfn, size_t itemsize,
			   void * ptr1, int len1, void * ptr2, int len2)
{
    int p;
    int lendiff;
    edist_state state;
    edist_seq seq1;
    edist_seq seq2;

    seq1.ptr = ptr1;
    seq1.len = len1;
    seq2.ptr = ptr2;
    seq2.len = len2;

    /* Make sure second sequence is longer (or same length) */
    if (seq1.len > seq2.len) {
	edist_seq tmp;
	tmp = seq1;
	seq1 = seq2;
	seq2 = tmp;
    }
    /* Special case for if one or both sequences are empty. */
    if (seq1.len == 0) return seq2.len;

    if (edist_init(&state, eqfn, itemsize, seq1, seq2) < 0)
	return -1;
    lendiff = seq2.len - seq1.len;

    p = lendiff; /* This is the minimum possible edit distance. */
    while(1) {
	int inc;
	int temp_p;
	for (temp_p = 0; temp_p != p; temp_p += 1) {
	    inc = p - temp_p;
	    if (abs(lendiff - inc) <= temp_p) {
		edist_calc_f_kp(&state, lendiff - inc, temp_p);
	    }
	    if (abs(lendiff + inc) <= temp_p) {
		edist_calc_f_kp(&state, lendiff + inc, temp_p);
	    }
	}
	edist_calc_f_kp(&state, lendiff, p);

	if (get_f_kp(&state, lendiff, p) == seq1.len) break;
	p = p + 1;
    }

    free(state.fkp);
    free(state.edits);
    return p;
}
