/* seqcmp.h: Sequence comparison algorithms.
 * Copyright (C) 2003  Richard Boulton
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/** @file seqcmp.h
 *  @brief Public header file defining types and functions.
 *
 *  This header file should be included by any program wishing to make use
 *  of the seqcmp library.  It contains all the typedefs and function
 *  prototypes that are neccessary for inclusion in such programs.
 */

#ifndef __editdistance_h__
#define __editdistance_h__

#include <stdlib.h> /* Should contain definition of size_t */

/** Type of functions used to compare two items in a sequence for equality.
 *  The function should return true (non-zero) if the items are equal and
 *  false (0) if they differ.
 */
typedef int (*seqcmp_eqfn)(void *, void *);

/** Calculate the edit distance between two sequences.
 *
 *  Edit distance is defined as the minimum number of edit operations
 *  required to move from one sequence to another.  The edit operations
 *  considered are:
 *   - Insertion of a character at an arbitrary position.
 *   - Deletion of a character at an arbitrary position.
 *   - Substitution of a character at an arbitrary position.
 *   - Transposition of two neighbouring characters at an arbitrary position
 *     in the string.
 *
 *  @param eqfn A function to compare two items in the sequence.  It should
 *              return true (ie, non-zero) if the items are equal, and 0 if
 *              they differ.
 *  @param itemsize The size, in bytes, of each item in the sequence.  No
 *              support is currently provided for varying item size.
 *  @param ptr1 A pointer to the start of the first sequence.
 *  @param len1 The length of the first sequence.
 *  @param ptr2 A pointer to the start of the second sequence.
 *  @param len2 The length of the second sequence.
 *
 *  @retval The edit distance from one item to the other.
 */
extern int edit_distance(seqcmp_eqfn eqfn, size_t itemsize,
			   void * ptr1, int len1, void * ptr2, int len2);

#endif

