#ifndef __list_h__
#define __list_h__
#include <stddef.h>

/* ---------------------------------------------------------
   LIST Interfaces header file.
   The list type is an opaque data type. Clients aren't
   allowed to use the internal structure of the vectors, only
   the interface to them.
   This file describes the single linked list data structure functions
*/
/*
Define the types of the allocator members
*/
#ifndef _GENERIC_TYPES_DEFINED
#define _GENERIC_TYPES_DEFINED
/* Unary operator. Should have a signature taking a pointer to its input
data, and should leave the results in *result. The return value should be
either NULL, or the result argument.
*/
typedef void *(unary_operator)(void *arg,void *result);
typedef void *(*binary_operator)(void *arg1,void *arg2,void *result);
#endif
/*
Memory allocator/free function pointers
*/
typedef struct _link {
	struct _link *Next;
	char Data[];
} link;

typedef struct _list {
		size_t ElementSize;
        size_t count;                    /* in elements units */
		link *Last;                     /* The last item */
        link *First;                    /* The contents of the list start here */
} list;
/* --------------------------------------------------------- */
/* Allocation                                                */
/* --------------------------------------------------------- */
/* Allocate a new list header.
   Arguments:
              size_t element_size: size of each element in bytes. Required.
              int (*pmalloc)(unsigned int) Defaults to malloc
              void (*pfree)(void *) Defaults to free.
*/
list *new_list(size_t elementsize);
/* Allocates a new list header using default values */
list *new_default_list(void);
/* Copy an existing vector into a freshly allocated one */
list *copy_list(list *);
/* destroy a list releasing all space associated with it */
void dispose_list(list *);
/* --------------------------------------------------------- */
/* Accessing the list properties                           */
/* --------------------------------------------------------- */
size_t get_size_list(list *l);
/* get a pointer to an element */
void *get_element_list(list *l,size_t position);
void *operator [](list *l,size_t position);

/* get a sub range from the given list. Returns a copy */
list *get_subrange_list(list *l,size_t start,size_t end);
/* --------------------------------------------------------- */
/* Setting the list properties                             */
/* --------------------------------------------------------- */
void *set_element_list(list *l,size_t position,void *pNewElement);
void *operator []=(list *l,size_t position,void *pNewElement);
list *push_back_list(list *l,void *pdata);
/* --------------------------------------------------------- */
/* Using a list                                              */
/* --------------------------------------------------------- */


/* Insert into the given vector, an element before the given position
   The result could be a new vector, and the old vector could have been
   freed.
   An example of usage is
	v = insert_list(l,5,data);
*/
list *insert_list(list *l,size_t position,void *pElement);

/* Eliminate element at the given position
   Returns: 1 if OK, 0 if error (position out of range)
*/
list *remove_list(list *l,size_t position);


/* Set the given list to empty.
   Returns the old size of the list;
*/
int clear_list(list *l);


/* If the given lists are equal returns 1, otherwise returns zero. Note that
   neither the links, nor the allocator/free members will be included in
   the comparison. Only element_size, and contents are tested.
   Two NULL lists are considered equal.
*/
int equal_list(list *l1,list *l2);


/* Apply the given function to each of the elements of the list, storing
   the results in a new list. The resulting list will have the same shape,
   size, and malloc/free members as the given vector.
   The called function receives a pointer to each element, and a pointer to
   memory where it can store the its results. Normally it should return the
   pointer to the result, or NULL to stop the iteration.
*/
list *map_list(list *v,unary_operator fn);


/* Call the given function for each element of the list until either the end
   of the list is reached, or the function returns zero.
   Returns a pointer to the last element analyzed or NULL if the end is
   reached.
*/
void *iterate_list(list *l,int (*fn)(void *));

/* Builds a new list with the results of the binary
   function "fn" applied to each element of both lists. The called function
   receives a pointer where it stores the result of the operation.
*/
list *binary_map_list(list *l1,list *l2,binary_operator fn);


/*
Cast list v2 into list v1 using the cast unary operator fn. The contents of
l1 are destroyed.
*/
list *cast_list(list *l1,list *l2,unary_operator fn);

/* Get a pointer to the actual contents of the list
*/
void *get_contents_list(list *l);


/* reverse a list */
list *reverse_list(list *l);

/* concatenate the second list to the first one */
list *concat_list(list *l1,list *l2);

/*
Remove an item from the list at the given position
*/
list *remove_list(list *l,size_t position);

/* add an item to the list, storing it at the beginning of the list */
list *push_front_list(list *l,void *pdata);

/* find the first data element that matches an element in the list */
void *find_list(list *l,void *ElementTofind,int (*fn)(void *anElement,void *ElementToFind));
#endif

