#include "containers.h"
#include "ccl_internal.h"
#ifndef DEFAULT_START_SIZE
#define DEFAULT_START_SIZE 20
#endif
/* Definition of the String Collection type */
struct _ArrayInt {
    ArrayIntInterface *VTable; /* The table of functions */
    size_t count;                  /* number of elements in the array */
    unsigned int Flags;             /* Read-only or other flags */
    int *contents;               /* The contents of the collection */
    size_t capacity;                /* allocated space in the contents vector */
    size_t ElementSize;		/* Size of the elements stored in this array. */
	unsigned timestamp;
    /* Element comparison function */
	CompareFunction CompareFn;
    /* Error function */
    ErrorFunction RaiseError;
	ContainerMemoryManager *Allocator;
} ;
static ArrayInt *Create(size_t elementsize,size_t startsize);

static void *DuplicateElement(ArrayInt *AL,void *str,size_t size,const char *functionName)
{
	void *result;
	if (size == 0)
		return str;
	if (str == NULL) {
		iError.RaiseError(functionName,CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = MALLOC(AL,sizeof(int));
	if (result == NULL) {
		AL->RaiseError(functionName,CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	else memcpy(result,str,size);
	return result;
}

#define AL_READONLY	1
#define AL_IGNORECASE	2
#define CHUNKSIZE 20
static int DefaultArrayIntCompareFunction(const void *left,const void *right,CompareInfo *ExtraArgs)
{
	size_t siz=((ArrayInt *)ExtraArgs->Container)->ElementSize;
	return memcmp(left,right,siz);
}

static size_t GetCount(const ArrayInt *AL)
{
	return AL->count;
}
static unsigned GetFlags(const ArrayInt *AL)
{
	return AL->Flags;
}
static unsigned SetFlags(ArrayInt *AL,unsigned newval)
{
	int oldval = AL->Flags;
	AL->Flags = newval;
	return oldval;
}


static int Resize(ArrayInt *AL)
{
	size_t newcapacity = AL->capacity + 1+AL->capacity/4;
	int *oldcontents = AL->contents;
	AL->contents = AL->Allocator->realloc(AL->contents,newcapacity*sizeof(int));
	if (AL->contents == NULL) {
		AL->RaiseError("Resize",CONTAINER_ERROR_NOMEMORY);
		AL->contents = oldcontents;
		return CONTAINER_ERROR_NOMEMORY;
	}
	AL->capacity = newcapacity;
	AL->timestamp++;
	return 1;
}

static int ResizeTo(ArrayInt *AL,size_t newcapacity)
{
	int *oldcontents;
	
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.ResizeTo",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	oldcontents = AL->contents;
	AL->contents = AL->Allocator->realloc(AL->contents,newcapacity*sizeof(int));
	if (AL->contents == NULL) {
		AL->RaiseError("iArrayInt.Resize",CONTAINER_ERROR_NOMEMORY);
		AL->contents = oldcontents;
		return CONTAINER_ERROR_NOMEMORY;
	}
	AL->capacity = newcapacity;
	AL->timestamp++;
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     Add ID:1
 Purpose:       Adds an item at the end of the ArrayInt
 Input:         The ArrayInt and the item to be added
 Output:        The number of items in the ArrayInt or negative if
                error
 Errors:
------------------------------------------------------------------------*/
static int Add(ArrayInt *AL,int newval)
{
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY) {
		AL->RaiseError("iArrayInt.Add",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}

	if (AL->count >= AL->capacity) {
		int r = Resize(AL);
		if (r <= 0)
			return r;
	}
	AL->contents[AL->count] = newval;
	AL->timestamp++;
	++AL->count;
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     AddRange ID:1
 Purpose:       Adds a range of data at the end of an existing
                arrraylist. If for some reason this process is
                interrupted in the middle, not all items will be
                added.
 Input:         The ArrayInt and the data to be added
 Output:        The number of items added or negative error code
 Errors:        ArrayInt must be writable and data must be
                different of NULL
------------------------------------------------------------------------*/
static int AddRange(ArrayInt * AL,size_t n,int *data)
{
	size_t newcapacity;

	if (AL == NULL) {
		iError.RaiseError("iArrayInt.AddRange",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY) {
		AL->RaiseError("iArrayInt.AddRange",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}
	if (data == NULL) {
		AL->RaiseError("iArrayInt.AddRange",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	newcapacity = AL->count+n;
	if (newcapacity >= AL->capacity-1) {
		int *newcontents;
		newcapacity += AL->count/4;
		newcontents = AL->Allocator->realloc(AL->contents,newcapacity*AL->ElementSize);
		if (newcontents == NULL) {
			AL->RaiseError("iVector.AddRange",CONTAINER_ERROR_NOMEMORY);
			return CONTAINER_ERROR_NOMEMORY;
		}
		AL->capacity = newcapacity;
		AL->contents = newcontents;
	}
	memcpy(AL->contents+AL->count,data,n*AL->ElementSize);
	AL->count += n;
	AL->timestamp++;
	return 1;
}

static ArrayInt *GetRange(ArrayInt *AL, size_t start,size_t end)
{
	ArrayInt *result;
	
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.GetRange",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = AL->VTable->Create(sizeof(int),AL->count);
	result->VTable = AL->VTable;
	if (AL->count == 0)
		return result;
	if (end >= AL->count)
		end = AL->count-1;
	if (start > end)
		return result;
	memcpy(result->contents,&AL->contents[start],sizeof(int)*(end-start));
	result->count = end-start;
	return result;
}

/*------------------------------------------------------------------------
 Procedure:     Clear ID:1
 Purpose:       Frees all memory from an array list object without
                freeing the object itself
 Input:         The array list to be cleared
 Output:        The number of objects that were in the array list
 Errors:        The array list must be writable
------------------------------------------------------------------------*/
static int Clear(ArrayInt *AL)
{
	if (AL->Flags & AL_READONLY)
		return 0;
	AL->Allocator->free(AL->contents);
	AL->contents = NULL;
	AL->count = 0;
	AL->timestamp = 0;
	AL->Flags = 0;
	AL->capacity = 0;
	return 1;
}

static int Contains(ArrayInt *AL,int data,void *ExtraArgs)
{
	size_t i;
	char *p;
	CompareInfo ci;

	if (AL == NULL) {
		iError.RaiseError("iArrayInt.Copy",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	p = (char *)AL->contents;
	if (ExtraArgs == NULL) {
		ExtraArgs = &ci;
		ci.Container = AL;
		ci.ExtraArgs = NULL;
	}
	for (i = 0; i<AL->count;i++) {
		if (AL->contents[i] == data)
			return 1;
	}
	return 0;
}

static int Equal(ArrayInt *AL1,ArrayInt *AL2)
{
	if (AL1 == AL2)
		return 1;
	if (AL1 == NULL || AL2 == NULL)
		return 0;
	if (AL1->count != AL2->count)
		return 0;
	if (AL1->Flags != AL2->Flags)
		return 0;
	if (AL1->ElementSize != AL2->ElementSize)
		return 0;
	if (AL1->count == 0)
		return 1;
	if (memcmp(AL1->contents,AL2->contents,sizeof(int)*AL1->count) != 0)
		return 0;
	return 1;
}

static ArrayInt *Copy(ArrayInt *AL)
{
	ArrayInt *result;
	size_t startsize,es;
	
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.Copy",CONTAINER_ERROR_BADARG);
		return NULL;
	}

	result = AL->Allocator->malloc(sizeof(*result));
	if (result == NULL) {
		iError.RaiseError("iArrayInt.Copy",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	memset(result,0,sizeof(*result));
	startsize = AL->count;
	if (startsize <= 0)
		startsize = DEFAULT_START_SIZE;
	es = startsize * sizeof(int);
	result->contents = CurrentMemoryManager->malloc(es);
	if (result->contents == NULL) {
		iError.RaiseError("newArrayInt",CONTAINER_ERROR_NOMEMORY);
		AL->Allocator->free(result);
		result = NULL;
	}
	else {
		memset(result->contents,0,es);
		result->capacity = startsize;
		result->VTable = &iArrayInt;
		result->ElementSize = sizeof(int);
		result->CompareFn = DefaultArrayIntCompareFunction;
		result->RaiseError = iError.RaiseError;
		result->Allocator = AL->Allocator;
	}

	if (result == NULL) {
		AL->RaiseError("iArrayInt.Copy",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	memcpy(result->contents,AL->contents,AL->count*sizeof(int));
	result->CompareFn = AL->CompareFn;
	result->Flags = AL->Flags;
	result->RaiseError = AL->RaiseError;
	result->VTable = AL->VTable;
	return result;
}

static int CopyElement(ArrayInt *AL,size_t idx, int *outbuf)
{
	if (AL == NULL || outbuf == NULL) {
		if (AL)
			AL->RaiseError("iArrayInt.CopyElement",CONTAINER_ERROR_BADARG);
		else
			iError.RaiseError("iArrayInt.CopyElement",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (idx >= AL->count) {
		AL->RaiseError("iArrayInt.CopyElement",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	*outbuf = AL->contents[idx];
	return 1;
}

static void **CopyTo(ArrayInt *AL)
{
	void **result;
	int *p;
		
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.CopyTo",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = MALLOC(AL,(1+AL->count)*sizeof(int));
	if (result == NULL) {
		AL->RaiseError("iArrayInt.CopyTo",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	p = AL->contents;
	p[0] = AL->count;
	memcpy(p+1,AL->contents,AL->count*sizeof(int));
	return result;
}

static int IndexOf(ArrayInt *AL,int data,void *ExtraArgs,size_t *result)
{
	size_t i;

	if (AL == NULL) {
		iError.RaiseError("iArrayInt.IndexOf",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	for (i=0; i<AL->count;i++) {
		if (AL->contents[i] == data) {
			*result = i;
			return 1;
		}
	}
	return CONTAINER_ERROR_NOTFOUND;
}

static int GetElement(ArrayInt *AL,size_t idx)
{
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.GetElement",CONTAINER_ERROR_BADARG);
		return 0;
	}
	if (idx >=AL->count ) {
		AL->RaiseError("GetElement",CONTAINER_ERROR_INDEX);
		return 0;
	}
	return AL->contents[idx];
}

/*------------------------------------------------------------------------
 Procedure:     InsertAt ID:1
 Purpose:       Inserts data at the given position, that should be
                between zero and the number of items in the array
                list. If the index is equal to the count of items in
                the array the item will be inserted at the end. Any
                insertion in the middle or at the beginning provokes
                a move of all items between the index and the end of
                the array list
 Input:         The array list, the index and the new data to be
                inserted
 Output:        Error code less than zero if error, or positive
                number bigger than zero otherwise
 Errors:        The index must be correct and the array list must be
                writable
------------------------------------------------------------------------*/
static int InsertAt(ArrayInt *AL,size_t idx,int newval)
{
	int *p;
	
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.InsertAt",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (idx > AL->count) {
		AL->RaiseError("iArrayInt.InsertAt",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	if (AL->count >= (AL->capacity-1)) {
		int r = Resize(AL);
		if (r <= 0)
			return r;
	}
	p = AL->contents;
	if (idx < AL->count) {
		memmove(p+(idx+1),
				p+idx,
				(AL->count-idx+1)*sizeof(int));
	}
	p[idx] = newval;
	AL->timestamp++;
	++AL->count;
	return 1;
}

static int Insert(ArrayInt *AL,int 	newval)
{
	return InsertAt(AL,0,newval);
}

static int InsertIn(ArrayInt *AL, size_t idx, ArrayInt *newData)
{
	size_t newCount;
	char *p;

	if (AL == NULL || newData == NULL) {
		iError.RaiseError("iArrayInt.InsertIn",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (idx > AL->count) {
		AL->RaiseError("iArrayInt.InsertIn",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	if (sizeof(int) != newData->ElementSize)
		return CONTAINER_ERROR_INCOMPATIBLE;
	newCount = AL->count + newData->count;
	if (newCount >= (AL->capacity-1)) {
		int r = ResizeTo(AL,newCount);
		if (r <= 0)
			return r;
	}
	p = (char *)AL->contents;
	if (idx < AL->count) {
		memmove(p+sizeof(int)*(idx+newData->count),
				p+sizeof(int)*idx,
				(AL->count-idx)*sizeof(int));
	}
	p += idx*sizeof(int);
	memcpy(p,newData->contents,newData->ElementSize*newData->count);
	AL->timestamp++;
	AL->count = newCount;
	return 1;
}

static int EraseAt(ArrayInt *AL,size_t idx)
{
	int *p;
	if (idx >= AL->count)
		return CONTAINER_ERROR_INDEX;
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (AL->count == 0)
		return -2;
	p = AL->contents;
	if (idx < (AL->count-1)) {
		memmove(p+idx,p+(idx+1),(AL->count-idx)*sizeof(int));
	}
	AL->count--;
	AL->timestamp++;
	return 1;
}

static int Remove(ArrayInt *AL,int element)
{
	size_t idx;
	int r = IndexOf(AL,element,NULL,&idx);
	if (r < 0)
		return r;
	return EraseAt(AL,idx);
}

static int PushBack(ArrayInt *AL,int str)
{
	return InsertAt(AL,AL->count,str);
}

/*------------------------------------------------------------------------
 Procedure:     Pop ID:1
 Purpose:       Removes the last item of the list and returns its
                data to the caller. It is the responsability of the
                caller to free this memory. This is almost the same as
				the "Remove" function, with the only difference that the
				data is NOT freed but returned to the caller.
 Input:         The array list to be popped
 Output:        a pointer to the last item's data
 Errors:        If the list is read-only or empty returns NULL
------------------------------------------------------------------------*/
static int PopBack(ArrayInt *AL,int *result)
{
    int *p;

    if (AL == NULL)
        return CONTAINER_ERROR_BADARG;
    if (AL->Flags&AL_READONLY) 
	return CONTAINER_ERROR_READONLY;
    if (AL->count == 0)
            return 0;
    AL->count--;
    if (result) {
        p = AL->contents;
		*result = p[AL->count];
    }
    AL->timestamp++;
    return 1;
}

/*------------------------------------------------------------------------
 Procedure:     Finalize ID:1
 Purpose:       Releases all memory associated with an ArrayInt,
                including the ArrayInt structure itself
 Input:         The ArrayInt to be destroyed
 Output:        The number of items that the ArrayInt contained or
                zero if error
 Errors:        Input must be writable
------------------------------------------------------------------------*/
static int Finalize(ArrayInt *AL)
{
	int result = Clear(AL);
	AL->Allocator->free(AL);
	return result;
}

static size_t GetCapacity(const ArrayInt *AL)
{
	return AL->capacity;
}

static int SetCapacity(ArrayInt *AL,size_t newCapacity)
{
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.SetCapacity",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->count != 0)
		return 0;
	if (newCapacity == 0)
		return CONTAINER_ERROR_BADARG;
	AL->Allocator->free(AL->contents);
	AL->contents = MALLOC(AL,newCapacity*sizeof(int));
	if (AL->contents == NULL) {
		AL->RaiseError("iArrayInt.SetCapacity",CONTAINER_ERROR_NOMEMORY);
		return 0;
	}
	memset(AL->contents,0,sizeof(void *)*newCapacity);
	AL->capacity = newCapacity;
	return 1;
}

static int Apply(ArrayInt *AL,int (*Applyfn)(int,void *),void *arg)
{
	size_t i;
	int *p;

	if (AL == NULL || Applyfn == NULL) {
		if (AL)
			AL->RaiseError("iArrayInt.ReplaceAt",CONTAINER_ERROR_BADARG);
		else
			iError.RaiseError("iArrayInt.ReplaceAt",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	p=AL->contents;
	for (i=0; i<AL->count;i++) {
		Applyfn(*p,arg);
		p++;
	}
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     ReplaceAt ID:1
 Purpose:       Replace the data at the given position with new
                data.
 Input:         The list and the new data
 Output:        Returns the new data if OK, NULL if error.
 Errors:        Index error or read only errors are caught
------------------------------------------------------------------------*/
static int ReplaceAt(ArrayInt *AL,size_t idx,int newval)
{
	if (AL == NULL) {
		if (AL)
			AL->RaiseError("iArrayInt.ReplaceAt",CONTAINER_ERROR_BADARG);
		else
			iError.RaiseError("iArrayInt.ReplaceAt",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY) {
		AL->RaiseError("iArrayInt.ReplaceAt",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}
	if (idx >= AL->count) {
		AL->RaiseError("iArrayInt.ReplaceAt",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	AL->contents[idx] = newval;
	AL->timestamp++;
	return 1;
}

static ArrayInt *IndexIn(ArrayInt *SC,ArrayInt *AL)
{
	ArrayInt *result = NULL;
	size_t i,top,idx;
	int r;

	if (SC == NULL || AL == NULL) {
		iError.RaiseError("iArrayInt.IndexIn",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	top = iArrayInt.Size(AL);
	result = iArrayInt.Create(SC->ElementSize,top);
	for (i=0; i<top;i++) {
		idx = *(size_t *)iArrayInt.GetElement(AL,i);
		r = GetElement(SC,idx);
		r = Add(result,r);
		if (r < 0)
			break;
	}
	return result;
}
static ErrorFunction SetErrorFunction(ArrayInt *AL,ErrorFunction fn)
{
	ErrorFunction old;
	old = AL->RaiseError;
	if (fn) AL->RaiseError = fn;
	return old;
}

static size_t Sizeof(ArrayInt *AL)
{
	if (AL == NULL)
		return sizeof(ArrayInt);
	return sizeof(ArrayInt) + (AL->count * sizeof(int));
}

/*------------------------------------------------------------------------
 Procedure:     SetCompareFunction ID:1
 Purpose:       Defines the function to be used in comparison of
                ArrayInt elements
 Input:         The ArrayInt and the new comparison function. If the new
                comparison function is NULL no changes are done.
 Output:        Returns the old function
 Errors:        None
------------------------------------------------------------------------*/
static CompareFunction SetCompareFunction(ArrayInt *l,CompareFunction fn)
{
	CompareFunction oldfn = l->CompareFn;

	if (fn != NULL) /* Treat NULL as an enquiry to get the compare function */
		l->CompareFn = fn;
	return oldfn;
}

static int Sort(ArrayInt *AL)
{
	CompareInfo ci;

	if (AL == NULL) {
		iError.RaiseError("iArrayInt.Sort",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	ci.Container = AL;
	ci.ExtraArgs = NULL;
	qsortEx(AL->contents,AL->count,sizeof(int),AL->CompareFn,&ci);
	return 1;
}

/* Proposed by PWO 
*/
static int Append(ArrayInt *AL1, ArrayInt *AL2)
{
	size_t newCount;
	char *p;

	if (AL1 == NULL) {
		iError.RaiseError("iArrayInt.Append",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL2 == NULL) {
		AL1->RaiseError("iArrayL.Append",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if ((AL1->Flags & AL_READONLY) || (AL2->Flags & AL_READONLY)) {
		AL1->RaiseError("iArrayInt.Append",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}
	if (AL2->ElementSize != AL1->ElementSize) {
		AL1->RaiseError("iArrayInt.Append",CONTAINER_ERROR_INCOMPATIBLE);
		return CONTAINER_ERROR_INCOMPATIBLE;
	}
	newCount = AL1->count + AL2->count;
	if (newCount >= (AL1->capacity-1)) {
		int r = ResizeTo(AL1,newCount);
		if (r <= 0)
			return r;
	}
	p = (char *)AL1->contents;
	p += AL1->count*AL1->ElementSize;
	memcpy(p,AL2->contents,AL2->ElementSize*AL2->count);
	AL1->timestamp++;
	AL1->count = newCount;
	AL2->Allocator->free(AL2);
    return 1;
}

/* Proposed by PWO 
*/
static int Reverse(ArrayInt *AL)
{
	size_t s;
	int *p, *q, *t;

	if (AL == NULL) {
		iError.RaiseError("iArrayInt.Reverse",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY) {
		AL->RaiseError("iArrayInt.Reverse",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}
	if (AL->count < 2)
		return 1;

	s = sizeof(int);
	p = AL->contents;
	q = p + s*(AL->count-1);
	t = malloc(s);
	if (t == NULL) {
		AL->RaiseError("iArrayInt.Reverse",CONTAINER_ERROR_NOMEMORY);
		return CONTAINER_ERROR_NOMEMORY;
	}
	while ( p < q ) {
		memcpy(t, p, s);
		memcpy(p, q, s);
		memcpy(q, t, s);
		p += s;
		q -= s;
	}
	free(t);
	AL->timestamp++;
	return 1;
}


/* ------------------------------------------------------------------------------ */
/*                                Iterators                                       */
/* ------------------------------------------------------------------------------ */

struct ArrayIntIterator {
	Iterator it;
	ArrayInt *AL;
	size_t index;
	size_t timestamp;
	unsigned long Flags;
	void *Current;
};

static void *GetNext(Iterator *it)
{
	struct ArrayIntIterator *ali = (struct ArrayIntIterator *)it;
	ArrayInt *AL = ali->AL;
	int *p;

	if (ali->index >= AL->count)
		return NULL;
	if (ali->timestamp != AL->timestamp) {
		AL->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	p = AL->contents;
	p += ali->index;
	ali->index++;
	ali->Current = p;
	return p;
}

static void *GetPrevious(Iterator *it)
{
	struct ArrayIntIterator *ali = (struct ArrayIntIterator *)it;
	ArrayInt *AL = ali->AL;
	int *p;

	if (ali->index >= AL->count || ali->index == 0)
		return NULL;
	if (ali->timestamp != AL->timestamp) {
		AL->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	p = AL->contents;
	p += ali->index;
	ali->index--;
	ali->Current = p;
	return p;
}

static void *GetLast(Iterator *it)
{
	struct ArrayIntIterator *ali = (struct ArrayIntIterator *)it;
	ArrayInt *AL = ali->AL;
	int *p;

	if (AL->count == 0)
		return NULL;
	if (ali->timestamp != AL->timestamp) {
		AL->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	p = AL->contents;
	p += (AL->count-1) ;
	ali->index = AL->count-1;
	ali->Current = p;
	return p;
}

static void *GetCurrent(Iterator *it)
{
	struct ArrayIntIterator *ali = (struct ArrayIntIterator *)it;
	return ali->Current;
}

static void *GetFirst(Iterator *it)
{
	struct ArrayIntIterator *ali = (struct ArrayIntIterator *)it;
	if (ali->AL->count == 0) {
		ali->Current = NULL;
		return NULL;
	}
	ali->index = 1;
	ali->Current = ali->AL->contents;
	return ali->Current;
}

static Iterator *newIterator(ArrayInt *AL)
{
	struct ArrayIntIterator *result = MALLOC(AL,sizeof(struct ArrayIntIterator));
	if (result == NULL)
		return NULL;
	result->it.GetNext = GetNext;
	result->it.GetPrevious = GetPrevious;
	result->it.GetFirst = GetFirst;
	result->it.GetCurrent = GetCurrent;
	result->it.GetLast = GetLast;
	result->AL = AL;
	result->Current = NULL;
	result->timestamp = AL->timestamp;
	return &result->it;
}

static int deleteIterator(Iterator * it)
{
	struct ArrayIntIterator *ali = (struct ArrayIntIterator *)it;
	ali->AL->Allocator->free(it);
	return 1;
}

static size_t DefaultSaveFunction(const void *element,void *arg, FILE *Outfile)
{
	const unsigned char *str = element;
	size_t len = *(size_t *)arg;

	return fwrite(str,1,len,Outfile);
}

static size_t DefaultLoadFunction(void *element,void *arg, FILE *Infile)
{
	size_t len = *(size_t *)arg;

	return fread(element,1,len,Infile);
}

static int Save(ArrayInt *AL,FILE *stream, SaveFunction saveFn,void *arg)
{
	size_t i;


	if (AL == NULL) {
		iError.RaiseError("iArrayInt.Save",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (stream == NULL) {
		AL->RaiseError("iArrayInt.Load",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (saveFn == NULL) {
		saveFn = DefaultSaveFunction;
	}
	if (fwrite(AL,1,sizeof(ArrayInt),stream) <= 0)
		return EOF;
	for (i=0; i< AL->count; i++) {
		int *p = AL->contents;

		p += i*sizeof(int);
		if (saveFn(p,arg,stream) <= 0)
			return EOF;
	}
	return 1;
}

static ArrayInt *Load(FILE *stream, ReadFunction loadFn,void *arg)
{
	size_t i;
	int *p;
	ArrayInt *result,AL;

	if (stream == NULL) {
		iError.RaiseError("iArrayInt.Load",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	if (loadFn == NULL) {
		loadFn = DefaultLoadFunction;
		arg = &AL.ElementSize;
	}
	if (fread(&AL,1,sizeof(ArrayInt),stream) <= 0)
		return NULL;
	p = malloc(AL.ElementSize);
	if (p == NULL)
		return NULL;
	result = Create(AL.ElementSize,AL.count);
	if (result) {
		result->Flags = AL.Flags;
		for (i=0; i< AL.count; i++) {
			if (loadFn(p,arg,stream) <= 0) {
				iError.RaiseError("iArrayInt.Load",CONTAINER_ERROR_FILE_READ);
				Finalize(result);
				result = NULL;
				break;
			}
			if (Add(result,*p) < 0) {
				Finalize(result);
				result = NULL;
				break;
			}
		}
	}
	free(p);
	return result;
}



static ArrayInt *Create(size_t elementsize,size_t startsize)
{
	ArrayInt *result;
	size_t es;
	
	/* This detects BOGUS sizes arising from passing a negative value */
	if ((int)elementsize <= 0) {
		iError.RaiseError("iArrayInt.Create",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	/* Round the element size to correctly align table elements.
	 This supposes that sizeo(void *) is a power of two
	 */
	elementsize = roundup(elementsize);
	/* Allocate space for the array list header structure */
	result = CurrentMemoryManager->malloc(sizeof(*result));
	if (result == NULL) {
		iError.RaiseError("iArrayInt.Create",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	memset(result,0,sizeof(*result));
	if (startsize <= 0)
		startsize = DEFAULT_START_SIZE;
	es = startsize * elementsize;
	result->contents = CurrentMemoryManager->malloc(es);
	if (result->contents == NULL) {
		iError.RaiseError("newArrayInt",CONTAINER_ERROR_NOMEMORY);
		CurrentMemoryManager->free(result);
		result = NULL;
	}
	else {
		memset(result->contents,0,es);
		result->capacity = startsize;
		result->VTable = &iArrayInt;
		result->ElementSize = elementsize;
		result->CompareFn = DefaultArrayIntCompareFunction;
		result->RaiseError = iError.RaiseError;
		result->Allocator = CurrentMemoryManager;
	}
	return result;
}

static size_t GetElementSize(const ArrayInt *AL)
{
	if (AL == NULL) {
		iError.RaiseError("iArrayInt.GetElementSize",CONTAINER_ERROR_BADARG);
		return 0;
	}
	return sizeof(int);
}

ArrayIntInterface iArrayInt = {
	GetCount,
	GetFlags, 
	SetFlags, 
	Clear,
	Contains,
	Remove, 
	Finalize,
	Apply,
	Equal,
	Copy,
	SetErrorFunction,
	Sizeof,
	newIterator,
	deleteIterator,
	Save,
	Load,
	GetElementSize,
/* Sequential container fields */
	Add, 
	GetElement,
	PushBack,
	PopBack,
	InsertAt,
	EraseAt, 
	ReplaceAt,
	IndexOf, 
/* ArrayInt specific fields */
	Insert,
	InsertIn,
	IndexIn,
	GetCapacity, 
	SetCapacity,
	SetCompareFunction,
	Sort,
	Create,

	AddRange, 
	GetRange,
	CopyElement,
	CopyTo, 
	Reverse,
	Append,
};

