#include "str.h"
#include "exceptions.h"
#include "str_private.h"
/* copy Function */

/** @section Strcpy Strcpy Family */
/** @addtogroup Copiers Copy Functions
@{
The purpose of all the Str(n)cpy function is to copy elements from
@a source to @a target. The following extensios were added:

- It is guaranteed that target has enough space for
the elements to be copied, so it's save to use Strcpy.
- The copy onto an InvalidString is well defined and will
yield a copy of the source string
- If the 'count' parameter is larger than the length of a
String to be copied an exception is raised. We do not fill
out extra space with '\0' as strncpy does.
- The Strc?cpy functions to not return a value. It's not
clear what  return value may give us if the target
was clobbered anyway.
*/

/** Examples can be found in @example examples/strcpy_examples.c
*/

/* static size_t adjust_count(String & string, size_t count){
size_t result = 0;
if (Strlen(string) < count){
result = Strlen(string);
} else {
result = count;
}
return result;
}
*/
/** @anchor Strncpy_fun
Copies exactly count elements from source to target
@param target of the copy. The target does not have to be
a proper String, and does not have to had enough space for the copy.
Resizing takes place as needed. If @a count is longer than
the len of either the String, char* or wchar_t *, an excption
will get risen.
@param source of the copy
@param count this many elements should get copied.
@pre count <= Strlen (source)
@pre Strvalid(source) or source != NULL
*/
void EXPORT overloaded Strncpy(StringA & target, StringA &source, size_t count)
{
	require(StrvalidA(source));
	require(source.count >= count);
	char *content;

	count = (count < source.count) ? count : source.count;
	if (target.capacity < count){
		target.capacity = calculate_new_capacity(target.capacity,count+1);
		target.content = allocate_proper_space(target.capacity,1);
	}
	target.count = count;
	strncpy(target.content, source.content, count);
	target.content[target.count] = 0;
}

void EXPORT overloaded Strncpy(StringW & target, StringW &source, unsigned int count)
{
	require(StrvalidW(source));
	require(source.count >= count);
	int code = 0;
	char *content;

	count = (count < source.count) ? count : source.count;
	if (target.capacity < count){
		target.capacity = calculate_new_capacity(target.capacity,count+1);
		target.content = allocate_proper_space(target.capacity,2);
	}
	wcsncpy(target.content,source.content, count);
	target.content[target.count] = 0;
}


/** @ref Strcpy family
Strncpy was overloaded to allow mixing
String, char* and wchar_t * types.
@param target of the copy.
@param source of the copy.
@param count elements to be copied.
@pre source != NULL
@pre len (source) >= count)
@pre count >= strlen(source)
see also @ref Strncpy_fun
*/
void EXPORT overloaded Strncpy(StringA & target, char* source, unsigned int count)
{
	require(source != NULL);
	int code = 0,slen;
	char *content;

	slen = strlen(source);
	count = (count < slen) ? count : slen;
	if (target.capacity < count){
		target.capacity = calculate_new_capacity(target.capacity,count+1);
		content = allocate_proper_space(target.capacity,1);
	}
	strncpy(target.content,source, count);
	target.content[target.count] = 0;
}


/** @ref Strcpy family
For the contstrains and meaning of the parameters see
@ref Strncpy_fun */
void EXPORT overloaded Strncpy(StringW & target, wchar_t * source, unsigned int count)
{
	require(source != NULL && wcslen(source) >= count);
	int code;

	if (target.capacity < count){
		Strresize(target,count+1);
	}
	wcsncpy (target.content, source, count);
	target.count = count;
	require(target.capacity > count);
}

/** @ref Strcpy family */
/** Safe Strcpy functions. Bounds are always checked and it will be
ensured that target can take all the needed Elements
@param target of the copy.
@param source of the copy.
@notes @a target is an I/O Parameter. It will always be proper copy
of source
*/

void EXPORT overloaded Strcpy(StringA & target, StringA & source)
{
	int code = 0;
	require(StrvalidA(source));

	if (!Strvalid(target)){
		target.count = 0;
		target.capacity = 0;

	}
	if (target.capacity < source.count) {
		int new_capacity = calculate_new_capacity(target.capacity,source.count+1);
		char *contents = allocate_proper_space(new_capacity,1);

		target.capacity = new_capacity;
		target.content = contents;
	}
	target.count = source.count;
	memcpy(target.content,source.content,source.count);
	target.content[target.count] = 0;
	require(target.capacity >= source.count);
}

void EXPORT overloaded Strcpy(StringW & target, StringW & source)
{
	require(StrvalidW(source));
	int code = 0;

	if (!StrvalidW(target)){
		target.count = 0;
		target.capacity= 0;

	}
	if (target.capacity < source.count) {
		int new_capacity = calculate_new_capacity(target.capacity,source.count+1);
		wchar_t *contents = allocate_proper_space(new_capacity,sizeof(wchar_t));

		if (contents == NULL) {
			return;
		}
		target.content = contents;
	}
	target.count = source.count;
	memcpy(target.content,source.content,source.count*sizeof(wchar_t));
	target.content[target.count] = 0;
	require(target.capacity >= source.count);
}

/** @ref Strcpy Family */
void EXPORT overloaded Strcpy(StringW & target, wchar_t *source)
{
	require(source != NULL);
	int code = 0,srclen;

	srclen = wcslen(source);
	if (target.capacity < srclen) {
		int new_capacity = calculate_new_capacity(target.capacity,srclen+1);
		wchar_t *contents = allocate_proper_space(new_capacity,sizeof(wchar_t));

		if (contents == NULL) {
			return;
		}
		target.content = contents;
	}
	target.count = srclen;
	memcpy(target.content,source,srclen);
	target.content[target.count] = 0;
	require(target.capacity >= srclen);
}

void EXPORT overloaded Strcpy(StringA & target, char *source)
{
	int code = 0,srclen;
	require(source != NULL);

	srclen = strlen(source);
	if (target.capacity < srclen) {
		int new_capacity = calculate_new_capacity(target.capacity,srclen+1);
		char *contents = allocate_proper_space(new_capacity,1);

		if (contents == NULL) {
			return;
		}
		target.content = contents;
	}
	target.count = srclen;
	memcpy(target.content,source,srclen);
	target.content[target.count] = 0;
	require(target.capacity >= srclen);
}

