#include "containers.h"
#ifndef DEFAULT_START_SIZE
#define DEFAULT_START_SIZE 20
#endif
/* Definition of the String Collection type */
struct _ArrayList {
    ArrayListInterface *VTable; /* The table of functions */
    size_t count;                  /* number of elements in the array */
    unsigned int Flags;             /* Read-only or other flags */
    void *contents;               /* The contents of the collection */
    size_t capacity;                /* allocated space in the contents vector */
    size_t ElementSize;		/* Size of the elements stored in this array. */
	unsigned timestamp;
    /* Element comparison function */
	CompareFunction CompareFn;
    /* Error function */
    ErrorFunction RaiseError;
} ;
static ArrayList *Create(size_t elementsize,size_t startsize);

static void *DuplicateElement(ArrayList *AL,void *str,size_t size,const char *functionName)
{
	void *result;
	if (size == 0)
		return str;
	if (str == NULL) {
		iError.RaiseError(functionName,CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = MALLOC(AL,size);
	if (result == NULL) {
		iError.RaiseError(functionName,CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	else memcpy(result,str,size);
	return result;
}

#define AL_READONLY	1
#define AL_IGNORECASE	2
#define CHUNKSIZE 20
static int DefaultArrayListCompareFunction(const void *left,const void *right,CompareInfo *ExtraArgs)
{
	size_t siz=((ArrayList *)ExtraArgs->Object)->ElementSize;
	return memcmp(left,right,siz);
}

#define roundup(x,n) (((x)+((n)-1))&(~((n)-1)))
static size_t GetCount(const ArrayList *AL)
{
	return AL->count;
}
static unsigned GetFlags(const ArrayList *AL)
{
	return AL->Flags;
}
static unsigned SetFlags(ArrayList *AL,unsigned newval)
{
	int oldval = AL->Flags;
	AL->Flags = newval;
	return oldval;
}


static int Resize(ArrayList *AL)
{
	size_t newcapacity = AL->capacity + 1+AL->capacity/4;
	void **oldcontents = AL->contents;
	AL->contents = REALLOC(AL,AL->contents,newcapacity*AL->ElementSize);
	if (AL->contents == NULL) {
		AL->RaiseError("Resize",CONTAINER_ERROR_NOMEMORY);
		AL->contents = oldcontents;
		return CONTAINER_ERROR_NOMEMORY;
	}
	AL->capacity = newcapacity;
	AL->timestamp++;
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     Add ID:1
 Purpose:       Adds an item at the end of the arraylist
 Input:         The arraylist and the item to be added
 Output:        The number of items in the arraylist or negative if
                error
 Errors:
------------------------------------------------------------------------*/
static int Add(ArrayList *AL,void *newval)
{
	char *p;

	if (AL == NULL) {
		iError.RaiseError("iArrayList.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (newval == NULL) {
		AL->RaiseError("Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->count >= AL->capacity) {
		int r = Resize(AL);
		if (r <= 0)
			return r;
	}
	p = AL->contents;
	p += AL->count*AL->ElementSize;
	memcpy(p,newval,AL->ElementSize);
	AL->timestamp++;
	++AL->count;
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     AddRange ID:1
 Purpose:       Adds a range of data at the end of an existing
                arrraylist. If for some reason this process is
                interrupted in the middle, not all items will be
                added.
 Input:         The arraylist and the data to be added
 Output:        The number of items added or negative error code
 Errors:        Arraylist must be writable and data must be
                different of NULL
------------------------------------------------------------------------*/
static int AddRange(ArrayList * AL,void **data)
{
	int i = 0;

	if (AL == NULL) {
		iError.RaiseError("iArrayList.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (data == NULL) {
		AL->RaiseError("AddRange",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	while (data[i] != NULL) {
		int r = Add(AL,data[i]);
		if (r < 0) {
			return r;
		}
		i++;
	}
	AL->timestamp++;
	return i;
}

static ArrayList *GetRange(ArrayList *AL, size_t start,size_t end)
{
	ArrayList *result;
	void *p;
	unsigned oldFlags;
	
	if (AL == NULL) {
		iError.RaiseError("iArrayList.Add",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = AL->VTable->Create(AL->ElementSize,AL->count);
	result->VTable = AL->VTable;
	if (AL->count == 0)
		return result;
	if (end >= AL->count)
		end = AL->count-1;
	if (start > end)
		return result;
	oldFlags = AL->Flags;
	AL->Flags = (unsigned)(~AL_READONLY);
	while (start <= end) {
		p = AL->VTable->GetElement(AL,start);
		result->VTable->Add(result,p);
		start++;
	}
	AL->Flags = result->Flags = oldFlags;
	return result;
}

/*------------------------------------------------------------------------
 Procedure:     Clear ID:1
 Purpose:       Frees all memory from an array list object without
                freeing the object itself
 Input:         The array list to be cleared
 Output:        The number of objects that were in the array list
 Errors:        The array list must be writable
------------------------------------------------------------------------*/
static int Clear(ArrayList *AL)
{
	if (AL->Flags & AL_READONLY)
		return 0;
	FREE(AL,AL->contents);
	AL->contents = NULL;
	AL->count = 0;
	AL->timestamp = 0;
	AL->Flags = 0;
	AL->capacity = 0;
	return 1;
}

static bool Contains(ArrayList *AL,void *data,void *ExtraArgs)
{
	size_t i;
	char *p;
	CompareInfo ci;

	if (data == NULL) {
		AL->RaiseError("Contains",CONTAINER_ERROR_BADARG);
		return 0;
	}
	p = (char *)AL->contents;
	if (ExtraArgs == NULL) {
		ExtraArgs = &ci;
		ci.Object = AL;
		ci.ExtraArgs = NULL;
	}
	for (i = 0; i<AL->count;i++) {
		if (!AL->CompareFn(p,data,ExtraArgs))
			return 1;
		p += AL->ElementSize;
	}
	return 0;
}

static int Equal(ArrayList *AL1,ArrayList *AL2)
{
	if (AL1 == NULL && AL2 == NULL)
		return 1;
	if (AL1 == NULL || AL2 == NULL)
		return 0;
	if (AL1 == AL2)
		return 1;
	if (AL1->count != AL2->count)
		return 0;
	if (AL1->Flags != AL2->Flags)
		return 0;
	if (AL1->ElementSize != AL2->ElementSize)
		return 0;
	if (AL1->count == 0)
		return 1;
	if (memcmp(AL1->contents,AL2->contents,AL1->ElementSize*AL1->count) != 0)
		return 0;
	return 1;
}

static ArrayList *Copy(ArrayList *AL)
{
	ArrayList *result = iArrayList.Create(AL->ElementSize,AL->count);
	if (result == NULL) {
		AL->RaiseError("Copy",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	memcpy(result->contents,AL->contents,AL->count*AL->ElementSize);
	return result;
}

static int CopyElement(ArrayList *AL,size_t idx, void *outbuf)
{
	char *p;
	if (AL == NULL || outbuf == NULL) {
		iError.RaiseError("iArrayList.CopyElement",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (idx >= AL->count) {
		AL->RaiseError("iArrayList.CopyElement",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	p = AL->contents;
	p += (1+idx)*AL->ElementSize;
	memcpy(outbuf,p,AL->ElementSize);
	return 1;
}

static void **CopyTo(ArrayList *AL)
{
	void **result = MALLOC(AL,(1+AL->count)*sizeof(void *));
	size_t i;
	char *p;
	if (result == NULL) {
		AL->RaiseError("CopyTo",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	p = AL->contents;
	for (i=0; i<AL->count;i++) {
		result[i] = DuplicateElement(AL,p,AL->ElementSize,"Copy");
		if (result[i] == NULL) {
			while (i > 0) {
				FREE(AL,result[i-1]);
				i--;
			}
			FREE(AL,result);
			return NULL;
		}
		p += AL->ElementSize;
	}
	result[i] = NULL;
	return result;
}

static size_t IndexOf(ArrayList *AL,void *data,void *ExtraArgs)
{
	size_t i;
	char *p;
	CompareInfo ci;

	if (AL == NULL) {
		iError.RaiseError("iArrayList.IndexOf",CONTAINER_ERROR_BADARG);
		return 0;
	}
	if (data == NULL)
		return 0;
	p = AL->contents;
	ci.Object = AL;
	ci.ExtraArgs = ExtraArgs;
	for (i=0; i<AL->count;i++) {
		if (!AL->CompareFn(p,data,&ci)) {
			return i+1;
		}
		p += AL->ElementSize;
	}
	return 0;
}

static void *GetElement(ArrayList *AL,size_t idx)
{
	char *p;
	if (AL == NULL) {
		iError.RaiseError("iArrayList.GetElement",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	if (AL->Flags&AL_READONLY) {
		AL->RaiseError("iArrayList.GetElement",CONTAINER_ERROR_READONLY);
		return NULL;
	}
	if (idx >=AL->count ) {
		AL->RaiseError("GetElement",CONTAINER_ERROR_INDEX);
		return NULL;
	}
	p = AL->contents;
	p += idx*AL->ElementSize;
	return p;
}

/*------------------------------------------------------------------------
 Procedure:     InsertAt ID:1
 Purpose:       Inserts data at the given position, that should be
                between zero and the number of items in the array
                list. If the index is equal to the count of items in
                the array the item will be inserted at the end. Any
                insertion in the middle or at the beginning provokes
                a move of all items between the index and the end of
                the array list
 Input:         The array list, the index and the new data to be
                inserted
 Output:        Error code less than zero if error, or positive
                number bigger than zero otherwise
 Errors:        The index must be correct and the array list must be
                writable
------------------------------------------------------------------------*/
static int InsertAt(ArrayList *AL,size_t idx,void *newval)
{
	char *p;
	
	if (AL == NULL) {
		iError.RaiseError("iArrayList.InsertAt",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (idx > AL->count) {
		AL->RaiseError("iArrayList.InsertAt",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	if (AL->count >= (AL->capacity-1)) {
		int r = Resize(AL);
		if (r <= 0)
			return r;
	}
	p = (char *)AL->contents;
	if (idx < AL->count) {
		memmove(p+AL->ElementSize*(idx+1),
				p+AL->ElementSize*idx,
				(AL->count-idx+1)*AL->ElementSize);
	}
	p += idx*AL->ElementSize;
	memcpy(p,newval,AL->ElementSize);
	AL->timestamp++;
	++AL->count;
	return 1;
}

static int Insert(ArrayList *AL,void *newval)
{
	return InsertAt(AL,0,newval);
}

static int RemoveAt(ArrayList *AL,size_t idx)
{
	char *p;
	if (idx >= AL->count)
		return CONTAINER_ERROR_INDEX;
	if (AL->Flags & AL_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (AL->count == 0)
		return -2;
	p = AL->contents;
	if (idx < (AL->count-1)) {
		memmove(p+AL->count*idx,p+AL->ElementSize*(idx+1),(AL->count-idx)*AL->ElementSize);
	}
	AL->count--;
	AL->timestamp++;
	return 1;
}

static int Remove(ArrayList *AL,void *str)
{
	size_t idx = IndexOf(AL,str,NULL);
	if (idx == 0)
		return 0;
	return RemoveAt(AL,idx-1);
}

static int PushBack(ArrayList *AL,void *str)
{
	return InsertAt(AL,AL->count,str);
}

/*------------------------------------------------------------------------
 Procedure:     Pop ID:1
 Purpose:       Removes the last item of the list and returns its
                data to the caller. It is the responsability of the
                caller to free this memory. This is almost the same as
				the "Remove" function, with the only difference that the
				data is NOT freed but returned to the caller.
 Input:         The array list to be popped
 Output:        a pointer to the last item's data
 Errors:        If the list is read-only or empty returns NULL
------------------------------------------------------------------------*/
static void * PopBack(ArrayList *AL)
{
    void *result;
	char *p;
    if ((AL->Flags&AL_READONLY) || AL->count == 0)
            return NULL;
    AL->count--;
	p = AL->contents;
	p += AL->ElementSize*AL->count;
    result = p;
	AL->timestamp++;
    return result;
}

/*------------------------------------------------------------------------
 Procedure:     Finalize ID:1
 Purpose:       Releases all memory associated with an arraylist,
                including the arraylist structure itself
 Input:         The arraylist to be destroyed
 Output:        The number of items that the arraylist contained or
                zero if error
 Errors:        Input must be writable
------------------------------------------------------------------------*/
static int Finalize(ArrayList *AL)
{
	int result = Clear(AL);
	FREE(AL,AL);
	return result;
}

static size_t GetCapacity(const ArrayList *AL)
{
	return AL->capacity;
}

static bool SetCapacity(ArrayList *AL,int newCapacity)
{
	if (AL->count != 0)
		return 0;
	FREE(AL,AL->contents);
	AL->contents = MALLOC(AL,newCapacity*sizeof(void *));
	if (AL->contents == NULL) {
		AL->RaiseError("SetCapacity",CONTAINER_ERROR_NOMEMORY);
		return 0;
	}
	memset(AL->contents,0,sizeof(void *)*newCapacity);
	AL->capacity = newCapacity;
	return 1;
}

static void Apply(struct _ArrayList *AL,int (*Applyfn)(void *,void *),void *arg)
{
	size_t i;
	char *p=AL->contents;
	for (i=0; i<AL->count;i++) {
		Applyfn(p,arg);
		p += AL->ElementSize;
	}
}

/*------------------------------------------------------------------------
 Procedure:     ReplaceAt ID:1
 Purpose:       Replace the data at the given position with new
                data.
 Input:         The list and the new data
 Output:        Returns the new data if OK, NULL if error.
 Errors:        Index error or read only errors are caught
------------------------------------------------------------------------*/
static void *ReplaceAt(ArrayList *AL,size_t idx,void *newval)
{
	char *p;

	if (AL->Flags & AL_READONLY)
		return 0;
	if (idx >= AL->count)
		return 0;
	p = AL->contents;
	p += idx*AL->ElementSize;
	memcpy(p,newval,AL->ElementSize);
	AL->timestamp++;
	return p;
}

static ErrorFunction SetErrorFunction(ArrayList *AL,ErrorFunction fn)
{
	ErrorFunction old;
	old = AL->RaiseError;
	AL->RaiseError = (fn) ? fn : iError.EmptyErrorFunction;
	return old;
}

static size_t Sizeof(ArrayList *AL)
{
	if (AL == NULL)
		return 0;
	return sizeof(ArrayList) + (AL->count * AL->ElementSize);
}

/*------------------------------------------------------------------------
 Procedure:     SetCompareFunction ID:1
 Purpose:       Defines the function to be used in comparison of
                Arraylist elements
 Input:         The Arraylist and the new comparison function. If the new
                comparison function is NULL no changes are done.
 Output:        Returns the old function
 Errors:        None
------------------------------------------------------------------------*/
static CompareFunction SetCompareFunction(ArrayList *l,CompareFunction fn)
{
	CompareFunction oldfn = l->CompareFn;

	if (fn != NULL) /* Treat NULL as an enquiry to get the compare function */
		l->CompareFn = fn;
	return oldfn;
}

static bool Sort(ArrayList *AL)
{
	CompareInfo ci;
	ci.Object = AL;
	ci.ExtraArgs = NULL;
	qsortEx(AL->contents,AL->count,AL->ElementSize,AL->CompareFn,&ci);
	return 1;
}

/* ------------------------------------------------------------------------------ */
/*                                Iterators                                       */
/* ------------------------------------------------------------------------------ */

struct ArrayListIterator {
	Iterator it;
	ArrayList *AL;
	size_t index;
	size_t timestamp;
	unsigned long Flags;
	void *Current;
};

static void *GetNext(Iterator *it)
{
	struct ArrayListIterator *ali = (struct ArrayListIterator *)it;
	ArrayList *AL = ali->AL;
	char *p;

	if (ali->index >= AL->count)
		return NULL;
	if (ali->timestamp != AL->timestamp) {
		AL->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	p = AL->contents;
	p += ali->index*AL->ElementSize;
	ali->index++;
	ali->Current = p;
	return p;
}

static void *GetPrevious(Iterator *it)
{
	struct ArrayListIterator *ali = (struct ArrayListIterator *)it;
	ArrayList *AL = ali->AL;
	char *p;

	if (ali->index >= AL->count || ali->index == 0)
		return NULL;
	if (ali->timestamp != AL->timestamp) {
		AL->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	p = AL->contents;
	p += ali->index*AL->ElementSize;
	ali->index--;
	ali->Current = p;
	return p;
}

static void *GetCurrent(Iterator *it)
{
	struct ArrayListIterator *ali = (struct ArrayListIterator *)it;
	return ali->Current;
}

static void *GetFirst(Iterator *it)
{
	struct ArrayListIterator *ali = (struct ArrayListIterator *)it;
	if (ali->AL->count == 0) {
		ali->Current = NULL;
		return NULL;
	}
	ali->index = 1;
	ali->Current = ali->AL->contents;
	return ali->Current;
}

static Iterator *newIterator(ArrayList *AL)
{
	struct ArrayListIterator *result = MALLOC(AL,sizeof(struct ArrayListIterator));
	if (result == NULL)
		return NULL;
	result->it.GetNext = GetNext;
	result->it.GetPrevious = GetPrevious;
	result->it.GetFirst = GetFirst;
	result->it.GetCurrent = GetCurrent;
	result->AL = AL;
	result->Current = NULL;
	result->timestamp = AL->timestamp;
	return &result->it;
}

static void deleteIterator(Iterator * it)
{
	struct ArrayListIterator *ali = (struct ArrayListIterator *)it;
	FREE(ali->AL,it);
}

static size_t DefaultSaveFunction(const void *element,void *arg, FILE *Outfile)
{
	const unsigned char *str = element;
	size_t len = *(size_t *)arg;

	return fwrite(str,1,len,Outfile);
}

static size_t DefaultLoadFunction(void *element,void *arg, FILE *Infile)
{
	size_t len = *(size_t *)arg;

	return fread(element,1,len,Infile);
}

static int Save(ArrayList *AL,FILE *stream, SaveFunction saveFn,void *arg)
{
	size_t i;
	if (saveFn == NULL) {
		saveFn = DefaultSaveFunction;
		arg = &AL->count;
	}
	if (fwrite(AL,1,sizeof(ArrayList),stream) <= 0)
		return EOF;
	for (i=0; i< AL->count; i++) {
		char *p = AL->contents;

		p += i*AL->ElementSize;
		if (saveFn(p,arg,stream) <= 0)
			return EOF;
	}
	return 0;
}

static ArrayList *Load(FILE *stream, ReadFunction loadFn,void *arg)
{
	size_t i;
	char *p;
	ArrayList *result,AL;
	if (loadFn == NULL) {
		loadFn = DefaultLoadFunction;
		arg = &AL.count;
	}
	if (fread(&AL,1,sizeof(ArrayList),stream) <= 0)
		return NULL;
	p = malloc(AL.ElementSize);
	if (p == NULL)
		return NULL;
	result = Create(AL.ElementSize,AL.count);
	result->Flags = AL.Flags;
	for (i=0; i< AL.count; i++) {
		if (loadFn(p,arg,stream) <= 0)
			break;
		Add(result,p);
	}
	free(p);
	return 0;
}


static ArrayList *Create(size_t elementsize,size_t startsize)
{
	ArrayList *result;
	size_t es;
	
	/* This detects BOGUS sizes arising from passing a negative value */
	if ((int)elementsize <= 0) {
		iError.RaiseError("newArrayList",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	/* Round the element size to correctly align table elements.
	 This supposes that sizeo(void *) is a power of two
	 */
	if (elementsize > sizeof(void *))
		elementsize = roundup(elementsize,sizeof(void *));
	/* Allocate space for the array list header structure */
	result = CurrentMemoryManager->malloc(sizeof(*result));
	if (result == NULL) {
		iError.RaiseError("newArrayList",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	memset(result,0,sizeof(*result));
	if (startsize <= 0)
		startsize = DEFAULT_START_SIZE;
	es = startsize * elementsize;
	result->contents = CurrentMemoryManager->malloc(es);
	if (result->contents == NULL) {
		iError.RaiseError("newArrayList",CONTAINER_ERROR_NOMEMORY);
		CurrentMemoryManager->free(result);
		result = NULL;
	}
	else {
		memset(result->contents,0,es);
		result->capacity = startsize;
		result->VTable = &iArrayList;
		result->ElementSize = elementsize;
		result->CompareFn = DefaultArrayListCompareFunction;
		result->RaiseError = iError.RaiseError;
	}
	return result;
}

static size_t GetElementSize(const ArrayList *AL)
{
	return AL->ElementSize;
}

ArrayListInterface iArrayList = {
	Create,
	GetElementSize,
	GetCount, GetFlags, SetFlags, Add, AddRange, 
	GetRange,
	Clear, Contains,
	Copy,
	CopyElement,
	CopyTo, IndexOf, Insert,
	InsertAt,
	Equal,
	GetElement, Remove, RemoveAt,
	Finalize, 
	GetCapacity, 
	SetCapacity,
	Apply, PushBack,
	PopBack,
	ReplaceAt,
	SetErrorFunction,
	SetCompareFunction,
	Sizeof,
	Sort,
	NULL,
	newIterator,deleteIterator,
	Save,
	Load,
};
