#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "containers.h"

/* Definition of the String Collection type */
struct _BitString {
    BitStringInterface *VTable; /* The table of functions */
    size_t count;                  /* number of elements in the array */
    BIT_TYPE *contents;               /* The contents of the collection */
    size_t capacity;                /* allocated space in the contents vector */
    unsigned long timestamp;
    unsigned int Flags;             /* Read-only or other flags */
} ;


#ifndef CHUNK_SIZE
#define CHUNK_SIZE 256
#endif

#if CHAR_BIT==8
static unsigned char mask[] = {
	255,127,63,31,15,7,3,1};
static unsigned char maskI[]= {
	255,1,3,7,15,31,63,127};
static unsigned char BitIndexMask[] = {
	1,2,4,8,16,32,64,128};
#elif CHAR_BIT==16
static unsigned char mask[] = {
	65535,32767,16383,8191,4095,2047,1023,511,255,127,63,31,15,7,3,1};
static unsigned char maskI[]= {
	65535,1,3,7,15,31,63,127,255,511,1023,2047,4095,8191,16383,32767};
static unsigned char BitIndexMask[] = {
	1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768};
static unsigned char InvertedBitIndexMask[] ={
	~1,~2,~4,~8,~16,~32,~64,~128,~256,~512,~1024,~2048,~4096,~8192,~16384,~32768};
#else
#error "Unimplemented CHAR_BIT bigger than 16"
#endif

#define BYTES_FROM_BITS(bitcount) ((bitcount)/CHAR_BIT)
#define BITPOS(idx) (idx & (CHAR_BIT-1))
#if 0
/* Returns the number of elements stored */
static size_t GetCount(BitString *AL);

/* Is this collection read only? */
static unsigned GetFlags(BitString *AL);

/* Sets this collection read-only or unsets the read-only flag */
static unsigned SetFlags(BitString *AL,unsigned flags);

/* Adds one element at the end. Given element is copied */
static int Add(BitString *AL,bool newval);

/* Clears all data and frees the memory */
static int Clear(BitString *AL);

/*Case sensitive search of a character string in the data */
static bool Contains(BitString *AL,BitString *sequence,void *ExtraArgs);

/*Returns the index of the given string or -1 if not found */
static size_t IndexOf(BitString *AL,bool bit,void *ExtraArgs);

/* Inserts a string at the position zero. */
static size_t Insert(BitString *b,bool bit);

/* Inserts a string at the given position */
static size_t InsertAt(BitString *AL,size_t idx,bool bit);

/* Returns the string at the given position */
static bool GetElement(BitString *AL,size_t idx);
/* Removes the given string if found */
static int Remove(BitString *b,bool bit);
/*Removes the bit at the indicated position */
static bool RemoveAt(BitString *b,size_t idx);
static int Finalize(BitString *b);
/* Returns the current capacity of the collection */
static int GetCapacity(BitString *b);
static bool SetCapacity(BitString *b,size_t newCapacity);
static bool Apply(BitString *b,int (*Applyfn)(bool,void *),void *arg);
static bool Pop(BitString *b);
static bool ReplaceAt(BitString *AL,size_t idx,bool newval);
static ErrorFunction SetErrorFunction(BitString *AL,ErrorFunction fn);
static 	BitString *Or(BitString *left,BitString *right);
static BitString * OrAssign(BitString *bsl,BitString *bsr);
static int Equal(BitString *bsl,BitString *bsr);
static BitString * And(BitString *bsl,BitString *bsr);
static BitString *AndAssign(BitString *bsl,BitString *bsr);
static BitString * Xor(BitString *bsl,BitString *bsr);
static BitString *XorAssign(BitString *bsl,BitString *bsr);
static BitString * Not(BitString *bsl);
static BitString *NotAssign(BitString *bsl);
void ContainerRaiseError(const char *fnname,int errcode);
static BitString *Reverse(BitString *b);
static BitString *GetRange(BitString *bs,size_t start,size_t end);
static BitString *StringToBitString(unsigned char * str);
static BitString *ObjectToBitString(unsigned char *p,size_t size);
static int LessEqual(BitString *bsl,BitString *bsr);
static bool SetElement(BitString *bs,size_t position,bool b);
static BitString * LeftShift(BitString *bs,size_t shift);
static BitString * RightShift(BitString *bs,size_t shift);
static int Print(BitString *b,size_t bufsiz,unsigned char *out);
static BitString *Append(BitString *b1,BitString *b2);
static size_t Sizeof(BitString *b);
static BitString *Copy(BitString *b);
static bool Set(BitString *b,size_t start,size_t stop,bool newval);
static uintmax_t PopulationCount(BitString *b);
static uintmax_t BitBlockCount(BitString *b);
static Iterator *newIterator(BitString *);
#endif

static size_t bytesizeFromBitLen(size_t bitlen){
	size_t bytesiz = BYTES_FROM_BITS(bitlen);
	bytesiz = (3+bytesiz) & ~3;
	return bytesiz;
}


/*------------------------------------------------------------------------
Procedure:     SetCapacity ID:1
Purpose:       Resizes a bit string without touching the data
Input:         The bitstring to be resized, and the new size in
bits
Output:        1 means no errors, 0 means no memory available for
the operation.
Errors:
------------------------------------------------------------------------*/
static bool SetCapacity(BitString *b,size_t bitlen){
	size_t bytesize,copy_size;
	unsigned char *oldData;
	if (b == NULL)		return 0;
	bytesize = bytesizeFromBitLen(bitlen);
	oldData = b->contents;
	b->contents = MALLOC(b,bytesize);
	if (b->contents == NULL) {
		b->contents = oldData;
		return 0;
	}
	memset(b->contents,0,bytesize);
	copy_size = 1+BYTES_FROM_BITS(b->count);
	copy_size = (copy_size > bytesize) ? bytesize : copy_size;
	memcpy(b->contents,oldData,copy_size);
	b->capacity = (unsigned)bytesize;
	FREE(b,oldData);
	return 1;
}

static size_t GetCapacity(BitString *b){
	return b->capacity;
}

static int Clear(BitString *b)
{
	if (b->Flags&CONTAINER_READONLY)
		return CONTAINER_ERROR_READONLY;
	FREE(b,b->contents);
	b->contents = NULL;
	b->count = 0;
	return 1;
}

static int Finalize(BitString *b)
{
	if (b->Flags&CONTAINER_READONLY)
		return CONTAINER_ERROR_READONLY;
	b->VTable->Clear(b);
	FREE(b,b);
	return 1;
}
static BitString *newBitString(size_t bitlen);
static BitString *Copy(BitString *b)
{
	BitString *result;

	if (b == NULL)		return NULL;
	result = newBitString(b->count);
	if (result == NULL)		return NULL;
	memcpy(result->contents,b->contents,1+BYTES_FROM_BITS(b->count));
	return result;
}

static size_t GetCount(BitString *b)
{
	return (b == NULL) ? 0 : b->count;
}


static bool SetElement(BitString *bs,size_t position,bool b)
{
	if (bs == NULL)
		return 0;

	if (bs->Flags&CONTAINER_READONLY)
		return CONTAINER_ERROR_READONLY;

	if (position >= bs->count)
		return b;
	if (b) {
		bs->contents[position >> 3] |= 1 << (position&7);
	}
	else
	    bs->contents[position >> 3] &= ~(1 << (position&7));
	return b;
}

static bool GetElement(BitString *bs,size_t position)
{
	if (position >= bs->count)
		return 0;
	return (bs->contents[BYTES_FROM_BITS(position)] &BitIndexMask[position&(CHAR_BIT-1)]) ? 1 : 0;
}

static BitString * RightShift(BitString *bs,size_t shift){
	size_t left = 0,tmp,tmp1;
	size_t len,bytesize;
	unsigned char *pdata;

	len = bs->count;
	pdata = bs->contents;
	bytesize = BYTES_FROM_BITS(len);
	if (shift > len) {
		memset(bs->contents,0,bytesize);
		return bs;
	}
	if (shift >= CHAR_BIT) {
		left = shift >> 3;
		memmove(pdata,pdata+left,bytesize-left);
		shift &= 7;
		memset(pdata+bytesize-left,0,left);
	}
	if (shift) {
		while (len >= 8) {
			tmp = *pdata;
			tmp >>= shift;
			tmp1 = pdata[1];
			tmp1 <<= (CHAR_BIT - shift);
			tmp |= tmp1;
			*pdata = (unsigned char)tmp;
			pdata++;
			len -= 8;
		}
		if (len) {
			tmp = *pdata;
			tmp >>= shift;
			*pdata = (unsigned char)tmp;
		}
	}
	return bs;
}

static BitString * LeftShift(BitString *bs,size_t shift){
	unsigned int tmp=0,tmp1,Mask;
	size_t left=0;
	unsigned char *pdata;
	size_t len,bytesize;

	len = bs->count;
	pdata = bs->contents;
	bytesize = BYTES_FROM_BITS(len);
	if (shift > len) {
		memset(bs->contents,0,bytesize);
		return bs;
	}
	if (shift >= CHAR_BIT) {
		left = shift/CHAR_BIT;
		memmove(pdata+left,pdata,bytesize-left);
		shift &= (CHAR_BIT-1);
		memset(pdata,0,left);
	}
	if (shift) {
		Mask = ~(((unsigned)-1) << shift);
		while (len >= CHAR_BIT) {
			tmp1 = *pdata;
			*pdata = (unsigned char)((*pdata << shift) | tmp);
			len -= CHAR_BIT;
			tmp = (tmp1 >> (CHAR_BIT-shift))&Mask;
			pdata++;
		}
		if (len) {
			*pdata = (unsigned char)((*pdata << shift) | tmp);
		}
	}
	return bs;
}

static BitString *GetRange(BitString *bs,size_t start,size_t end)
{
	size_t t,len;
	BitString *result;
	size_t startbyte,endbyte,shiftamount;

	if (bs == NULL)		return NULL;
	if (end >= bs->count)		end = bs->count;
	if (start > end) {
		t = start;
		start = end;
		end = t;
	}
	len  = end-start;
	result = newBitString(len);
	startbyte = BYTES_FROM_BITS(start);
	endbyte = BYTES_FROM_BITS(end);
	shiftamount = start&(CHAR_BIT-1);
	/* Copy the first byte */
	result->contents[0] = (bs->contents[0] >> shiftamount);
	memcpy(result->contents,bs->contents+startbyte,endbyte-startbyte);
	if (shiftamount)		RightShift(result,shiftamount);
	return result;
}

static int LessEqual(BitString *bsl,BitString *bsr){
	unsigned int *pdataL, *pdataR;
	size_t i,j,k,l;

	if ((bsl->count == 0 && bsr->count != 0) ||
		    (bsl->count == bsr->count && bsr->count == 0))
		return 1;
	if (bsr->count == 0 && bsl->count != 0)
		return 0;
	pdataL = (unsigned int *)bsl->contents;
	pdataR = (unsigned int *)bsr->contents;
	i = bytesizeFromBitLen(bsl->count) >> 3;
	j = bytesizeFromBitLen(bsr->count) >> 3;
	l = k = (i < j) ? i : j;
	while (--k >= 0) {
		if (pdataL[k] &  ~pdataR[k])
			return 0;
	}
	if (i == j)		return 1;
	/* The rest must be zero */
	k = (i < j) ? j : i;
	pdataL = (i < j) ? pdataR : pdataL;
	while (l < k) {
		if (pdataL[l])
			return 0;
		l++;
	}
	return 1;
}

static int Equal(BitString *bsl,BitString *bsr)
{
	unsigned int *pdataL,*pdataR;
	size_t top;

	if (bsl->count != bsr->count)		return 0;
	pdataL = (unsigned int *)bsl->contents;
	pdataR = (unsigned int *)bsr->contents;
	top = BYTES_FROM_BITS(bsl->count);
	if (memcmp(pdataL,pdataR,top))		return 0;
	if (bsl->count&(CHAR_BIT-1)) {
		if ( (pdataL[top]&maskI[bsl->count&(CHAR_BIT-1)]) != (pdataR[top]&maskI[bsl->count&(CHAR_BIT-1)]) )
			return 0;
	}
	return 1;
}

static BitString * Or(BitString *bsl,BitString *bsr)
{
	size_t len,i,resultlen;
	BitString *result,*src;

	len = bsl->count < bsr->count ? bsl->count : bsr->count;
	len = BYTES_FROM_BITS(len);
	resultlen = bsl->count < bsr->count ? bsr->count : bsl->count;
	result = newBitString(resultlen);
	for (i=0; i<len;i++) {
		result->contents[i] = bsl->contents[i] | bsr->contents[i];
	}
	if (bsl->count < bsr->count)		src = bsr;
	else
	    src = bsl;
	while (i < resultlen) {
		result->contents[i] |= src->contents[i];
		i++;
	}
	return result;
}

static BitString * OrAssign(BitString *bsl,BitString *bsr)
{
	size_t len,i;

	len = bsl->count < bsr->count ? bsl->count : bsr->count;
	len = BYTES_FROM_BITS(len);
	for (i=0; i<len;i++) {
		bsl->contents[i] |=  bsr->contents[i];
	}
	return bsl;
}

static BitString * And(BitString *bsl,BitString *bsr)
{
	size_t len,i,resultlen;
	BitString *result,*src;

	len = bsl->count < bsr->count ? bsl->count : bsr->count;
	resultlen = bsl->count < bsr->count ? bsr->count : bsl->count;
	result = newBitString(resultlen);
	for (i=0; i<len;i++) {
		result->contents[i] = bsl->contents[i] & bsr->contents[i];
	}
	if (bsl->count < bsr->count)
		src = bsr;
	else
	    src = bsl;
	while (i < resultlen) {
		result->contents[i] = src->contents[i];
		i++;
	}
	return result;
}
static BitString * AndAssign(BitString *bsl,BitString *bsr)
{
	size_t len,i;

	len = bsl->count < bsr->count ? bsl->count : bsr->count;
	for (i=0; i<len;i++) {
		bsl->contents[i] &= bsr->contents[i];
	}
	return bsl;
}

static BitString * Xor(BitString *bsl,BitString *bsr)
{
	size_t len,i,resultlen;
	BitString *result,*src;

	len = bsl->count < bsr->count ? bsl->count : bsr->count;
	resultlen = bsl->count < bsr->count ? bsr->count : bsl->count;
	result = newBitString(resultlen);
	for (i=0; i<len;i++) {
		result->contents[i] = bsl->contents[i] ^ bsr->contents[i];
	}
	if (bsl->count < bsr->count)		src = bsr;
	else
	    src = bsl;
	while (i < resultlen) {
		result->contents[i] = src->contents[i];
		i++;
	}
	return result;
}
static BitString * XorAssign(BitString *bsl,BitString *bsr)
{
	size_t len,i;

	len = bsl->count < bsr->count ? bsl->count : bsr->count;
	for (i=0; i<len;i++) {
		bsl->contents[i] ^= bsr->contents[i];
	}
	return bsl;
}

static BitString * Not(BitString *bsl)
{
	size_t len,i;
	BitString *result;

	len = bsl->count;
	result = newBitString(len);
	for (i=0; i<len;i++) {
		result->contents[i] = ~bsl->contents[i];
	}
	return result;
}
static BitString * NotAssign(BitString *bsl)
{
	size_t len,i;

	len = bsl->count;
	for (i=0; i<len;i++) {
		bsl->contents[i] = ~bsl->contents[i];
	}
	return bsl;
}


static BitString *ObjectToBitString(unsigned char *p,size_t siz)
{
	BitString *result = newBitString(CHAR_BIT*siz);
	memcpy(result->contents,p,siz);
	return result;
}


static BitString * StringToBitString(unsigned char * str)
{
	int i=0;
	BitString *result;
	unsigned char *save = str;
	if (str == NULL)		return NULL;
	if (*str == '0' && (str[1] == 'B' || str[1] == 'b'))
		str += 2;
	while (*str) {
		while (*str == ' ' || *str == '\t' || *str == ',')
			str++;
		if (*str == 0)
			break;
		if (*str == '1' || *str == '0') {
			i++;
			str++;
		}
		else break;
	}
	if (i == 0)
		return NULL;
	result = newBitString(i);
	str--;
	i=0;
	while (str >= save) {
		while (*str == ' ' || *str == '\t' || *str == ',')
			str--;
		if (*str == 0)
			break;
		if (*str == '1' || *str == '0') {
			if (*str == '1')
				result->contents[BYTES_FROM_BITS(i)] |= BitIndexMask[i&(CHAR_BIT-1)];
			str--;
			i++;
		}
		else break;
	}
	return result;
}

static size_t Print(BitString *b,size_t bufsiz,unsigned char *out)
{
	size_t result=0,j=b->count;
	unsigned char *top = out+bufsiz-1;
	size_t i;

	for (i=b->count-1; j>0; i--) {
		if (out >= top)
			break;
		if (i!= b->count-1 && 0 == (j&3))
			*out++ = ' ';
		if (i!= b->count-1 && 0 == (j&7))
			*out++ = ' ';
		if (b->contents[BYTES_FROM_BITS(i)] & BitIndexMask[i&7])
			*out = '1';
		else
		    *out = '0';
		out++;
		if (out >= top)
			break;
		result += 2;
		j--;
	}
	*out = 0;
	return result;
}


static BitString *Append(BitString *b1,BitString *b2)
{
	BitString *result;
	size_t bytesize,i,idx;

	if (b1 == NULL)
		return Copy(b2);
	if (b2 == NULL)
		return Copy(b1);
	result = newBitString(b1->count+b2->count);
	if (result == NULL)
		return NULL;
	bytesize = bytesizeFromBitLen(b1->count);
	memcpy(result->contents,b1->contents,bytesize);
	bytesize = bytesizeFromBitLen(result->count);
	for (i=b1->count;i<bytesize;i++) {
		idx = i-b1->count;
		if ((b2->contents[idx >>3] >>(idx&7))&1) {
			result->contents[i >>3] |= 1 << (i&7);
		}
	}
	return result;
}

static BitString *Reverse(BitString *b)
{
	BitString *result;
	size_t i,j;

	if (b == NULL)		return NULL;
	result = newBitString(b->count);
	i = 0;
	j = b->count - 1;
	while (i < result->count) {
		if ((b->contents[j >> 3] >> (j&7))&1) {
			result->contents[i >> 3] |= 1 << (i&7);
		}
		i++;
		j--;
	}
	return result;
}

static size_t Sizeof(BitString *b)
{
	size_t result;
	if (b == NULL)		return 0;
	result = sizeof(BitString);
	result += b->capacity;
	return result;
}


static unsigned GetFlags(BitString *b)
{
	return b->Flags;
}

static unsigned SetFlags(BitString *b,unsigned newval)
{
	int result = b->Flags;
	b->Flags = newval;
	return result;
}

static size_t expandBitstring(BitString *b)
{
	BIT_TYPE *oldContents = b->contents;
	b->contents = realloc(b->contents,b->capacity+CHUNK_SIZE);
	if (b->contents == NULL) {
		b->contents = oldContents;
		return 0;
	}
	b->capacity += CHUNK_SIZE;
	return b->capacity;
}

static int Add(BitString *b,bool newval)
{
	size_t bytepos,bitpos;

	/* Test if the new size + 1 will fit */
	bytepos = BYTES_FROM_BITS(b->count+1);
	if (bytepos >= b->capacity) {
		if (!expandBitstring(b))			return 0;
	}
	bytepos = BYTES_FROM_BITS(b->count);
	bitpos = b->count&(CHAR_BIT-1);
	if (newval)		b->contents[bytepos] |= 1 << bitpos;
	else
		b->contents[bytepos] &= ~(1 << bitpos);
	b->count++;
	return 1;
}


static int ReplaceAt(BitString *b,size_t idx,bool newval)
{
	size_t bytepos,bitpos;

	if (idx >= b->count) {
		iError.RaiseError("ReplaceAt",CONTAINER_ERROR_INDEX);
		return 0;

	}
	bytepos = BYTES_FROM_BITS(idx);
	bitpos = b->count&(CHAR_BIT-1);
	if (newval)		b->contents[bytepos] |= BitIndexMask[bitpos];
	else
		b->contents[bytepos] &= ~BitIndexMask[bitpos];
	return 1;
}

static bool Pop(BitString *b){
	size_t bytepos,bitpos;

	if (b->count == 0) {
		iError.RaiseError("Pop",CONTAINER_ERROR_INDEX);
		return 0;
	}
	bytepos = BYTES_FROM_BITS(b->count);
	bitpos = b->count&(CHAR_BIT-1);
	b->count--;
	if (b->contents[bytepos]&(1 << bitpos))		return 1;
	return 0;
}

static bool Apply(BitString *b,int (*Applyfn)(bool,void *),void *arg){
	size_t i,bytepos,bitpos;

	for (i=0; i<b->count;i++) {
		bytepos = BYTES_FROM_BITS(i);
		bitpos = i & (CHAR_BIT-1);
		Applyfn(b->contents[bytepos]&(1 << bitpos),arg);
	}
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     bitBitstr ID:1
 Purpose:       Finds the bit position of the given bit-pattern in
                the given text. The algorithm is as follows:
                (1) It builds a table of CHAR_BIT lines, each line
                holding a one bit right shifted pattern.
                (2) For each byte in the text it will compare the
                byte with each byte in the shifted pattern table. A
                match is found if all bytes are equal.
                (3) The first byte and the last byte must be masked
                because they contain some bit positions with unused
                bits.
 Input:         The bit-string to be searched and its length, and
                the bit-pattern and its length. Both lengths in bits
 Output:        Zero if the pattern isn't found, or 1+bit-index of
                the starting bit of the pattern if the pattern is
                found.
 Errors:        If there is no memory it will return "pattern not
                found" (zero).
------------------------------------------------------------------------*/
static size_t bitBitstr(BitString *Text,BitString *Pat)
{
	unsigned char *text = Text->contents;
	size_t textSize = Text->count;
	size_t patSize = Pat->count;
	size_t patByteLength = BYTES_FROM_BITS(patSize);
	size_t i,j,k,stop;
	unsigned char *ShiftTable,*ptable,*pText;

	if (patSize > textSize) /* eliminate wrong arguments */
		return 0;
	/* Allocate the shift table. If no memory abort with zero. */
	ShiftTable = MALLOC(Text,patByteLength*CHAR_BIT);
	if (ShiftTable == NULL) {
		return 0;
	}
	/* The first line of the shift table is unshifted. Just copy it */
	memcpy(ShiftTable,text,patByteLength);
	ptable = ShiftTable + patByteLength; /* Advance poinbter to next line */
	for (i=1; i<CHAR_BIT; i++) {
		/* Copy the last line into the new line */
		memcpy(ptable,ptable-patByteLength,patByteLength);
		/* Now shift it by one bit */
		k=patByteLength;
		for (j = 0; k--; ptable++) {
			j = (j << CHAR_BIT) | *ptable;
			*ptable = (unsigned char)(j >> 1);
		}
	}

	/* We should stop when the number of bytes left is less than the pattern length */
	stop = BYTES_FROM_BITS(textSize) - patByteLength;

	/* For each byte in the input text */
	for (k=0; k <= stop; k++) {
		/* For each line in the shift table */
		for (i=0; i<CHAR_BIT;i++) {
			ptable = ShiftTable+i*patByteLength; /* will point to the table line */
			pText = text+k; /* Will point at the current start character in the text */
			/* The first character needs a mask since there could be
			unused bits */
			if ((*pText++ & mask[i]) != *ptable++)				
				break;
			/* Search each byte in the input text until last byte minus 1 */
			for (j=1; j<patByteLength-1; j++) {
				if (*pText++ != *ptable++)					
					break;
			}
			if (j == patByteLength-1) {
				/* Test the last byte with the complemented mask */
				if ((*pText & maskI[patSize&(CHAR_BIT-1)]) == *ptable) {
					free(ShiftTable);
					return 1+(k*CHAR_BIT)+i;
				}
			}
		}
	}
	/* There was no match */
	FREE(Text,ShiftTable);
	return 0;
}

static int Contains(BitString *txt,BitString *pattern,void *Args)
{
	return bitBitstr(txt,pattern) ? 1 : 0;
}

static size_t IndexOf(BitString *b,bool bit,void *ExtraArgs)
{
	size_t bitpos;
	size_t bytepos,i;

	for (i=0; i<b->count;i++) {
		bytepos = BYTES_FROM_BITS(i);
		bitpos = i&(CHAR_BIT-1);
		if (bit) {
			if (b->contents[bytepos]&(1 << bitpos))
				return 1+i;
		}
		else if ((b->contents[bytepos]&(1 << bitpos))==0)
			return i+1;
	}
	return 0;
}
#if 0
/*------------------------------------------------------------------------
Procedure:     ShiftRightByOne ID:1
Purpose:       Shift the given bit pattern 1 bit to the right. This
was proposed by Peter Nilsson in the comp.lang.c
group. It uses a barrel shifter.
Input:         A pattern, and its count in bits
Output:        The shifted pattern. This is done destructively.
Errors:        None
------------------------------------------------------------------------*/
static int ShiftRightByOne(unsigned char *p, size_t z,unsigned carry)
{

	while ( z-- ) {
		carry = (carry << CHAR_BIT) | *p;
		*p++ = carry >> 1;
	}

	return carry & 1;
}
#endif

static int ShiftLeftByOne(unsigned char *p, size_t z,unsigned carry)
{

	while ( z-- ) {
		*p = (unsigned char)((*p << 1)|carry);
		p++;
		carry = ((*p)>>(CHAR_BIT-1))?1:0;
	}

	return carry & 1;
}


static size_t InsertAt(BitString *b,size_t idx,bool value)
{
	size_t bitpos,newval;
	unsigned carry;
	bool oldval;
	size_t bytepos,count,bytesToShift,bytesUsed;

	count = b->count+1;
	bytesUsed = BYTES_FROM_BITS(count);
	if (bytesUsed >= b->capacity) {
		if (!expandBitstring(b))
			return 0;
	}
	bytepos = BYTES_FROM_BITS(idx);
	bitpos = idx & (CHAR_BIT-1);
	oldval = b->contents[bytepos];
	newval=0;
	if (value)
		newval |= BitIndexMask[bitpos];
	if (bitpos > 0)
		newval |= (oldval&mask[bitpos-1]);
	if (bitpos < (CHAR_BIT-1))
		newval |= (oldval << 1) & (((unsigned)-1) << (bitpos+1));
	carry = (oldval << (CHAR_BIT-1))?1: 0;
	b->contents[bytepos] = (unsigned char )newval;
	idx += CHAR_BIT-bitpos;
	bytesToShift = bytesUsed - (BYTES_FROM_BITS(idx));
	bytepos++;
	ShiftLeftByOne(&b->contents[bytepos],bytesToShift,carry);
	b->count++;
	return b->count;

}

static size_t Insert(BitString *b,bool bit){
	return InsertAt(b,0,bit);
}

/*------------------------------------------------------------------------
Procedure:     RemoveAt ID:1
Purpose:       Removes the indicated bit from the bitstring, making
               it shorter by 1 bit.
Input:         The bitstring and the index
Output:        Either 1 (success) or zero (failure)
Errors:        No error functions are called. It returns just zero.
               The rationale behind is that you can remove bits in
               a loop until there are none.
------------------------------------------------------------------------*/
static bool RemoveAt(BitString *bitStr,size_t idx)
{
	size_t bitpos,oldval,newval;
	size_t bytepos,bytesize;
	unsigned char Mask[] = { 0,0x80,0xc0,0xe0,0xf0,0xf8,0xfc,0xfe	};

	if (bitStr->count == 0) /* If bit string empty there is nothing to remove */
		return 0;
	if (bitStr->count < idx) /* if the index is beyond the data return failure */
		return 0;
	bytepos = BYTES_FROM_BITS(idx);
	bitpos = BITPOS(idx);
	oldval = bitStr->contents[bytepos];
	newval = oldval >> 1;
	if (bitpos > 0) {
		newval &= Mask[1+bitpos];
		newval |= oldval&(~Mask[1+bitpos]);
	}
	bytesize = BYTES_FROM_BITS(bitStr->count);
	if (bytesize > 1)
		newval |= (bitStr->contents[bytepos+1]&1) << (CHAR_BIT-1);
	bitStr->contents[bytepos] = (unsigned char)newval;
	bytepos++;
	while (bytepos < (bytesize-1)) {
		unsigned tmp = bitStr->contents[bytepos+1] &1;
		bitStr->contents[bytepos] =  (unsigned char)((bitStr->contents[bytepos] >> 1)|tmp);
		bytepos++;
	}
	bitStr->contents[bytepos] >>= 1;
	bitStr->count--;
	return 1;

}

static int Remove(BitString *bitStr,bool val){
	size_t idx = IndexOf(bitStr,val,NULL);
	if (idx == 0)		return 0;
	idx--;
	RemoveAt(bitStr,idx);
	return 1;
}

static bool Set(BitString *b,size_t start,size_t stop,bool newval)
{
	BIT_TYPE *contents;
	size_t startbyte,stopbyte;

	contents = b->contents;
	startbyte = BYTES_FROM_BITS(start);
	stopbyte = BYTES_FROM_BITS(stop);
	if (!newval) {
		if (startbyte == stopbyte) {
			contents[startbyte] &= ((0xff >> (CHAR_BIT - (start&(CHAR_BIT-1)))) |
				    (0xff << ((stop&0x7) + 1)));
		}
		else {
			contents[startbyte] &= 0xff >> (8 - (start&0x7));
			while (++startbyte < stopbyte)
                 contents[startbyte] = 0;
			contents[stopbyte] &= 0xff << ((stop&0x7) + 1);
		}
	}
	else {
        if (startbyte == stopbyte) {
                contents[startbyte] |= ((0xff << (start&0x7)) &
                                    (0xff >> (7 - (stop&0x7))));
        } else {
                contents[startbyte] |= 0xff << ((start)&0x7);
                while (++startbyte < stopbyte)
                        contents[startbyte] = 0xff;
                contents[stopbyte] |= 0xff >> (7 - (stop&0x7));
        }
	}
	return 1;
}

static size_t bitcount(uintmax_t x)
{
	if (64 == sizeof(x)*CHAR_BIT) {
	    x -=  (x>>1) & 0x5555555555555555ULL;                                /* 0-2 in 2 bits */
	    x  = ((x>>2) & 0x3333333333333333ULL) + (x & 0x3333333333333333ULL);  /* 0-4 in 4 bits */
	    x  = ((x>>4) + x) & 0x0f0f0f0f0f0f0f0fULL;                           /* 0-8 in 8 bits */
	    x *= 0x0101010101010101ULL;
	    return   x>>56;
	}
	else { /* Assume that 32 bit unsigned long exists. */
	    x -=  (x>>1) & 0x55555555UL;                        /* 0-2 in 2 bits */
	    x  = ((x>>2) & 0x33333333UL) + (x & 0x33333333UL);  /* 0-4 in 4 bits */
	    x  = ((x>>4) + x) & 0x0f0f0f0fUL;                   /* 0-8 in 8 bits */
	    x *= 0x01010101UL;
	    return  (unsigned)(x>>24);
	}
}

static uintmax_t PopulationCount(BitString *b)
{
	uintmax_t *pumax,x,result=0;
	size_t iterations;
	size_t count;
	unsigned char *p;

	if (b == NULL || b->count == 0)
		return 0;
	/* Treat 32 or 64 bits at once */
	pumax = (uintmax_t *)b->contents;
	iterations = b->count/(sizeof(x)*CHAR_BIT);
	/* Position the pointer p to get the last few bytes
	   that are left once the uintmax_t pointer finishes.
	*/
	p = b->contents+(iterations*sizeof(x));
	while (iterations > 0) {
		x = *pumax++;
		if (x != 0)
			result += bitcount(x);
		iterations--;
	}
	count = (b->count%(sizeof(x)*CHAR_BIT));
	if (count) {
		x=0;
		/* count is NOT unsigned! This supposes that an
		   integer can hold CHAR_BIT-1 bits.
		*/
		while (count > 0) {
			x = (x << CHAR_BIT) | *p++;
			count -= CHAR_BIT;
		}
		result += bitcount(x);
	}
	return result;
}

static uintmax_t BitBlockCount(BitString *b)
{
	uintmax_t x=0,result=0;
	unsigned char *pbits;
	size_t iterations;
	size_t count;

	if (b == NULL || b->count == 0)
		return 0;
	/* Treat 32 or 64 bits at once */
	pbits = b->contents;
	iterations = BYTES_FROM_BITS(b->count);
	/* Position the pointer p to get the last few bytes
	   that are left once the uintmax_t pointer finishes.
	*/
	while (iterations > 0) {
		x = *pbits++;
		if (x != 0)
			result += (x & 1) + bitcount( (x^(x>>1)) ) / 2;
		if (iterations > 1 && (x&1) ) {
			if (*pbits >> (CHAR_BIT-1)) {
				result--;
			}
		}
		iterations--;
	}
	count = BITPOS(b->count);
	if (count) {
		x = (*pbits)&mask[count];
		result += (x & 1) + bitcount( (x^(x>>1)) ) / 2;
	}
	return result;
}

/* ------------------------------------------------------------------------------ */
/*                                Iterators                                       */
/* ------------------------------------------------------------------------------ */

struct BitstringIterator {
	Iterator it;
	BitString *Bits;
	size_t index;
	size_t timestamp;
	int bit;
};

static void *GetNext(Iterator *it)
{
	struct BitstringIterator *bi = (struct BitstringIterator *)it;
	BitString *b = bi->Bits;

	if (bi->timestamp != b->timestamp) {
		return NULL;
	}
	if (bi->index >= b->count)
		return NULL;
	bi->bit = GetElement(b,bi->index++);
	return &bi->bit;
}
static void *GetCurrent(Iterator *it)
{
	struct BitstringIterator *bi = (struct BitstringIterator *)it;
	return &bi->bit;
}

static void *GetFirst(Iterator *it)
{
	struct BitstringIterator *bi = (struct BitstringIterator *)it;
	if (bi->Bits->count == 0)
		return NULL;
	bi->index = 0;
	return GetNext(it);
}

static Iterator *newIterator(BitString *bitstr)
{
	struct BitstringIterator *result = MALLOC(bitstr,sizeof(struct BitstringIterator));
	if (result == NULL)
		return NULL;
	result->it.GetNext = GetNext;
	result->it.GetPrevious = GetNext;
	result->it.GetFirst = GetFirst;
	result->it.GetCurrent = GetCurrent;
	result->Bits = bitstr;
	result->timestamp = bitstr->timestamp;
	return &result->it;
}
static int Save(BitString *bitstr,FILE *stream, SaveFunction saveFn,void *arg)
{
	if (fwrite(bitstr,1,sizeof(BitString),stream) <= 0)
		return EOF;
	if (fwrite(bitstr->contents, 1,BYTES_FROM_BITS(bitstr->count), stream) <= 0)
		return EOF;
	return 0;
}

static BitString *Load(FILE *stream, ReadFunction readFn,void *arg)
{
	BitString tmp,*r;
	
	if (fread(&tmp,1,sizeof(BitString),stream) <= 0)
		return NULL;
	r = newBitString(tmp.count);
	if (r == NULL)
		return NULL;
	r->count = tmp.count;
	r->Flags = tmp.Flags;
	if (fread(r->contents,1,BYTES_FROM_BITS(r->count),stream) <= 0) {
		iError.RaiseError("Bitstring.Load",CONTAINER_ERROR_FILE_READ);
		return NULL;
	}
	return r;
}


BitStringInterface iBitString = {
	newBitString,
	GetCount,
	GetFlags,
	SetFlags,
	Add,
	Clear,
	Contains,
	IndexOf,
	Insert,
	InsertAt,
	GetElement,
	SetElement,
	Remove,
	RemoveAt,
	Finalize,
	GetCapacity,
	SetCapacity,
	Apply,
	Add, /* Add and push are identical */
	Pop,
	ReplaceAt,
	NULL,
	newIterator,
	Sizeof,
	Or,
	OrAssign,
	And,
	AndAssign,
	Xor,
	XorAssign,
	Not,
	NotAssign,
	PopulationCount,
	BitBlockCount,
	Equal,
	LessEqual,
	Reverse,
	GetRange,
	StringToBitString,
	ObjectToBitString,
	LeftShift,
	RightShift,
	Print,
	Append,
	Copy,
	Set,
	Save,
	Load,
};



static BitString *newBitString(size_t bitlen)
{
	BitString *set;
	size_t bytesiz;

	set = 	CurrentMemoryManager->malloc(sizeof(BitString));
	if (set == NULL)		return NULL;
	set->count = bitlen;
	bytesiz = bytesizeFromBitLen(bitlen);
	set->contents = CurrentMemoryManager->malloc(bytesiz);
	if (set->contents == NULL	) {
		CurrentMemoryManager->free(set);
		return NULL;
	}
	memset(set->contents,0,bytesiz);
	set->VTable = &iBitString;
	set->VTable->Allocator = CurrentMemoryManager;
	return set;
}
