#ifndef __containers_h__
#define __containers_h__
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <limits.h>
/* ------------------------------------------------------------
   Default settings: NO_C99
---------------------------------------------------------------*/

#define NO_GC
#define NO_C99

#ifdef NO_C99
/* No stdbool */
typedef int bool;
/* No flexible arrays */
#define MINIMUM_ARRAY_INDEX	1
#else
/* Use C99 features */
#define MINIMUM_ARRAY_INDEX
#include <stdbool.h>
#endif

/* General container flags */
#define CONTAINER_READONLY	1
/************************************************************************** */
/*                                                                          */
/*                          ErrorHandling                                   */
/*                                                                          */
/************************************************************************** */
#define CONTAINER_ERROR_INDEX           -3
#define CONTAINER_ERROR_READONLY        -4
#define CONTAINER_ERROR_FILEOPEN        -5
#define CONTAINER_ERROR_NOTFOUND        -6
#define CONTAINER_ERROR_NO_COMPAREFUNCTION_SET -7
#define CONTAINER_INTERNAL_ERROR        -8
#define CONTAINER_ERROR_OBJECT_CHANGED  -9
#define CONTAINER_ERROR_NOT_EMPTY       -10
#define CONTAINER_ERROR_FILE_READ       -11
#define CONTAINER_ERROR_FILE_WRITE      -12
#define CONTAINER_FULL                  -13
#define CONTAINER_ASSERTION_FAILED      -14
#define CONTAINER_ERROR_BADARG	        -15
#define CONTAINER_ERROR_NOMEMORY        -16
#define CONTAINER_ERROR_NOENT           -17


typedef void (*ErrorFn)(const char *,int);

typedef struct tagError {
	ErrorFn RaiseError;
	void (*EmptyErrorFunction)(const char *fname,int errcode);
	char *(*StrError)(int errcode);
	ErrorFn (*SetErrorFunction)(ErrorFn);
} ErrorInterface;

extern ErrorInterface iError;
/************************************************************************** */
/*                                                                          */
/*                          Memory allocation object                        */
/*                                                                          */
/************************************************************************** */

typedef struct tagMemoryManager {
	void *(*malloc)(size_t);        /* Function to allocate memory */
	void (*free)(void *);           /* Function to release it      */
	void *(*realloc)(void *,size_t);/* Function to resize a block of memory */
	void *(*calloc)(size_t,size_t);
} ContainerMemoryManager;
extern ContainerMemoryManager * CurrentMemoryManager;
ContainerMemoryManager *SetCurrentMemoryManager(ContainerMemoryManager *in);
extern ContainerMemoryManager iDebugMalloc;
#define MALLOC(container,siz) 	((container->VTable->Allocator) ? \
               container->VTable->Allocator->malloc(siz) : \
               CurrentMemoryManager->malloc(siz))

#define REALLOC(container,obj,siz) 	((container->VTable->Allocator) ? \
                container->VTable->Allocator->realloc(obj,siz) : \
                CurrentMemoryManager->realloc(obj,siz))

#define FREE(container,obj) 	((container->VTable->Allocator) ? \
container->VTable->Allocator->free(obj) : \
CurrentMemoryManager->free(obj))

/************************************************************************** */
/*                         Iterator objects                                 */
/************************************************************************** */

typedef struct _Iterator {
    void *(*GetNext)(struct _Iterator *);
    void *(*GetPrevious)(struct _Iterator *);
    void *(*GetFirst)(struct _Iterator *);
    void *(*GetCurrent)(struct _Iterator *);
} Iterator;

/* Type definition of the compare function */
typedef struct tagCompareInfo {
	void *ExtraArgs;
	void *Object;
} CompareInfo;
typedef int (*CompareFunction)(const void *elem1, const void *elem2,CompareInfo *ExtraArgs);
typedef size_t (*SaveFunction)(const void *element, void *arg, FILE *OutputStream);
typedef size_t (*ReadFunction)(void *element, void *arg, FILE *InputStream);

typedef struct tagErrorInfo {
	int ErrorCode;
	const char *OperationName;
} ErrorInfo;

typedef void (*ErrorFunction)(const char *,int);
/************************************************************************** */
/*                                                                          */
/*                      Heap allocator objects                              */
/*                                                                          */
/************************************************************************** */
typedef struct tagHeapObject ContainerHeap;
typedef struct _HeapAllocatorInterface {
	ContainerHeap *(*Create)(size_t ElementSize,ContainerMemoryManager *m);
	void *(*newObject)(ContainerHeap *heap);
	void (*AddToFreeList)(ContainerHeap *heap,void *element);
	void (*DestroyFreeList)(ContainerHeap *heap);
	void (*Destroy)(ContainerHeap *heap);
	size_t (*Sizeof)(ContainerHeap *heap);
} HeapInterface;

extern HeapInterface iHeap;

/************************************************************************** */
/*                                                                          */
/*                            Pool Allocator object interface               */
/* This interface comes in two flavors: normal and debug versions           */
/************************************************************************** */
typedef struct Pool Pool;
typedef struct _tagPoolAllocatorInterface {
	Pool  *(*Create)(ContainerMemoryManager *m);
	void  *(*Alloc)(Pool *pool,size_t size);
	void  *(*Calloc)(Pool *pool,size_t n,size_t size);
	void   (*Clear)(Pool *);
	void   (*Destroy)(Pool *);
} PoolAllocatorInterface;

extern PoolAllocatorInterface iPool;

typedef struct _tagPoolAllocatorDebugInterface {
	Pool  *(*Create)(const char *file_line);
	void  *(*Alloc)(Pool *pool,size_t size,const char *file_line);
	void  *(*Calloc)(Pool *pool,size_t n,size_t size,const char *file_line);
	void   (*Clear)(Pool *,const char *file_line);
	void   (*Destroy)(Pool *,const char *file_line);
	int    (*FindPoolFromData)(Pool *pool, void *data);
	void   (*SetMaxFree)(Pool *pool, size_t size);
	size_t (*Sizeof)(Pool *);
} PoolAllocatorDebugInterface;

extern PoolAllocatorDebugInterface iPoolDebug;

/************************************************************************** */
/*                                                                          */
/*                            String Collections                            */
/*                                                                          */
/************************************************************************** */

/*
A string collection is a table of zero terminated strings that will grow
automatically when you add elements to it. Actually, it is the same container
as the vector but for variable length items: zero terminated strings.
*/

/* We define first an empty structure, that will be fully defined later, to be
able to define the functions in our interface
*/
typedef struct StringCollection StringCollection;
/* Forward declaration of the string collection type */
typedef struct _ArrayList ArrayList;

typedef int (*StringCompareFn)(unsigned char *s1,unsigned char *s2,CompareInfo *info);
/* This function is needed to read a line from a file.
   The resulting line is allocated with the given memory manager
*/
extern int getline(char **LinePointer,int *lineLength,FILE *stream,ContainerMemoryManager *mm);
/* Definition of the functions associated with this type. */
typedef struct {
	StringCollection *(*Create)(size_t startsize);
	StringCollection *(*CreateFromFile)(unsigned char *fileName);
    /* Returns the number of elements stored */
    size_t (*Size)(StringCollection *SC);
    /* Is this collection read only? */
    unsigned (*GetFlags)(StringCollection *SC);
    /* Sets this collection read-only or unsets the read-only flag */
    unsigned (*SetFlags)(StringCollection *SC,unsigned flags);
    /* Adds one element at the end. Given string is copied */
    int (*Add)(StringCollection *SC,char *newval);
    /* Adds a NULL terminated table of strings */
    int (*AddRange)(StringCollection *SC,char **newvalues);
    /* Clears all data and frees the memory */
    int (*Clear)(StringCollection *SC);
    /*Case sensitive search of a character string in the data */
    bool (*Contains)(StringCollection *SC,char *str);
    /* Copies all strings into a NULL terminated vector */
    char **(*CopyTo)(StringCollection *SC);
    /*Returns the index of the given string or -1 if not found */
    size_t (*IndexOf)(StringCollection *SC,char *SearchedString);
    /* Inserts a string at the position zero. */
    int (*Insert)(StringCollection *SC,char *);
    /* Inserts a string at the given position */
    int (*InsertAt)(StringCollection *SC,size_t idx,char *newval);
    /* Returns the string at the given position */
    char *(*GetElement)(StringCollection *SC,size_t idx);
    /* erases the given string if found */
    int (*Erase)(StringCollection *SC,char *);
    /*erases the string at the indicated position */
    int (*EraseAt)(StringCollection *SC,size_t idx);
    /* Frees the memory used by the collection */
    int (*Finalize)(StringCollection *SC);
    /* Returns the current capacity of the collection */
    size_t (*GetCapacity)(StringCollection *SC);
    /* Sets the capacity if there are no items in the collection */
    bool (*SetCapacity)(StringCollection *SC,int newCapacity);
    /* Calls the given function for all strings. "Arg" is a used supplied argument */
    /* that can be NULL that is passed to the function to call */
    void (*Apply)(StringCollection *SC,int (*Applyfn)(char *,void * arg),void *arg);
    /* Calls the given function for each string and saves all results
       in an integer vector */
    int *(*Map)(StringCollection *SC,int (*Applyfn)(char *));
    /* Pushes a string, using the collection as a stack */
    int (*Push)(StringCollection *SC,char *str);
    /* Pops the first (position zero) string of the collection */
    size_t (*Pop)(StringCollection *SC,char *outbuf,size_t buflen);
    /* Replaces the character string at the given position with a new one */
    char *(*ReplaceAt)(StringCollection *SC,size_t idx,char *newval);
    /* Compares two string collections */
    bool (*Equal)(StringCollection *SC1,StringCollection *SC2);
    /* Copy a string collection */
    StringCollection *(*Copy)(StringCollection *SC);
    /* Set or unset the error function */
    ErrorFunction (*SetErrorFunction)(StringCollection *SC,ErrorFunction fn);
    /* Sort the strings */
    bool (*Sort)(StringCollection *SC);
    /* Memory used by the string collection object and its data */
    size_t (*Sizeof)(StringCollection *SC);
    ContainerMemoryManager *Allocator;
    Iterator *(*newIterator)(StringCollection *SC);
	void (*deleteIterator)(Iterator *);
	/* Writes the string collection in binary form to a stream */
	int (*Save)(StringCollection *SC,FILE *stream, SaveFunction saveFn,void *arg);
	StringCollection *(*Load)(FILE *stream, ReadFunction readFn,void *arg);
	struct _ArrayList *(*CastToArray)(StringCollection *SC);
	size_t (*FindFirstText)(StringCollection *SC,char *text);
	size_t (*FindNextText)(StringCollection *SC, char *text,size_t start);
	ArrayList *(*FindText)(StringCollection *SC,char *text);
	int (*WriteToFile)(StringCollection *SC,unsigned char *filename);
} StringCollectionInterface;

extern StringCollectionInterface iStringCollection;

/* -------------------------------------------------------------------------
* LIST Interface                                                           *
 * This describes the single linked list data structure functions          *
 * Each list objects has a list composed of list elements.                 *
 *-------------------------------------------------------------------------*/
typedef struct _List List;
typedef struct {
	List *(*Create)(size_t element_size);
	size_t (*GetElementSize)(List *l);
    size_t (*Size)(List *L);        /* Returns the number of elements stored */
    unsigned (*GetFlags)(List *L);                      /* Gets the flags */
	unsigned (*SetFlags)(List *L,unsigned flags);       /* Sets the flags */
    int (*Add)(List *L,void *newval);              /* Adds element at end */
    int (*Clear)(List *L);                         /* Clears all elements */
    bool (*Contains)(List *L,void *element);       /* Searches an element */
    List *(*Copy)(List *L);           /* Copies all items into a new list */
    /*Returns the index of the given data or -1 if not found */
    size_t (*IndexOf)(List *L,void *SearchedElement,void *ExtraArgs);
    /* Inserts a value at the given position */
    int (*InsertAt)(List *L,size_t idx,void *newval);
    /* Returns the data at the given position */
    void *(*GetElement)(List *L,size_t idx);
	int  (*CopyElement)(List *list,size_t idx,void *OutBuffer);
    /* erases the given data if found */
    int (*Erase)(List *L,void *);
    /*erases the string at the indicated position */
	int (*EraseRange)(List *L,size_t start,size_t end);
    int (*EraseAt)(List *L,size_t idx);
    /* Frees the memory used by the collection and destroys the list */
    int (*Finalize)(List *L);
    /* Pushes an element, using the collection as a stack */
    int (*PushFront)(List *L,void *str);
    /* Pops the first element the list */
    int (*PopFront)(List *L,void *result);
    /* Replaces the element at the given position with a new one */
    int (*ReplaceAt)(List *L,size_t idx,void *newval);
    /* Sorts the list in place */
    int (*Sort)(List *l);
    /* Reverses the list */
    int (*Reverse)(List *l);
    /* Gets an element range from the list */
    List *(*GetRange)(List *l,size_t start,size_t end);
    /* Compares two lists (Shallow comparison) */
    bool (*Equal)(List *l1,List *l2);
    /* Add a list at the end of another */
    int (*Append)(List *l1,List *l2);
    /* Set The comparison function */
    CompareFunction (*SetCompareFunction)(List *l,CompareFunction fn);
    /* Call a callback function with each element of the list and an extra argument */
    int (*Apply)(List *L,int(Applyfn)(void *,void *),void *arg);
    ErrorFunction (*SetErrorFunction)(List *L,ErrorFunction); /* Set/unset the error function */
    CompareFunction Compare;
    size_t (*Sizeof)(List *l);
    ContainerMemoryManager *Allocator;
    Iterator *(*newIterator)(List *L);
	void *(*Seek)(Iterator *it,size_t pos);
	int (*deleteIterator)(Iterator *);
	int (*UseHeap)(List *L, ContainerMemoryManager *m);
	List *(*Load)(FILE *stream, ReadFunction loadFn,void *arg);
	int (*Save)(List *L,FILE *stream, SaveFunction saveFn,void *arg);
} ListInterface;

extern ListInterface iList;
/* -------------------------------------------------------------------
 *                           QUEUES
 * -------------------------------------------------------------------*/
typedef struct _Queue Queue;

typedef struct _QueueInterface {
	Queue *(*Create)(size_t elementSize);
	size_t (*Size)(Queue *Q);
	size_t (*Sizeof)(Queue *Q);
	unsigned (*GetFlags)(Queue *Q);
	unsigned (*SetFlags)(Queue *Q,unsigned newFlags);
	int (*Enqueue)(Queue *Q, void *Element);
	int (*Dequeue)(Queue *Q,void *result);
	int (*Clear)(Queue *Q);
	int (*Finalize)(Queue *Q);
} QueueInterface;

extern QueueInterface iQueue;
/* --------------------------------------------------------------------
 *                           Double Ended QUEues                      *
 * -------------------------------------------------------------------*/
typedef struct deque_t Deque;
typedef struct _DeQueueInterface {
	Deque *(*Create)(size_t elementSize);
	size_t (*Size)(Deque *Q);
	unsigned (*GetFlags)(Deque *Q);
	unsigned (*SetFlags)(Deque *Q,unsigned newFlags);
	int (*Add)(Deque *Q, void *Element);
	int (*AddLeft)(Deque *Q, void *Element);
	int (*Erase)(Deque * d, void* item);
	int (*Clear)(Deque *Q);
	int (*Finalize)(Deque *Q);
	size_t (*Contains)(Deque * d, void* item);
	int (*Reverse)(Deque * d);
	Deque *(*Copy)(Deque *d);
	int (*PopBack)(Deque *d,void *outbuf);
	int (*PeekBack)(Deque *d,void *outbuf);
	int (*PeekFront)(Deque *d,void *outbuf);
	int (*PopFront)(Deque *d,void *outbuf);
	int (*Rotate)(Deque*d,int n);
	int (*Rotateleft)(Deque *d,size_t);
	int (*RotateRight)(Deque *d,size_t);
	int (*Save)(Deque *d,FILE *stream, SaveFunction saveFn,void *arg);
	Deque *(*Load)(FILE *stream, ReadFunction readFn,void *arg);
} DequeInterface;

extern DequeInterface iDeque;

/* ------------------------------------------------------------------------
 * Double linked list interface                                            *
 * This describes the double linked list data structure functions          *
 * Each list objects has a list composed of list elemants.                 *
 *-------------------------------------------------------------------------*/
typedef struct Dlist Dlist;
typedef struct {
	Dlist *(*Create)(size_t elementsize);
	size_t (*GetElementSize)(Dlist *);
    size_t (*Size)(Dlist *dl);             /* Returns the number of elements stored */
	size_t (*Sizeof)(Dlist *dl);
    unsigned (*GetFlags)(Dlist *AL);                               /* Gets the flags */
    unsigned (*SetFlags)(Dlist *AL,unsigned flags);                    /* Sets flags */
    int (*Add)(Dlist *dl,void *newval);              /* Adds one element at the end. */
    int (*Clear)(Dlist *dl);                 /* Clears all data and frees the memory */
    bool (*Contains)(Dlist *dl,void *element);                /* Searches an element */
    Dlist *(*Copy)(Dlist *dl);                  /* Copies all items into a new Dlist */
	int (*CopyElement)(Dlist *l,size_t idx,void *outbuf);/* Copies element to buffer */
    size_t (*IndexOf)(Dlist *AL,void *SearchedElement);     /* index or -1 if not found */
    int (*InsertAt)(Dlist *AL,size_t idx,void *newval); /* Inserts value at position */
    void *(*GetElement)(Dlist *AL,int idx);             /* Returns  data at position */
    int (*Erase)(Dlist *AL,void *);                /* erases the given data if found */
    int (*EraseAt)(Dlist *AL,size_t idx);              /* erases element at position */
    int (*Finalize)(Dlist *AL);                                 /* Frees memory used */
    int (*PushFront)(Dlist *AL,void *str);             /* Pushes an element at start */
    int (*PopFront)(Dlist *AL,void *result);     /* Pops the first element the Dlist */
    int (*PushBack)(Dlist *AL,void *str);                /* Pushes an element at end */
    int (*PopBack)(Dlist *AL,void *result);       /* Pops the last element the Dlist */
    void *(*ReplaceAt)(Dlist *AL,size_t idx,void *newval); /* Replace element at pos */
	                                                  /* Inserts a list into another */
	Dlist *(*Splice)(Dlist *list,void *pos,Dlist *toInsert,int direction);
    bool (*Sort)(Dlist *l);                                        /* Sorts the list */
    Dlist *(*Reverse)(Dlist *l);                                /* Reverses the list */
    Dlist *(*GetRange)(Dlist *l,size_t start,size_t end);            /* Gets a range */
    bool (*Equal)(Dlist *l1,Dlist *l2);   /* Compares two lists (Shallow comparison) */
    int (*Append)(Dlist *l1,Dlist *l2);         /* Add a Dlist at the end of another */
	                                   /* Sets the comparison function for this list */
    CompareFunction (*SetCompareFunction)(Dlist *l,CompareFunction fn);
                 /* Call a callback function with each element and an extra argument */
    void (*Apply)(Dlist *L,int(Applyfn)(void *elem,void *extraArg),void *extraArg);
    ErrorFunction (*SetErrorFunction)(Dlist *L,ErrorFunction);
    ContainerMemoryManager *Allocator;
	Iterator *(*newIterator)(Dlist *);
	void (*deleteIterator)(Iterator *);
	int (*UseHeap)(Dlist *L, ContainerMemoryManager *m);
	Dlist *(*Load)(FILE *stream, ReadFunction loadFn,void *arg);
	int (*Save)(Dlist *L,FILE *stream, SaveFunction saveFn,void *arg);

} DlistInterface;

extern DlistInterface iDlist;

/****************************************************************************
 *                                                                          *
 *           The array list                                                 *
 ****************************************************************************/

/* Forward declaration of the string collection type */

/* Definition of the functions associated with this type. */
typedef struct {
	ArrayList *(*Create)(size_t elementsize,size_t startsize);
	size_t (*GetElementSize)(const ArrayList *AL);
    /* Returns the number of elements stored */
    size_t (*Size)(const ArrayList *AL);
    /* Is this collection read only? */
    unsigned (*GetFlags)(const ArrayList *AL);
    /* Sets this collection read-only or unsets the read-only flag */
    unsigned (*SetFlags)(ArrayList *AL,unsigned flags);
    /* Adds one element at the end. Given element is copied */
    int (*Add)(ArrayList *AL,void *newval);
    /* Adds a NULL terminated table of strings */
    int (*AddRange)(ArrayList *AL,void **newvalues);
	ArrayList *(*GetRange)(ArrayList *AL, size_t start, size_t end);
    /* Clears all data and frees the memory */
    int (*Clear)(ArrayList *AL);
    /*Case sensitive search of a character string in the data */
    bool (*Contains)(ArrayList *AL,void *str,void *ExtraArgs);
	    /* Copies an array list */
    ArrayList *(*Copy)(ArrayList *AL);
	int (*CopyElement)(ArrayList *AL,size_t idx,void *outbuf);
    /* Copies all strings into a NULL terminated vector */
    void **(*CopyTo)(ArrayList *AL);
    /*Returns the 1 based index of the given string or 0 if not found */
    size_t (*IndexOf)(ArrayList *AL,void *data,void *ExtraArgs);
    /* Inserts a string at the position zero. */
    int (*Insert)(ArrayList *AL,void *);
    /* Inserts a string at the given position */
    int (*InsertAt)(ArrayList *AL,size_t idx,void *newval);
    /* Returns the string at the given position */
	int (*Equal)(ArrayList *first,ArrayList *second);
    void *(*GetElement)(ArrayList *AL,size_t idx);
    /* erases the given string if found */
    int (*Erase)(ArrayList *AL,void *);
    /*erases the string at the indicated position */
    int (*EraseAt)(ArrayList *AL,size_t idx);
    /* Frees the memory used by the collection */
    int (*Finalize)(ArrayList *AL);
    /* Returns the current capacity of the collection */
    size_t (*GetCapacity)(const ArrayList *AL);
    /* Sets the capacity if there are no items in the collection */
    bool (*SetCapacity)(ArrayList *AL,int newCapacity);
    /* Calls the given function for all strings in the given collection.
       "Arg" is a user supplied argument (can be NULL) that is passed
       to the function to call */
    void (*Apply)(ArrayList *AL,int (*Applyfn)(void *element,void * arg),void *arg);
    /* Pushes a string, using the collection as a stack */
    int (*Push)(ArrayList *AL,void *str);
    /* Pops the last string off the collection */
    void * (*Pop)(ArrayList *AL);
    /* Replaces the character string at the given position with a new one */
    void *(*ReplaceAt)(ArrayList *AL,size_t idx,void *newval);
	ErrorFunction (*SetErrorFunction)(ArrayList *AL,ErrorFunction);
	CompareFunction (*SetCompareFunction)(ArrayList *l,CompareFunction fn);
	size_t (*Sizeof)(ArrayList *AL);
	bool (*Sort)(ArrayList *AL);
	ContainerMemoryManager *Allocator;
	Iterator *(*newIterator)(ArrayList *AL);
	void (*deleteIterator)(Iterator *);
	/* Writes the string collection in binary form to a stream */
	int (*Save)(ArrayList *AL,FILE *stream, SaveFunction saveFn,void *arg);
	ArrayList *(*Load)(FILE *stream, ReadFunction readFn,void *arg);
} ArrayListInterface;

extern ArrayListInterface iArrayList;

/* ----------------------------------Dictionary interface -----------------------*/

typedef struct _Dictionary Dictionary;

typedef struct {
	Dictionary *(*Create)(size_t ElementSize,size_t hint);
	size_t (*GetElementSize)(Dictionary *d);
    /* Returns the number of elements stored */
    size_t (*Size)(Dictionary *Dict);
    /* Get the flags */
    unsigned (*GetFlags)(Dictionary *Dict);
    /* Sets the flags */
    unsigned (*SetFlags)(Dictionary *Dict,unsigned flags);
    /* Adds one element. Given string is copied */
    int (*Add)(Dictionary *Dict,const unsigned char *key,void *Data);
    /* Clears all data and frees the memory */
    int (*Clear)(Dictionary *Dict);
	bool (*Contains)(Dictionary *dict,const unsigned char *key);
    /* Returns the string at the given position */
    const void *(*GetElement)(const Dictionary *Dict,const unsigned char *Key);
    /* erases the given string if found */
	int (*Equal)(const Dictionary *d1,const Dictionary *d2);
    bool (*Erase)(Dictionary *Dict,const unsigned char *);
    /* Frees the memory used by the dictionary */
    int (*Finalize)(Dictionary *Dict);
    /* Calls the given function for all strings. "Arg" is a used supplied argument */
    /* (that can be NULL) that is passed to the function to call */
    bool (*Apply)(Dictionary *Dict,int (*Applyfn)(const unsigned char *Key,
	                                              const void *data,void *arg),void *arg);
    /* Set or unset the error function */
    ErrorFunction (*SetErrorFunction)(Dictionary *Dict,ErrorFunction fn);
    StringCollection *(*GetKeys)(Dictionary *Dict);
    size_t (*Sizeof)(Dictionary *dict);
    ContainerMemoryManager *Allocator;
	Iterator *(*newIterator)(Dictionary *dict);
	void (*deleteIterator)(Iterator *);
	ArrayList *(*CastToArray)(Dictionary *);
	int (*Save)(Dictionary *Dict,FILE *stream, SaveFunction saveFn,void *arg);
	Dictionary * (*Load)(FILE *stream, ReadFunction readFn, void *arg);
	Dictionary *(*Copy)(Dictionary *dict);
	int (*CopyElement)(const Dictionary *Dict,const unsigned char *Key,void *outbuf);
} DictionaryInterface;

extern DictionaryInterface iDictionary;
/* ----------------------------------Hash table interface -----------------------*/
typedef struct _HashTable HashTable;
typedef unsigned int (*HashFunction)(const char *char_key, size_t *klen);

typedef struct {
	/* Construction */
	HashTable *(*Create)(size_t ElementSize);
    /* Returns the number of elements stored */
    size_t (*Size)(const HashTable *HT);
	size_t (*Sizeof)(const HashTable *HT);
    /* Get the flags */
    unsigned (*GetFlags)(const HashTable *HT);
    /* Sets th flags */
    unsigned (*SetFlags)(HashTable *HT,unsigned flags);
    /* Adds one element. Given string is copied */
    int (*Add)(HashTable *HT,const void *key,size_t klen,const void *Data);
    /* Clears all data and frees the memory */
    int (*Clear)(HashTable *HT);
    /* Returns the element with the given key */
    void *(*GetElement)(const HashTable *HT,const void *Key,size_t klen);
	int (*Search)(HashTable *ht,int (*Comparefn)(void *rec, const void *key,size_t klen,const void *value), void *rec);
    /* erases the given string if found */
    int (*Erase)(HashTable *HT,void *key,size_t klen);
    /* Frees the memory used by the HashTable */
    int (*Finalize)(HashTable *HT);
    /* Calls the given function for all strings. "Arg" is a used supplied argument */
    /* that can be NULL that is passed to the function to call */
    bool (*Apply)(HashTable *HT,int (*Applyfn)(void *Key,size_t klen,void *data,void *arg),void *arg);
    /* Set or unset the error function */
    ErrorFunction (*SetErrorFunction)(HashTable *HT,ErrorFunction fn);
    int (*Resize)(HashTable *HT,size_t newSize);
	int (*Replace)(HashTable *HT,const void *key, size_t klen,const void *val);
	HashTable *(*Copy)(const HashTable *Orig,Pool *pool);
    ContainerMemoryManager *Allocator;
	HashFunction (*SetHashFunction)(HashTable *ht, HashFunction hf);
	/* Copy hash table base into overlay. If duplicates, table "base" wins*/
	HashTable *(*Overlay)(Pool *p, const HashTable *overlay, const HashTable *base);
	/*
	 Merge two hash tables. If duplicate keys are found, call the "merger" callback to decide
	 which element should be kept in the final hash table
	 */
	HashTable *(*Merge)(Pool *p, const HashTable *overlay, const HashTable *base,
						void * (*merger)(Pool *p,
										 const void *key,
										 size_t klen,
										 const void *h1_val,
										 const void *h2_val,
										 const void *data),
						const void *data);
	Iterator *(*newIterator)(HashTable *);
	void (*deleteIterator)(Iterator *);
	int (*Save)(HashTable *HT,FILE *stream, SaveFunction saveFn,void *arg);
	HashTable *(*Load)(FILE *stream, ReadFunction readFn, void *arg);
} HashTableInterface;

extern HashTableInterface iHashTable;
void qsortEx(void *base, size_t num, size_t width,CompareFunction cmp, CompareInfo *ExtraArgs);

/* ---------------------------------------------------------------------------
 *                                                                           *
 *                         Binary searchtrees                                *
 * ------------------------------------------------------------------------  */
typedef struct tagBinarySearchTree BinarySearchTree;

typedef struct tagBinarySearchTreeInterface {
	BinarySearchTree *(*Create)(size_t ElementSize);
    /* Returns the number of elements stored */
    size_t (*Size)(BinarySearchTree *ST);
    /* Get the flags */
    unsigned (*GetFlags)(BinarySearchTree *ST);
    /* Sets the flags */
    unsigned (*SetFlags)(BinarySearchTree *ST, unsigned flags);
    /* Adds one element. Given data is copied */
    int (*Add)(BinarySearchTree *ST, const void *Data);
	/* Like Add but allows for an extra argument to be passed to the
	   comparison function */
	int (*Insert)(BinarySearchTree * ST, const void *Data, void *ExtraArgs);
    /* Clears all data and frees the memory */
    int (*Clear)(BinarySearchTree *ST);
    /* erases the given data if found */
    bool (*Erase)(BinarySearchTree *ST, const void *,void *);
    /* Frees the memory used by the tree */
    int (*Finalize)(BinarySearchTree *ST);
    /* Calls the given function for all nodes. "Arg" is a used supplied argument */
    /* that can be NULL that is passed to the function to call */
    bool (*Apply)(BinarySearchTree *ST,int (*Applyfn)(const void *data,void *arg),void *arg);
	/* Find a given data in the tree */
	void *(*Find)(BinarySearchTree *ST,void *data);
    /* Set or unset the error function */
    ErrorFunction (*SetErrorFunction)(BinarySearchTree *ST, ErrorFunction fn);
	CompareFunction (*SetCompareFunction)(BinarySearchTree *ST,CompareFunction fn);
	size_t (*Sizeof)(BinarySearchTree *ST);
	CompareFunction Compare;
	BinarySearchTree *(*Merge)(BinarySearchTree *left, BinarySearchTree *right, const void *data);
	bool (*Equal)(const BinarySearchTree *left, const BinarySearchTree *right);
    ContainerMemoryManager *Allocator;
	Iterator *(*newIterator)(BinarySearchTree *);
	void (*deleteIterator)(Iterator *);
} BinarySearchTreeInterface;

extern BinarySearchTreeInterface iBinarySearchTree;
/* -------------------------------------------------------------------------------------
 *                                                                                     *
 *                                Red Black trees                                      *
 * ----------------------------------------------------------------------------------  */

typedef struct tagRedBlackTree RedBlackTree;

typedef struct tagRedBlackTreeInterface {
	RedBlackTree *(*Create)(size_t ElementSize,size_t KeySize);
	size_t (*GetElementSize)(RedBlackTree *b);
    size_t (*Size)(RedBlackTree *ST); /* Returns the number of elements stored */
    unsigned (*GetFlags)(RedBlackTree *ST); /* Gets the flags */
    unsigned (*SetFlags)(RedBlackTree *ST, unsigned flags);
    /* Adds one element. Given data is copied */
    int (*Add)(RedBlackTree *ST, const void *Key,const void *Data);
	/* Like Add but allows for an extra argument to be passed to the
	 comparison function */
	int (*Insert)(RedBlackTree *RB, const void *Key, const void *Data, void *ExtraArgs);
    /* Clears all data and frees the memory */
    int (*Clear)(RedBlackTree *ST);
    /* erases the given data if found */
    bool (*Erase)(RedBlackTree *ST, const void *,void *);
    /* Frees the memory used by the tree */
    int (*Finalize)(RedBlackTree *ST);
    /* Calls the given function for all nodes. "Arg" is a used supplied argument */
    /* that can be NULL that is passed to the function to call */
    bool (*Apply)(RedBlackTree *ST,int (*Applyfn)(const void *data,void *arg),void *arg);
	/* Find a given data in the tree */
	void *(*Find)(RedBlackTree *ST,void *key_data,void *ExtraArgs);
    /* Set or unset the error function */
    ErrorFunction (*SetErrorFunction)(RedBlackTree *ST, ErrorFunction fn);
	CompareFunction (*SetCompareFunction)(RedBlackTree *ST,CompareFunction fn);
	size_t (*Sizeof)(RedBlackTree *ST);
	CompareFunction Compare;
	ContainerMemoryManager *Allocator;
	Iterator *(*newIterator)(RedBlackTree *);
	void (*deleteIterator)(Iterator *);
} RedBlackTreeInterface;

extern RedBlackTreeInterface iRedBlackTree;
/* -------------------------------------------------------------------------------------
 *                                                                                     *
 *                            Balanced Binary trees                                    *
 *                                                                                     *
 * ----------------------------------------------------------------------------------  */
typedef struct tagBB_Tree BB_Tree;

typedef struct tagBB_TreeInterface {
	BB_Tree *(*Create)(size_t ElementSize,ContainerMemoryManager *m);
	size_t (*GetElementSize)(BB_Tree *d);
    size_t (*Size)(BB_Tree *ST);   /* Returns the number of elements stored */
    unsigned (*GetFlags)(BB_Tree *ST); /* Get the flags */
    unsigned (*SetFlags)(BB_Tree *ST, unsigned flags); /* Sets the flags */
    int (*Add)(BB_Tree *ST, void *Data); /* Adds one element. Given data is copied */
	/* Like Add but allows for an extra argument to be passed to the
	 comparison function */
	int (*Insert)(BB_Tree *RB, const void *Data, void *ExtraArgs);
    int (*Clear)(BB_Tree *ST); /* Clears all elements */
    bool (*Erase)(BB_Tree *tree, void *element);  /* erases the given data if found */
    int (*Finalize)(BB_Tree *ST);  /* Frees the memory used by the tree */
    /* Calls the given function for all nodes. "Arg" is a used supplied argument */
    /* that can be NULL that is passed to the function to call */
    bool (*Apply)(BB_Tree *ST,int (*Applyfn)(const void *data,void *arg),void *arg);
	/* Find a given data in the tree */
	void *(*Find)(BB_Tree *tree,void *element,void *ExtraArgs);
    /* Set or unset the error function */
    ErrorFunction (*SetErrorFunction)(BB_Tree *ST, ErrorFunction fn);
	CompareFunction (*SetCompareFunction)(BB_Tree *ST,CompareFunction fn);
	size_t (*Sizeof)(BB_Tree *ST);
	ContainerMemoryManager *Allocator;
	Iterator *(*newIterator)(BB_Tree *);
	void (*deleteIterator)(Iterator *);
} BB_TreeInterface;

extern BB_TreeInterface iBB_Tree;
/* -------------------------------------------------------------------------------------
 *                                                                                     *
 *                            Bit strings                                              *
 *                                                                                     *
 * ----------------------------------------------------------------------------------  */
typedef struct _BitString BitString;
#define BIT_TYPE unsigned char
#define BITSTRING_READONLY	1

/* Definition of the functions associated with this type. */
typedef struct {
	/* creates a bit string */
	BitString *(*Create)(size_t bitlen);
    size_t (*Size)(BitString *BitStr); /* Returns the number of elements stored */
    unsigned (*GetFlags)(BitString *BitStr); /* gets the flags */
    unsigned (*SetFlags)(BitString *BitStr,unsigned flags);/* Sets flags */
    int (*Add)(BitString *BitStr,bool);     /* Adds one bit at the end. */
    /* Clears all data and frees the memory */
    int (*Clear)(BitString *BitStr);
    /* Searches a bit in the data */
    int (*Contains)(BitString *BitStr,BitString *str,void *ExtraArgs);
    /*Returns the index of the given string or -1 if not found */
    size_t (*IndexOf)(BitString *BitStr,bool SearchedBit,void *ExtraArgs);
    /* Inserts a string at the position zero. */
    size_t (*Insert)(BitString *BitStr,bool bit);
    /* Inserts a string at the given position */
    size_t (*InsertAt)(BitString *BitStr,size_t idx,bool bit);
    /* Returns the string at the given position */
    int (*GetElement)(BitString *BitStr,size_t idx);
	/* sets the given element to a new value */
	bool (*SetElement)(BitString *bs,size_t position,bool b);
    /* erases the given string if found */
    int (*Erase)(BitString *BitStr,bool bit);
    /*erases the string at the indicated position */
    bool (*EraseAt)(BitString *BitStr,size_t idx);
    /* Frees the memory used by the collection */
    int         (*Finalize)(BitString *BitStr);
    /* Returns the current capacity of the collection */
    size_t      (*GetCapacity)(BitString *BitStr);
    /* Sets the capacity if there are no items in the collection */
    int         (*SetCapacity)(BitString *BitStr,size_t newCapacity);
    /* Calls the given function for all strings in the given collection.
	 "Arg" is a user supplied argument (can be NULL) that is passed
	 to the function to call */
    bool        (*Apply)(BitString *BitStr,int (*Applyfn)(bool ,void * arg),void *arg);
    /* Pushes a bit */
    bool        (*Push)(BitString *BitStr,int val);
    /* Pops the last bit off */
    bool       (*Pop)(BitString *BitStr);
    /* Replaces the bit at the given position with a new one */
    int        (*ReplaceAt)(BitString *BitStr,size_t idx,bool newval);
	ContainerMemoryManager *Allocator;
	Iterator  *(*newIterator)(BitString *);
	size_t     (*Sizeof)(BitString *b);
	/* Bit string specific functions */
	BitString *(*Or)(BitString *left,BitString *right);
	BitString *(*OrAssign)(BitString *bsl,BitString *bsr);
	BitString *(*And)(BitString *bsl,BitString *bsr);
	BitString *(*AndAssign)(BitString *bsl,BitString *bsr);
	BitString *(*Xor)(BitString *bsl,BitString *bsr);
	BitString *(*XorAssign)(BitString *bsl,BitString *bsr);
	BitString *(*Not)(BitString *bsl);
	BitString *(*NotAssign)(BitString *bsl);
	uintmax_t  (*PopulationCount)(BitString *b);
	uintmax_t  (*BitBlockCount)(BitString *b);
	bool       (*Equal)(BitString *bsl,BitString *bsr);
	bool       (*LessEqual)(BitString *bsl,BitString *bsr);
	BitString *(*Reverse)(BitString *b);
	BitString *(*GetRange)(BitString *b,size_t start,size_t end);
	BitString *(*StringToBitString)(unsigned char *);
	BitString *(*ObjectToBitString)(unsigned char *p,size_t size);
	BitString *(*LeftShift)(BitString *bs,size_t shift);
	BitString *(*RightShift)(BitString *bs,size_t shift);
	size_t     (*Print)(BitString *b,size_t bufsiz,unsigned char *out);
	BitString *(*Append)(BitString *left,BitString *right);
	BitString *(*Copy)(BitString *);
	bool       (*Set)(BitString *,size_t start,size_t stop,bool newval);
	int        (*Save)(BitString *bitstr,FILE *stream, SaveFunction saveFn,void *arg);
	BitString *(*Load)(FILE *stream, ReadFunction saveFn,void *arg);
} BitStringInterface;

extern BitStringInterface iBitString;

/* -------------------------------------------------------------------------------------
 *                                                                                     *
 *                            Bloom filter                                             *
 *                                                                                     *
 * ----------------------------------------------------------------------------------  */
typedef struct tagBloomFilter BloomFilter;
typedef struct tagBloomFilterInterface {
	BloomFilter *(*Create)(size_t MaxElements,double probability);
	size_t (*Add)(BloomFilter *b,const void *key,size_t keylen);
	int (*Find)(BloomFilter *b,const void *key,size_t keylen);
	int (*Clear)(BloomFilter *b);
	int (*Finalize)(BloomFilter *b);
} BloomFilterInterface;
extern BloomFilterInterface iBloomFilter;

int decode_ule128(FILE *stream, size_t *val);
int encode_ule128(FILE *stream, size_t val);
#endif
