#ifndef _WIN64
#include "exceptions.h"
#include <seh.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <gc.h>

/* one GLOBAL structure */
struct ExceptionGlobals GLOBALS;

/* private functions */

static void DefaultHandler (struct Exception *exc){
	char *violation_message = NULL;
	assert(exc);
	switch (exc->code){
		case PRECONDITION_VIOLATION:
			violation_message = "PRECONDITION VIOLATION";
			break;
		case POSTCONDITION_VIOLATION:
			violation_message = "POSTCONDITION_VIOLATION";
			break;
		case INVARIANT_VIOLATION:
			violation_message = "INVARIANT VIOLATION";
			break;
		case CHECK_VIOLATION:
			violation_message = "CHECK FAILED";
			break;
		default:
			violation_message = "EXCEPTION";
			break;
	}
	fprintf(stderr, "%s in %s,\nfunction '%s', line %u.\n",
			violation_message, exc->file_name, exc->function_name,
			exc->line);

}


void DefaultHandlerAbort (struct Exception *exc){
	DefaultHandler(exc);
	abort();
}

void DefaultHandlerWarn (struct Exception *exc){
	DefaultHandler(exc);
}

void DefaultHandlerExit (struct Exception *exc, int exit_code){
	DefaultHandler(exc);
	exit (exit_code);
}

static void InitializeHandlers(void)
{
	for (int i = 0; i < MAX_HANDLERS; ++i){
		GLOBALS.Handlers[i] = DefaultHandlerAbort;
	}
}

int HandlerIndexInRange (unsigned long num)
{
	if(num > USER_EXCEPTION_OFFSET)
		return 0;
	int result = num < (MAX_HANDLERS - USER_EXCEPTION_OFFSET);
	return result;
}



static void InitializeExceptionsField(void){
	struct Exception *exc = NULL;
	int i;
	for (i = 0; i < MAX_EXCEPTIONS; ++i){
		exc = GC_MALLOC(sizeof(*exc));
		if (NULL != exc){
			GLOBALS.ExeptionsField[i] = exc;
		} else {
			abort();
		}
	}
}


void InitializeExceptions(void){
	InitializeHandlers();
	InitializeExceptionsField();
}


static long calculate_place (unsigned long place){
	assert(place > USER_EXCEPTION_OFFSET);
	return place - USER_EXCEPTION_OFFSET;
}


int RegisterHandler (unsigned long place, ExceptionHandler handler){
	long act_place = calculate_place(place);
	int result = 0;
	if (HandlerIndexInRange(act_place)){
		GLOBALS.Handlers[act_place] = handler;
		result = 1;
	}
	return result;
}


int DeregisterHandler (unsigned long place){
	long act_place = calculate_place(place);
	int result = 0;
	if (HandlerIndexInRange(act_place)){
		GLOBALS.Handlers[act_place] = NULL;
		result = 1;
	}
	return result;
}


ExceptionHandler GetHandler(unsigned long place){
	long act_place = calculate_place(place);
	return GLOBALS.Handlers[act_place];
}

void __declspec(dllexport) RunHandler(unsigned long code)
{
	EXCEPTION_POINTERS *pexp = GetExceptionInformation();
	struct Exception *exc = NULL;
	ExceptionHandler handler = NULL;
	if (HandlerIndexInRange(code)){
		handler = GetHandler(code);
	} else {
		handler = DefaultHandlerAbort;
	}
	exc = (struct Exception*) pexp->ExceptionRecord->ExceptionInformation[0];
	if (exc)
		handler(exc);
	else {
		fprintf(stderr,"Exception 0x%x\n",code);
		abort();
	}
}


struct Exception * GetAnExceptionStructure(void)
{
	static int counter;
	static int initialized;

	if (initialized == 0) {
		InitializeExceptions();
		initialized = 1;
	}

	++counter;
	if (counter == MAX_EXCEPTIONS ){
		counter = 0;
	}
	return GLOBALS.ExeptionsField[counter];
}

#endif 



