/* -------------------------------------------------------------------
                        List routines sample implementation
                        -----------------------------------
This routines handle the List container class. This is a very general implementation
and efficiency considerations aren't yet primordial. Lists can have elements
of any size. This implement single linked Lists.
   The design goals here are just correctness and showing how the implementation of
the proposed interface COULD be done.
---------------------------------------------------------------------- */

#ifndef DEFAULT_START_SIZE
#define DEFAULT_START_SIZE 20
#endif
#include "containers.h"
#ifndef INT_MAX
#define INT_MAX (((unsigned)-1) >> 1)
#endif
typedef struct _list_element {
    struct _list_element *Next;
    char Data[MINIMUM_ARRAY_INDEX];
} list_element;
struct _List {
    ListInterface *VTable;      /* Methods table */
    size_t count;               /* in elements units */
    unsigned Flags;	unsigned timestamp;         /* Changed at each modification */
    size_t ElementSize;         /* Size (in bytes) of each element */
    list_element *Last;         /* The last item */
    list_element *First;        /* The contents of the list start here */
    CompareFunction Compare;    /* Element comparison function */
    ErrorFunction RaiseError;   /* Error function */
	ContainerHeap *Heap;
	ContainerMemoryManager *Allocator;
};
static size_t IndexOf(List *AL,void *SearchedElement,void *ExtraArgs);
static int RemoveAt(List *AL,size_t idx);
#define CONTAINER_LIST_READONLY	1
#define CONTAINER_LIST_SMALL	2
#define CHUNK_SIZE	1000
/* The function Push is identical to Insert */
#define Push Insert

/*------------------------------------------------------------------------
 Procedure:     new_link ID:1
 Purpose:       Allocation of a new list element. If the element
                size is zero, we have an heterogenous
                list, and we allocate just a pointer to the data
                that is maintained by the user.
                Note that we allocate the size of a list element
                plus the size of the data in a single
                block. This block should be passed to the FREE
                function.

 Input:         The list where the new element should be added and a
                pointer to the data that will be added (can be
                NULL).
 Output:        A pointer to the new list element (can be NULL)
 Errors:        If there is no memory returns NULL
------------------------------------------------------------------------*/
static list_element *new_link(List *li,void *data,const char *fname)
{
	list_element *result;

	if (li->Flags & CONTAINER_LIST_SMALL || li->Heap == NULL) {
		result = MALLOC(li,sizeof(*result)+li->ElementSize);
	}
	else result = iHeap.newObject(li->Heap);
	if (result == NULL) {
		li->RaiseError(fname,CONTAINER_ERROR_NOMEMORY);
	}
	else {
		result->Next = NULL;
		memcpy(&result->Data,data,li->ElementSize);
	}
	return result;
}
static int DefaultListCompareFunction(const void *left,const void *right,CompareInfo *ExtraArgs)
{
        size_t siz=((List *)ExtraArgs->Object)->ElementSize;
        return memcmp(left,right,siz);
}

/*------------------------------------------------------------------------
 Procedure:     Contains ID:1
 Purpose:       Determines if the given data is in the container
 Input:         The list and the data to be searched
 Output:        Returns 1 (true) if the data is in there, false
                otherwise
 Errors:        The same as the function IndexOf
------------------------------------------------------------------------*/
static bool Contains(List *l,void *data)
{
	return (IndexOf(l,data,NULL) == 0) ? 0 : 1;
}

/*------------------------------------------------------------------------
 Procedure:     Clear ID:1
 Purpose:       Reclaims all memory used by a list. The list header
                itself is not reclaimed but zeroed. Note that the
                list must be writable.
 Input:         The list to be cleared
 Output:        Returns the number of elemnts that were in the list
                if OK, CONTAINER_ERROR_READONLY otherwise.
 Errors:        None
------------------------------------------------------------------------*/
static int Clear(List *l)
{
	if (l == NULL) {
		iError.RaiseError("iList.Create",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (l->Flags & CONTAINER_LIST_READONLY) {
		l->RaiseError("iList.Clear",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}
#ifdef NO_GC
	if (l->Heap)
		iHeap.Destroy(l->Heap);
	else {
		list_element *rvp = l->First,*tmp;

		while (rvp) {
			tmp = rvp;
			rvp = rvp->Next;
			FREE(l,tmp);
		}
	}
#endif
	memset(l,0,sizeof(List));
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     Add ID:1
 Purpose:       Adds an element to the list
 Input:         The list and the elemnt to be added
 Output:        Returns the number of elements in the list or a
                negative error code if an error occurs.
 Errors:        The element to be added can't be NULL, and the list
                must be writable.
------------------------------------------------------------------------*/
static int Add(List *l,void *elem)
{
	list_element *newl;

	if (l == NULL) {
		iError.RaiseError("iList.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (elem == NULL) {
		l->RaiseError("iList.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (l->Flags &CONTAINER_LIST_READONLY) {
		return CONTAINER_ERROR_READONLY;
	}
	newl = new_link(l,elem,"iList.Add");
	if (newl == 0)
		return CONTAINER_ERROR_NOMEMORY;
	if (l->count ==  0) {
		l->First = newl;
	}
	else {
		l->Last->Next = newl;
	}
	l->Last = newl;
	l->timestamp++;
	++l->count;
	return 1;
}

#if 0
static int Memset(List *l, void *elem, size_t repeat)
{
	list_element *newl;
	if (l == NULL) {
		iError.RaiseError("iList.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (elem == NULL) {
		l->RaiseError("iList.Add",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (l->Flags &CONTAINER_LIST_READONLY) {
		return CONTAINER_ERROR_READONLY;
	}
	l->timestamp++;
	while (repeat > 0) {
		newl = new_link(l, elem, "iList.Add");
		if (newl == NULL) {
			return CONTAINER_ERROR_NOMEMORY;
		}
		if (l->count ==  0) {
			l->First = newl;
		}
		else {
			l->Last->Next = newl;
		}
		l->Last = newl;
		l->count++;
		repeat--;
	}
	return 1;
}
#endif

/*------------------------------------------------------------------------
 Procedure:     SetReadOnly ID:1
 Purpose:       Sets/Unsets the read only flag.
 Input:         The list and the new value
 Output:        The old value of the flag
 Errors:        None
------------------------------------------------------------------------*/
static unsigned SetFlags(List *l,unsigned newval)
{
	unsigned result;

	result = l->Flags;
	l->Flags = newval;
	return result;
}

/*------------------------------------------------------------------------
 Procedure:     IsReadOnly ID:1
 Purpose:       Queries the read only flag
 Input:         The list
 Output:        The state of the flag
 Errors:        None
------------------------------------------------------------------------*/
static unsigned GetFlags(List *l)
{
	return l->Flags;
}

/*------------------------------------------------------------------------
 Procedure:     Size ID:1
 Purpose:       Returns the number of elements in the list
 Input:         The list
 Output:        The number of elements
 Errors:        None
------------------------------------------------------------------------*/
static size_t Size(List *l)
{
	return l->count;
}

/*------------------------------------------------------------------------
 Procedure:     SetCompareFunction ID:1
 Purpose:       Defines the function to be used in comparison of
                list elements
 Input:         The list and the new comparison function. If the new
                comparison function is NULL no changes are done.
 Output:        Returns the old function
 Errors:        None
------------------------------------------------------------------------*/
static CompareFunction SetCompareFunction(List *l,CompareFunction fn)
{
	CompareFunction oldfn = l->Compare;


	if (l == NULL) {
		iError.RaiseError("iList.Copy",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	if (fn != NULL) { /* Treat NULL as an enquiry to get the compare function */
		if (l->Flags&CONTAINER_LIST_READONLY) {
			l->RaiseError("SetCompareFunction",CONTAINER_LIST_READONLY);
		}
		else l->Compare = fn;
	}
	return oldfn;
}

/*------------------------------------------------------------------------
 Procedure:     Copy ID:1
 Purpose:       Copies a list. The result is fully allocated (list
                elements and data).
 Input:         The list to be copied
 Output:        A pointer to the new list
 Errors:        None. Returns NULL if therfe is no memory left.
------------------------------------------------------------------------*/
static List *Copy(List *l)
{
	List *result;
	list_element *rvp,*newRvp,*last;

	if (l == NULL) {
		iError.RaiseError("iList.Copy",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = iList.Create(l->ElementSize);
	if (result == NULL)
		return NULL;
	rvp = l->First;
	result->Flags = l->Flags;   /* Same flags */
	result->VTable = l->VTable; /* Copy possibly subclassed methods */
	newRvp = new_link(result,rvp->Data,"Copy");
	if (newRvp == NULL) {
		goto memError;
	}
	last = result->First = result->Last = newRvp;
	result->count++;
	rvp = rvp->Next;
	while (rvp) {
		newRvp = new_link(result,rvp->Data,"Copy");
		if (newRvp == NULL) {
memError:
			l->RaiseError("iList.Copy",CONTAINER_ERROR_NOMEMORY);
			result->VTable->Finalize(result);
			return NULL;
		}
		last->Next = newRvp;
		last = last->Next;
		result->Last = newRvp;
		result->count++;
		rvp = rvp->Next;
	}
	return result;
}

/*------------------------------------------------------------------------
 Procedure:     Finalize ID:1
 Purpose:       Reclaims all memory for a list. The data, list
                elements and the list header are reclaimed.
 Input:         The list to be destroyed. It should NOT be read-
                only.
 Output:        Returns the old count or a negative value if an
                error occurs (list not writable)
 Errors:        Needs a writable list
------------------------------------------------------------------------*/
static int Finalize(List *l)
{
	int t=0;

	t = Clear(l);
	if (t < 0)
		return t;
	CurrentMemoryManager->free(l);
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     GetElement ID:1
 Purpose:       Returns the data associated with a given position
 Input:         The list and the position
 Output:        A pointer to the data
 Errors:        NULL if error in the positgion index
------------------------------------------------------------------------*/
static void * GetElement(List *l,size_t position)
{
	list_element *rvp;

	if (l == NULL) {
		iError.RaiseError("iList.GetElement",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	if (position >= l->count) {
		l->RaiseError("GetElement",CONTAINER_ERROR_INDEX);
		return NULL;
	}
	rvp = l->First;
	while (position) {
		rvp = rvp->Next;
		position--;
	}
	return rvp->Data;
}

/*------------------------------------------------------------------------
 Procedure:     CopyElement ID:1
 Purpose:       Returns the data associated with a given position
 Input:         The list and the position
 Output:        A pointer to the data
 Errors:        NULL if error in the positgion index
 ------------------------------------------------------------------------*/
static int CopyElement(List *l,size_t position,void *outBuffer)
{
	list_element *rvp;

	if (l == NULL || outBuffer == NULL) {
		iError.RaiseError("iList.CopyElement",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (position >= l->count) {
		l->RaiseError("iList.CopyElement",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	rvp = l->First;
	while (position) {
		rvp = rvp->Next;
		position--;
	}
	memcpy(outBuffer,rvp->Data,l->ElementSize);
	return 1;
}

static bool ReplaceAt(List *l,size_t position,void *data)
{
	list_element *rvp;

	/* Error checking */
	if (l == NULL || data == NULL) {
		iError.RaiseError("iList.ReplaceAt",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (position >= l->count ) {
		l->RaiseError("ReplaceAt",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	if (l->Flags & CONTAINER_LIST_READONLY)
		return CONTAINER_LIST_READONLY;
	/* Position at the right data item */
	if (position == l->count-1)
		rvp = l->Last;
	else if (position == 0)
		rvp = l->First;
	else  {
		rvp = l->First;
		while (position) {
			rvp = rvp->Next;
			position--;
		}
	}
	/* Replace the data there */
	memcpy(&rvp->Data , data,l->ElementSize);
	l->timestamp++;
	return 1;
}

/*------------------------------------------------------------------------
 Procedure:     GetRange ID:1
 Purpose:       Gets a consecutive sub sequence of the list within
                two indexes.
 Input:         The list, the subsequence start and end
 Output:        A new freshly allocated list. If the source list is
                empty or the range is empty, the result list is
                empty too.
 Errors:        None
------------------------------------------------------------------------*/
static List *GetRange(List *l,size_t start,size_t end)
{
	size_t counter;
	List *result;
	list_element *rvp;;

	if (l == NULL) {
		iError.RaiseError("iList.GetRange",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	result = iList.Create(l->ElementSize);
	result->VTable = l->VTable;
	if (l->count == 0)
		return result;
	if (end >= l->count)
		end = l->count-1;
	if (start > end || start > l->count)
		return result;
	if (start == l->count-1)
		rvp = l->Last;
	else {
		rvp = l->First;
		counter = 0;
		while (counter < start) {
			rvp = rvp->Next;
			counter++;
		}
	}
	while (start <= end && rvp != NULL) {
		int r = result->VTable->Add(result,&rvp->Data);
		if (r < 0)
			break;
		rvp = rvp->Next;
		start++;
	}
	return result;
}

/*------------------------------------------------------------------------
 Procedure:     Equal ID:1
 Purpose:       Compares the data of two lists. They must be of the
                same length, the elements must have the same size
                and the comparison functions must be the same. If
                all this is true, the comparisons are done.
 Input:         The two lists to be compared
 Output:        Returns 1 if the two lists are equal, zero otherwise
 Errors:        None
------------------------------------------------------------------------*/
static bool Equal(List *l1,List *l2)
{
	list_element *link1,*link2;
	CompareFunction fn;
	CompareInfo ci;

	if (l1 == NULL && l2 == NULL)
		return 1;
	if (l1 == l2)
		return 1;
	if (l1 == NULL || l2 == NULL)
		return 0;
	if (l1->count != l2->count)
		return 0;
	if (l1->ElementSize != l2->ElementSize)
		return 0;
	if (l1->count == 0)
		return 1;
	if (l1->Compare != l2->Compare)
		return 0;
	fn = l1->Compare;
	link1 = l1->First;
	link2 = l2->First;
	ci.Object = l1;
	ci.ExtraArgs = NULL;
	while (link1 && link2) {
		if (fn(link1->Data,link2->Data,&ci))
			return 0;
		link1 = link1->Next;
		link2 = link2->Next;
	}
	if (link1 || link2)
		return 0;
	return 1;
}


/*------------------------------------------------------------------------
 Procedure:     PushFront ID:1
 Purpose:       Inserts an element at the start of the list
 Input:         The list and the data to insert
 Output:        The count of the list after insertion, or CONTAINER_ERROR_READONLY
                if the list is not writable or CONTAINER_ERROR_NOMEMORY if there is
				no more memory left.
 Errors:        None
------------------------------------------------------------------------*/
static int PushFront(List *l,void *pdata)
{
	list_element *rvp;

	if (l == NULL || pdata == NULL) {
		iError.RaiseError("iList.PushFront",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (l->Flags & CONTAINER_LIST_READONLY)
		return CONTAINER_ERROR_READONLY;
	rvp = new_link(l,pdata,"Insert");
	if (rvp == NULL)
		return CONTAINER_ERROR_NOMEMORY;
	rvp->Next = l->First;
	l->First = rvp;
	if (l->Last == NULL)
		l->Last = rvp;
	l->count++;
	l->timestamp++;
	return 1;
}


/*------------------------------------------------------------------------
 Procedure:     Pop ID:1
 Purpose:       Takes out the first element of a list.
 Input:         The list
 Output:        The first element or NULL if there is none or the
                list is read only
 Errors:        None
------------------------------------------------------------------------*/
static int PopFront(List *l,void *result)
{
	list_element *le;

	if (l == NULL || result == NULL) {
		iError.RaiseError("iList.PopFront",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (l->Flags & CONTAINER_LIST_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (l->count == 0)
		return 0;
	le = l->First;
	if (l->count == 1) {
		l->First = l->Last = NULL;
	}
	else l->First = l->First->Next;
	l->count--;
	memcpy(result,&le->Data,l->ElementSize);
	if (l->Heap) {
		iHeap.AddToFreeList(l->Heap,le);
	}
	else FREE(l,le);
	l->timestamp++;
	return 1;
}

static int InsertAt(List *l,size_t pos,void *pdata)
{

	if (l == NULL || pdata == NULL) {
		iError.RaiseError("iList.InsertAt",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (pos > l->count) {
		l->RaiseError("iList.InsertAt",CONTAINER_ERROR_INDEX);
		return CONTAINER_ERROR_INDEX;
	}
	if (l->Flags & CONTAINER_LIST_READONLY)
		return CONTAINER_ERROR_READONLY;
	if (pos == l->count) {
		l->VTable->Add(l,pdata);
	}
	else {
		list_element *elem = new_link(l,pdata,"InsertAt");
		if (elem == NULL) {
			l->RaiseError("iList.InertAt",CONTAINER_ERROR_NOMEMORY);
			return CONTAINER_ERROR_NOMEMORY;
		}
		if (pos == 0) {
			elem->Next = l->First;
			l->First = elem;
		}
		else {
			list_element *rvp;
			rvp = l->First;
			while (pos > 0) {
				rvp = rvp->Next;
				pos--;
			}
			elem->Next = rvp->Next;
			rvp->Next = elem;
		}
	}
	l->timestamp++;
	return 1;
}

static int Remove(List *l,void *elem)
{
	size_t idx;

	idx = IndexOf(l,elem,NULL);
	if (idx == 0)
		return 0;
	return RemoveAt(l,idx-1);
}

static int RemoveRange(List *l,size_t start,size_t end)
{
	list_element *rvp,*start_pos,*tmp;
	size_t toremove;
	if (l == NULL) {
		iError.RaiseError("iList.RemoveRange",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (end > l->count)
		end = l->count-1;
	if (start >= l->count)
		return 0;
	if (start > end)
		return 0;
	toremove = end - start+1;
	rvp = l->First;
	while (rvp && start > 0) {
		rvp = rvp->Next;
		start--;
	}
	start_pos = rvp;
	if (rvp == NULL) {
		iError.RaiseError("iList.RemoveRange",CONTAINER_ASSERTION_FAILED);
		return CONTAINER_ASSERTION_FAILED;
	}
	while (toremove > 0) {
		tmp = rvp->Next;
		if (l->Heap)
			iHeap.AddToFreeList(l->Heap,rvp);
		else {
			FREE(l,rvp);
		}
		rvp = tmp;
		toremove--;
	}
	start_pos->Next = rvp;
	return 1;
}

static int RemoveAt(List *l,size_t position)
{
	list_element *rvp,*last,*removed;

	if (l == NULL) {
		iError.RaiseError("iList.RemoveRange",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (position >= l->count)
		l->RaiseError("RemoveAt",CONTAINER_ERROR_BADARG);
	if (l->Flags & CONTAINER_LIST_READONLY)
		return CONTAINER_ERROR_READONLY;
	rvp = l->First;
	if (position == 0) {
		removed = l->First;
		if (l->count == 1) {
			l->First = l->Last = NULL;
		}
		else {
			l->First = l->First->Next;
		}
	}
	else if (position == l->count - 1) {
		while (rvp->Next != l->Last)
			rvp = rvp->Next;
		removed = rvp->Next;
		rvp->Next = NULL;
		l->Last = rvp;
	}
	else {
		position--;
		last = rvp;
		rvp = rvp->Next;
		while (position > 0) {
			rvp = rvp->Next;
			position --;
		}
		removed = rvp;
		last->Next = rvp->Next;
	}
	if (l->Heap) {
		iHeap.AddToFreeList(l->Heap,removed);
	}
	else FREE(l,removed);
	l->timestamp++;
	--l->count;
	return 1;
}


static int Append(List *l1,List *l2)
{

	if (l1 == NULL || l2 == NULL) {
		iError.RaiseError("iList.Append",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if ((l1->Flags & CONTAINER_LIST_READONLY) || (l2->Flags & CONTAINER_LIST_READONLY)) {
		l1->RaiseError("iList.Append",CONTAINER_ERROR_READONLY);
		return CONTAINER_ERROR_READONLY;
	}
	if (l2->First)
		l1->Last->Next = l2->First;
	if (l2->Last)
		l1->Last = l2->Last;
    l1->count += l2->count;
	l1->timestamp++;
	l2->Allocator->free(l2);
    return 1;
}

static int Reverse(List *l)
{
	list_element *tmp,*prev,*rvp;

	if (l == NULL) {
		iError.RaiseError("Reverse",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (l->Flags & CONTAINER_LIST_READONLY) {
		return CONTAINER_LIST_READONLY;
	}
	if (l->count < 2)
		return 1;
	rvp = l->First;
	l->First = l->Last;
	prev = NULL;
	while (rvp) {
		tmp = rvp->Next;
		rvp->Next = prev;
	    prev = rvp;
	    rvp = tmp;
	}
	l->Last = prev;
	l->timestamp++;
	return 1;
}



/* Searches a List for a given data item
   Returns a positive integer if found, negative if the end is reached
*/
static size_t IndexOf(List *l,void *ElementToFind,void *ExtraArgs)
{
	list_element *rvp;
	int r,i=1;
	CompareFunction fn;
	CompareInfo ci;

	if (l == NULL || ElementToFind == NULL) {
		iError.RaiseError("IndexOf",CONTAINER_ERROR_BADARG);
		return 0;
	}
	rvp = l->First;
	fn = l->Compare;
	ci.Object = l;
	ci.ExtraArgs = ExtraArgs;
	while (rvp) {
		r = fn(*(char **)rvp->Data,ElementToFind,&ci);
		if (r == 0)
			return i;
		rvp = rvp->Next;
		i++;
	}
	return 0;
}

static bool lcompar (const void *elem1, const void *elem2,CompareInfo *ExtraArgs)
{
	list_element *Elem1 = *(list_element **)elem1;
	list_element *Elem2 = *(list_element **)elem2;
	List *l = (List *)ExtraArgs->Object;
	CompareFunction fn = l->Compare;
	return fn(Elem1->Data,Elem2->Data,ExtraArgs);
}

static int Sort(List *l)
{
	list_element **tab;
	size_t i;
	list_element *rvp;
	CompareInfo ci;

	if (l->count < 2)
		return 1;
	tab = MALLOC(l,l->count * sizeof(list_element *));
	if (tab == NULL) {
		l->RaiseError("Sort",CONTAINER_ERROR_NOMEMORY);
		return CONTAINER_ERROR_NOMEMORY;
	}
	rvp = l->First;
	for (i=0; i<l->count;i++) {
		tab[i] = rvp;
		rvp = rvp->Next;
	}
	ci.Object = l;
	ci.ExtraArgs = NULL;
	qsortEx(tab,l->count,sizeof(list_element *),lcompar,&ci);
	for (i=0; i<l->count-1;i++) {
		tab[i]->Next = tab[i+1];
	}
	tab[l->count-1]->Next = NULL;
	l->Last = tab[l->count-1];
	l->First = tab[0];
	return 1;

}
static int Apply(List *L,int (Applyfn)(void *,void *),void *arg)
{
	list_element *le;
	void *pElem=NULL;

	if (L == NULL || Applyfn == NULL) {
		iError.RaiseError("Apply",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	le = L->First;
	if (L->Flags&CONTAINER_LIST_READONLY) {
		pElem = malloc(L->ElementSize);
		if (pElem == NULL) {
			return CONTAINER_ERROR_NOMEMORY;
		}
	}
	while (le) {
		if (pElem) {
			memcpy(pElem,le->Data,L->ElementSize);
			Applyfn(pElem,arg);
		}
		else Applyfn(le->Data,arg);
		le = le->Next;
	}
	return 1;
}

static ErrorFunction SetErrorFunction(List *l,ErrorFunction fn)
{
	ErrorFunction old;
	old = l->RaiseError;
	l->RaiseError = (fn) ? fn : iError.EmptyErrorFunction;
	return old;
}

static size_t Sizeof(List *l)
{
	return sizeof(List) + l->ElementSize * l->count + l->count *sizeof(list_element);
}

static int UseHeap(List *L, ContainerMemoryManager *m)
{
	if (L == NULL) {
		iError.RaiseError("iList.UseHeap",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	if (L->Heap || L->count) {
		L->RaiseError("UseHeap",CONTAINER_ERROR_NOT_EMPTY);
		return CONTAINER_ERROR_NOT_EMPTY;
	}
	if (m == NULL)
		m = CurrentMemoryManager;
	L->Heap = iHeap.Create(L->ElementSize+sizeof(list_element), m);
	return 1;
}

/* ------------------------------------------------------------------------------ */
/*                                Iterators                                       */
/* ------------------------------------------------------------------------------ */

struct ListIterator {
	Iterator it;
	List *L;
	size_t index;
	list_element *Current;
	size_t timestamp;
};

static void *Seek(Iterator *it,size_t idx)
{
	struct ListIterator *li = (struct ListIterator *)it;
	list_element *rvp = li->L->First;
	if (idx >= li->L->count-1) {
		li->index = li->L->count-1;
		li->Current = li->L->Last;
	}
	else if (idx == 0) {
		li->index = 0;
		li->Current = li->L->First;
	}
	else {
		li->index = idx;
		while (idx > 0) {
			rvp = rvp->Next;
			idx--;
		}
		li->Current = rvp;
	}
	return li->Current;
}

static void *GetNext(Iterator *it)
{
	struct ListIterator *li = (struct ListIterator *)it;
	List *L = li->L;
	void *result;

	if (li->index >= L->count || li->Current == NULL)
		return NULL;
	if (li->timestamp != L->timestamp) {
		L->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	result = li->Current->Data;
	li->Current = li->Current->Next;
	li->index++;
	return result;
}

static void *GetPrevious(Iterator *it)
{
	struct ListIterator *li = (struct ListIterator *)it;
	List *L = li->L;
	list_element *rvp;
	size_t i;

	if (li->index >= L->count || li->index == 0)
		return NULL;
	if (li->timestamp != L->timestamp) {
		L->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	rvp = L->First;
	i=0;
	li->index--;
	if (li->index > 0) {
		while (rvp && i < li->index) {
			rvp = rvp->Next;
			i++;
		}
	}
	li->Current = rvp;
	return rvp->Data;
}

static void *GetCurrent(Iterator *it)
{
	struct ListIterator *li = (struct ListIterator *)it;
    return li->Current;
}
static void *GetFirst(Iterator *it)
{
	struct ListIterator *li = (struct ListIterator *)it;
	List *L = li->L;
	if (L->count == 0)
		return NULL;
	if (li->timestamp != L->timestamp) {
		L->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	li->index = 1;
	li->Current = L->First->Next;
	return L->First->Data;
}

static Iterator *newIterator(List *L)
{
	struct ListIterator *result = MALLOC(L,sizeof(struct ListIterator));
	if (result == NULL)
		return NULL;
	result->it.GetNext = GetNext;
	result->it.GetPrevious = GetPrevious;
	result->it.GetFirst = GetFirst;
	result->it.GetCurrent = GetCurrent;
	result->L = L;
	result->timestamp = L->timestamp;
	return &result->it;
}
static int deleteIterator(Iterator *it)
{
	struct ListIterator *li;
	List *L;

	if (it == NULL) {
		iError.RaiseError("deleteIterator",CONTAINER_ERROR_BADARG);
		return CONTAINER_ERROR_BADARG;
	}
	li = (struct ListIterator *)it;
	L = li->L;
	FREE(L,it);
	return 1;
}

static size_t DefaultSaveFunction(const void *element,void *arg, FILE *Outfile)
{
	const unsigned char *str = element;
	size_t len = *(size_t *)arg;

	return fwrite(str,1,len,Outfile);
}

static int Save(List *L,FILE *stream, SaveFunction saveFn,void *arg)
{
	size_t i;
	list_element *rvp;

	if (saveFn == NULL) {
		saveFn = DefaultSaveFunction;
		arg = &L->ElementSize;
	}
	if (fwrite(L,1,sizeof(List),stream) <= 0)
		return EOF;
	rvp = L->First;
	for (i=0; i< L->count; i++) {
		char *p = rvp->Data;

		if (saveFn(p,arg,stream) <= 0)
			return EOF;
		rvp = rvp->Next;
	}
	return 1;
}

static size_t DefaultLoadFunction(void *element,void *arg, FILE *Infile)
{
	size_t len = *(size_t *)arg;

	return fread(element,1,len,Infile);
}

static List *Load(FILE *stream, ReadFunction loadFn,void *arg)
{
	size_t i;
	List *result,L;
	char *buf;

	if (stream == NULL) {
		iError.RaiseError("iList.Load",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	if (loadFn == NULL) {
		loadFn = DefaultLoadFunction;
		arg = &L.count;
	}
	if (fread(&L,1,sizeof(List),stream) <= 0) {
		return NULL;
	}
	buf = malloc(L.ElementSize);
	if (buf == NULL) {
		iError.RaiseError("iList.Load",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	result = iList.Create(L.ElementSize);
	result->Flags = L.Flags;
	for (i=0; i < L.count; i++) {
		if (loadFn(buf,arg,stream) <= 0)
			break;
		if (Add(result,buf) < 0) {
			iList.Finalize(result);
			result = NULL;
			break;
		}
	}
	free(buf);
	return result;
}

static size_t GetElementSize(List *l)
{
	if (l) {
		return l->ElementSize;
	}
	iError.RaiseError("iList.GetElementSize",CONTAINER_ERROR_BADARG);
	return 0;
}

/*------------------------------------------------------------------------
 Procedure:     Create ID:1
 Purpose:       Allocates a new list object header, initializes the
                VTable field and the element size
 Input:         The size of the elements of the list.
 Output:        A pointer to the newly created list or NULL if
                there is no memory.
 Errors:        If element size is smaller than zero an error
				routine is called. If there is no memory result is
				NULL.
 ------------------------------------------------------------------------*/
static List *Create(size_t elementsize)
{
	List *result;

	if (elementsize == 0) {
		iError.RaiseError("iList.Create",CONTAINER_ERROR_BADARG);
		return NULL;
	}
	else if (elementsize < sizeof(void *))
		elementsize = sizeof(void *);
	result = CurrentMemoryManager->malloc(sizeof(List));
	if (result == NULL) {
		iError.RaiseError("iList.Create",CONTAINER_ERROR_NOMEMORY);
		return NULL;
	}
	memset(result,0,sizeof(List));
	result->ElementSize = elementsize;
	result->VTable = &iList;
	result->Compare = DefaultListCompareFunction;
	result->RaiseError = iError.RaiseError;
	result->Allocator = CurrentMemoryManager;
	return result;
}

ListInterface iList = {
	Create,
	GetElementSize,
	Size,GetFlags,SetFlags,Add,Clear,Contains,Copy,
	IndexOf,InsertAt,GetElement,CopyElement,
	Remove,
	RemoveRange,
	RemoveAt,Finalize,
	PushFront,PopFront,
	ReplaceAt,
	Sort,
	Reverse,
	GetRange,
	Equal,
	Append,
	SetCompareFunction,
	Apply,
	SetErrorFunction,
	DefaultListCompareFunction,
	Sizeof,
	NULL,
	newIterator,
	Seek,
	deleteIterator,
	UseHeap,
	Load,
	Save,
};
/* -------------------------------------------------------------------
 *                           QUEUES
 * -------------------------------------------------------------------*/
typedef struct _Queue {
	QueueInterface *VTable;
	List *Items;
	/* In memory constrained environments, the Allocator of the list */
	/* could be used */
	ContainerMemoryManager *Allocator;
} _Queue;

static size_t QSize(Queue *Q)
{
	return Q->Items->count;
}

static size_t QSizeof(Queue *q)
{
	return sizeof(*q) + q->Items->VTable->Sizeof(q->Items);
}

static int QFinalize(Queue *Q)
{
	Q->Items->VTable->Finalize(Q->Items);
	Q->Allocator->free(Q);
	return 1;
}

static int QClear(Queue *Q)
{
	return Q->Items->VTable->Clear(Q->Items);
}

static unsigned QSetFlags(Queue *Q,unsigned newval)
{
	int result;

	result = Q->Items->Flags;
	Q->Items->Flags = newval;
	return result;
}

static unsigned QGetFlags(Queue *Q)
{
	return Q->Items->Flags;
}

static int Dequeue(Queue *Q,void *result)
{
	return Q->Items->VTable->PopFront(Q->Items,result);
}

static int Enqueue(Queue *Q,void *newval)
{
	return Q->Items->VTable->Add(Q->Items,newval);
}


static Queue *CreateQ(size_t ElementSize)
{
	Queue *result = CurrentMemoryManager->malloc(sizeof(Queue));

	if (result == NULL)
		return NULL;
	result->Items = iList.Create(ElementSize);
	if (result->Items == NULL) {
		CurrentMemoryManager->free(result);
		return NULL;
	}
	result->VTable = &iQueue;
	return result;
}

QueueInterface iQueue = {
	CreateQ,
	QSize,
	QSizeof,
	QGetFlags,
	QSetFlags,
	Enqueue,
	Dequeue,
	QClear,
	QFinalize,
};
