#include <assert.h>
#include "str.h"
#include "exceptions.h"
#include "str_private.h"

#ifndef require
#error require undefined
#endif

/** @section Strcat Strcat Family */
/** @addtogroup Concatenation Concatenation Functions
@{

All the functions from this Group do the following: They
concatenate elements from @a from to the @a to String. The following
deviations were done:

- It is guarenteed
that @a to always has enough space for
taking all the elements.
- The behavior on an invalid @a to is well defined. If @a to is
invalid from will get copied @emph iff @a from is valid.
-
*/

/** Examples see @example examples/strcat_examples.c
*/

static int Strcat_tplA(StringA &to, char *from,size_t elements_to_cat)
{

	if (!Strvalid(to)){
		to.count = 0;
		to.content = NULL;
		to.capacity = 0;
	}
	size_t needed_space = to.count + elements_to_cat + 1;
	to = str_newA(needed_space,to);
	assert(to.capacity >= needed_space);
	if (to.count)
		memcpy(to.content+to.count, from, elements_to_cat);
	to.count = needed_space-1;
	to.content[needed_space] = 0;
	return elements_to_cat;
}

static int Strcat_tplW(StringW &to, wchar_t *from,size_t elements_to_cat)
{

	if (!StrvalidW(to) || to.count == 0){
		to.count = 0;
		to.content = NULL;
		to.capacity = 0;
	}
	size_t needed_space = to.count + elements_to_cat + 1;
	to = str_newW(needed_space,to);
	assert(to.capacity >= needed_space);
	wcsncat(to.content, from, elements_to_cat);
	to.count = needed_space-1;
	to.content[needed_space] = 0;
	return elements_to_cat;
}


/** @ref Strcat family */
int EXPORT overloaded Strcat(StringA &to, char* from)
{
	StringA s;

	if (from == NULL)
		return 0;
	s.count = strlen(from);
	s.capacity = s.count+1;
	s.content = from;
	return Strncat(to, s, s.count);
}

/** @ref Strcat family */
int EXPORT overloaded Strcat(StringW &to, wchar_t* from)
{
	StringW s;

	if (from == NULL)
		return 0;
	s.count = wcslen(from);
	s.capacity = s.count+1;
	s.content = from;
	return Strncat(to, s, s.count);
}

/** @ref Strcat family */
int EXPORT overloaded Strcat(StringA & to, StringA &from)
{
	int code = 0;
	if (!Strvalid(from))
		return 0;
	return Strncat(to, from, from.count);
}

int EXPORT overloaded Strcat(StringW & to, StringW &from)
{
	int code = 0;
	if (!StrvalidW(from))
		return 0;
	return Strcat_tplW(to, from.content, from.count);
}
StringA EXPORT overloaded Strcatv(char *s,...)
{
	int len = strlen(s);
	va_list ap,save;
	char *next,*p;
	StringA result;

	va_start(ap,s);
	save = ap;
	next = va_arg(ap,char *);
	while (next) {
		len += strlen(next);
		next = va_arg(ap,char *);
	}
	va_start(ap,s);
	result = new_stringA(len);
	strcat(result.content,s);
	p = result.content + strlen(s);
	next = va_arg(ap,char *);
	while (next) {
		while (*next) {
			*p++ = *next++;
		}
		next = va_arg(ap,char *);
	}
	*p = 0;
	result.count = p - result.content;
	return result;
}

StringW EXPORT overloaded Strcatv(wchar_t *s,...)
{
	int len = wcslen(s);
	va_list ap,save;
	wchar_t *next,*p;
	StringW result;

	va_start(ap,s);
	save = ap;
	next = va_arg(ap,wchar_t *);
	while (next) {
		len += wcslen(next);
		next = va_arg(ap,wchar_t *);
	}
	va_start(ap,s);
	result = new_stringW(len+1);
	wcscat(result.content,s);
	p = result.content + wcslen(s);
	next = va_arg(ap,wchar_t *);
	while (next) {
		while (*next) {
			*p++ = *next++;
		}
		next = va_arg(ap,wchar_t *);
	}
	*p = 0;
	result.count = p - result.content;
	return result;
}


/** @ref Strcat family
Parameters and constraints on them see @ref Strncat_fun
*/
int EXPORT overloaded Strncat(StringA &to,char * from,size_t elements_to_cat)
{
	return Strcat_tplA(to, from, elements_to_cat);
}

/** @ref Strcat family
Parameters and constraints on them see @ref Strncat_fun */
int EXPORT overloaded Strncat(StringW &to,wchar_t* from,size_t elements_to_cat)
{
	require(from && wcslen(from) <= elements_to_cat);
	return Strcat_tplW(to, from, elements_to_cat);
}

int EXPORT overloaded Strncat(StringA &to,FILE *file,size_t n)
{
	unsigned cb;
	char *tmp;

	if (file == NULL)
		return 0;
	if (!StrvalidA(to) || n >= (to.capacity-to.count)) {
		cb = 2*to.count+n;
		tmp = allocate_proper_space(cb,1);
		memcpy(tmp,to.content,to.count);
		tmp[to.count] = 0;
		to.content = tmp;
		to.capacity = cb;
	}
	cb = fread(to.content+to.count,1,n,file);
	if (cb>0) {
		to.content[to.count+cb] = 0;
		to.count += cb;
	}
	return cb;
}

/** @ref Strcat family
@anchor Strncat_fun
Catenates @a elements_to_cat from @a from onto @a to.
If @a to is invalid a copy is made from @a from.
@exception GEF_EXCEPTION_PRECONDITON_VIOLATION
If from is either not a valid
String, or 'from' == NULL or if 'elements_to_cat' >
length(from)
@pre Strvalid(from) || from != NULL
@pre length(from) >= elements_to_cat
*/
int EXPORT overloaded Strncat(StringA & to,StringA & from,size_t elements_to_cat)
{
	if (!StrvalidA(from))
		return false;
	if (!StrvalidA(to)){
		to.count = from.count;
		to.capacity = from.capacity;
		to.content = GC_malloc(from.capacity);
		memcpy(to.content,from.content,from.count);
		return true;
	}
	size_t needed_space = elements_to_cat + to.count+1;
	to = str_newA(needed_space, to);
	strncat(to.content, from.content, elements_to_cat);
	to.count = needed_space-1;
	to.content[to.count] = 0;
	return true;
}

int EXPORT overloaded Strncat(StringW & to,StringW & from,size_t elements_to_cat)
{
	if (!StrvalidW(to)){
		if (StrvalidW(from)) {
			to.count = from.count;
			to.capacity = from.capacity;
			to.content = GC_malloc(from.capacity);
			memcpy(to.content,from.content,from.count);
			return true;
		}
		return false;
	}
	require(StrvalidW(from));

	size_t needed_space = elements_to_cat + to.count+1;
	to = str_newW(needed_space, to);
	wcsncat(to.content, from.content, elements_to_cat);
	to.count = needed_space-1;
	to.content[to.count] = 0;
	return true;
}
