#include "str.h"
#include "exceptions.h"
#include "str_private.h"


/** Strcmp result is the same as for the normal strcmp functions
in C.
@anchor Strcmp_fun
@pre valid_strings -> Strvalid(s1) && Strvalid (s2)
@pre valid_strings -> Strvalid(s1) && s2 != NULL (for
s2 beeing of type char* or wchar_t *
	@param s1 First String or char* or wchar_t* to compare.
@param s2 Second String or char* or wchar_t* to compare.
@result Strcmp(s1,s2) < 0 => s1 is less than s2 <br>
Strcmp(s1,s2) == 0 => s1 is equal to s2 <br>
Strcmp(s1,s2) > 0 => s1 is greater than s2

@note  This overloading orgy is annoying and we may try another
extension to C in which a cast to the actual parameter
will get searched on a mismatch. This is done
for operators but does not work for normal functions yet.
*/

/** Examples see @example examples/strcmp_examples.c
*/

/** other examples in @example examples/strcomparison_examples.c
*/
int EXPORT overloaded Strcmp(StringA & s1, StringA & s2)
{
	require(StrvalidA(s1) && StrvalidA(s2));

	if (!StrvalidA(s1) || !StrvalidA(s2))
		return 0;
	/* conversion widening the smaller to the wider string type */
	if (0 == s1.count && s2.count == 0){
		return 0;
	}
	if (0 == s1.count){
		return -1;
	}
	if (0 == s2.count){
		return 1;
	}
	// int len = s1.count < s2.count ? s1.count : s2.count;
	return strcmp(s1.content, s2.content);
}

int EXPORT overloaded Strcmp(StringW & s1, StringW & s2)
{
	require(StrvalidW(s1) && StrvalidW(s2));

	if (0 == s1.count && s2.count == 0){
		return 0;
	}
	if (0 == s1.count){
		return -1;
	}
	if (0 == s2.count){
		return 1;
	}
	// int len = s1.count < s2.count ? s1.count : s2.count;
	// len *= sizeof(wchar_t);
	return wcscmp(s1.content, s2.content);
}

/** Parameters and constraints on them see
@ref Strcmp_fun "Strcmp Function"
*/
int EXPORT overloaded Strcmp(StringA & string, const char* str)
{
	require(StrvalidA(string) && str != NULL);
	size_t len = strlen(str);
	if (0 == string.count && len == 0) return 0;
	if (0 == string.count) return -1;
	if (0 == len) return 1;
	return strcmp(string.content,str);
}

int EXPORT overloaded Strcmp(StringW & string,const wchar_t *str)
{
	require(StrvalidW(string) && str != NULL);
	size_t len = wcslen(str);
	if (0 == string.count && len == 0) return 0;
	if (0 == string.count) return -1;
	if (0 == len) return 1;
	return wcscmp (string.content,str);
}

int EXPORT overloaded Strcmp(StringpA s1,StringpA s2)
{
	require(StrvalidpA(s1) && StrvalidpA(s2));
	return strcmp(s1.content,s2.content);
}

int EXPORT overloaded Strcmp(StringpW s1,StringpW s2)
{
	require(StrvalidpW(s1) && StrvalidpW(s2));
	return wcscmp(s1.content,s2.content);
}

int EXPORT overloaded Strcmp(StringpA s1,char *s2)
{
	require(StrvalidpA(s1) && NULL != s2);
	return strcmp(s1.content,s2);
}

int EXPORT overloaded Strcmp(StringpW s1,wchar_t *s2)
{
	require(StrvalidpW(s1) && NULL != s2);
	return wcscmp(s1.content,s2);
}


/**
	Parameters and constraints on them see
@ref Strcmp_fun "Strcmp Function"
*/
int EXPORT overloaded Strcmp(const char* str, StringA & string)
{
	require(StrvalidA(string) && NULL != str);
	return strcmp(string.content, str);
}


int EXPORT overloaded Strcmp(const wchar_t* str, StringW& string)
{
	require(StrvalidW(string) && NULL != str);
	return wcscmp(string.content,str);
}

/** Strncmp compare some elements from two Strings.
@anchor Strncmp_fun
@param s1 First String to compare.
@param s2 Second String to compare.
@param elements compars this many elements.
@returns see @ref Strcmp_fun
@pre Strvalid(s1) && Strvalid(s2)
@pre elements_in_bounds: elements <= Strlen(s1) && <br>
elements <= Strlen(s2)
@notes It would be good adding comparisons to
char* and wchar_t * too.
*/
int overloaded Strncmp(StringA &s1, StringA &s2, size_t elements)
{
	require(StrvalidA(s1) && StrvalidA(s2));
	elements = (elements > s2.count) ? s2.count : elements;
	elements = (elements > s1.count) ? s1.count : elements;

	return memcmp(s1.content, s2.content, elements);
}

int overloaded Strncmp(StringW &s1, StringW &s2, size_t elements)
{
	require(StrvalidW(s1) && StrvalidW(s2));
	elements = (elements > s2.count) ? s2.count : elements;
	elements = (elements > s1.count) ? s1.count : elements;

	return memcmp(s1.content, s2.content, elements);
}

bool EXPORT Strequalrange(String & s1, String & s2,
	int nelements,
	int s1_from = 0,
	int s2_from = 0){
	/* anders schreiben ? frido 2002-09-20 */
	bool result = true;
	int code = 0;
	require(Strvalid(s1) && Strvalid(s2));
	int count = 0;
	int diff_s1 = 0;
	int diff_s2 = 0;
	if (s1_from + nelements > s1.count ||
			s2_from + nelements > s2.count){
		diff_s1 = s1.count - s1_from;
		diff_s2 = s2.count - s2_from;
		count  = MIN (diff_s1, diff_s2);
	}
	else {
		count = nelements;
	}

	CHECK(count >= 0, exc);
	for (int i = 0; i < count && result; i++){
		if (s1[s1_from + i] != s2[s2_from + i]){
			result = false;
		}
	}
	return result;
}

/** Strcmpi compares Strings case independendly.
@param s1, s2 Strings to be compared
@pre valid Parmaters => Strvalid (s1) && Strvalid(s2)
@result -1 if s1 lexically smaller than s2
0 if s1 == s2 (case insensitive)
1 else
*/
int EXPORT overloaded Stricmp(StringA & s1, StringA & s2)
{
	require(StrvalidA(s1) && StrvalidA(s2));
	return strcmpi(s1.content, s2.content);
}

int EXPORT overloaded Stricmp(StringW & s1, StringW & s2)
{
	require(StrvalidW(s1) && StrvalidW(s2));
	return wcsicmp(s1.content, s2.content);
}

