/*------------------------------------------------------------------------
Module:        strcollection.c
Author:        jacob
Project:       Containers
State:
Creation Date:
Description:   Implements the string collection container.
Added fixes from Gerome. Oct. 20 2005
------------------------------------------------------------------------*/
#include "containers.h"

/* Definition of the String Collection type */
struct StringCollection {
    StringCollectionInterface *VTable; /* The table of functions */
    size_t count;                  /* in element size units */
    unsigned int Flags;             /* Read-only or other flags */
    char **contents;               /* The contents of the collection */
    size_t capacity;                /* in element_size units */
	size_t timestamp;
	/* Error function */
	ErrorFunction RaiseError;
//	CompareFunction Compare;
};

#ifdef UNIX
#define stricmp strcasecmp
#endif
#ifdef _MSC_VER
#define stricmp _stricmp
#else
extern int stricmp(const char *,const char *);
#endif

#ifndef DEFAULT_START_SIZE
#define DEFAULT_START_SIZE 20
#endif

static int NullPtrError(const char *fnName)
{
	char buf[256];
	sprintf(buf,"iStrCollection.%s",fnName);
	iError.RaiseError(buf,CONTAINER_ERROR_BADARG);
	return CONTAINER_ERROR_BADARG;
}

static int CompareStrings(const void *s1,const void *s2)
{
	const char *S1 = *(const char **)s1;
	const char *S2 = *(const char **)s2;
	return strcmp(S1,S2);
}

static int CaseCompareStrings(const void *s1,const void *s2)
{
	const char *S1 = *(const char **)s1;
	const char *S2 = *(const char **)s2;
	return stricmp(S1,S2);
}


/*------------------------------------------------------------------------
Procedure:     DuplicateString ID:1
Purpose:       Functional equivalent of strdup with added argument
               in case of error
Input:         The string to duplicate and the name of the calling function
Output:        The duplicated string
Errors:        Duplication of NULL is allowed and returns NULL. If no more
               memory is available the error function is called.
------------------------------------------------------------------------*/
static char * DuplicateString(StringCollection *SC,char *str,const char *fnname)
{
    char *result;
    if (str == NULL)      return NULL;
    result = MALLOC(SC,strlen(str)+1);
    if (result == NULL) {
        iError.RaiseError(fnname,CONTAINER_ERROR_NOMEMORY);
        return NULL;
    }
    strcpy(result,str);
    return result;
}

#define SC_IGNORECASE   (CONTAINER_READONLY << 1)
#define CHUNKSIZE 20



static size_t GetCount(StringCollection *SC){ return SC->count;}

/*------------------------------------------------------------------------
 Procedure:     IsReadOnly ID:1
 Purpose:       Reads the read-only flag
 Input:         The collection
 Output:        the state of the flag
 Errors:        None
------------------------------------------------------------------------*/
static unsigned GetFlags(StringCollection *SC)
{
	return SC->Flags;
}

/*------------------------------------------------------------------------
 Procedure:     SetReadOnly ID:1
 Purpose:       Sets the value of the read only flag
 Input:         The collection to be changed
 Output:        The old value of the flag
 Errors:        None
------------------------------------------------------------------------*/
static unsigned SetFlags(StringCollection *SC,unsigned newval)
{
    int oldval = SC->Flags;
    SC->Flags = newval;
    return oldval;
}

/*------------------------------------------------------------------------
 Procedure:     Resize ID:1
 Purpose:       Grows a string collection by CHUNKSIZE
 Input:         The collection
 Output:        One if successfull, zero othewise
 Errors:        If no more memory is available, nothing is changed
                and returns zero
------------------------------------------------------------------------*/
static int Resize(StringCollection *SC)
{
    size_t newcapacity = SC->capacity + 1+SC->capacity/4;
    char **oldcontents = SC->contents;
    SC->contents = MALLOC(SC,newcapacity*sizeof(char *));
    if (SC->contents == NULL) {
		SC->RaiseError("StringCollection: Resize",CONTAINER_ERROR_NOMEMORY);
		SC->contents = oldcontents;
		return CONTAINER_ERROR_NOMEMORY;
    }
    memset(SC->contents,0,sizeof(char *)*newcapacity);
    memcpy(SC->contents,oldcontents,SC->count*sizeof(char *));
    SC->capacity = newcapacity;
    FREE(SC,oldcontents);
	SC->timestamp++;
    return 1;
}

/*------------------------------------------------------------------------
 Procedure:     Add ID:1
 Purpose:       Adds a string to the string collection.
 Input:         The collection and the string to be added to it
 Output:        Number of items in the collection or <= 0 if error
 Errors:
------------------------------------------------------------------------*/
static int Add(StringCollection *SC,char *newval)
{
	if (SC == NULL) {
		return NullPtrError("Add");
	}

	if (SC->Flags & CONTAINER_READONLY)
        return CONTAINER_ERROR_READONLY;
    if (SC->count >= SC->capacity) {
		int r = Resize(SC);
        if (r <= 0)
            return r;
    }

    if (newval) {
        SC->contents[SC->count] = DuplicateString(SC,newval,"Add");
        if (SC->contents[SC->count] == NULL) {
            return 0;
        }
    }
    else
        SC->contents[SC->count] = NULL;
	SC->timestamp++;
    ++SC->count;
	return 1;
}

static int AddRange(StringCollection *SC,char **data)
{
    int i = 0;

	if (SC == NULL) {
		return NullPtrError("Add");
	}
    if (SC->Flags & CONTAINER_READONLY)
        return CONTAINER_ERROR_READONLY;
	if (data[0] != NULL)
		SC->timestamp++;
    while (data[i] != NULL) {
        int r = Add(SC,data[i]);
        if (r <= 0) {
            return r;
        }
        i++;
    }
    return i;
}

static int Clear(StringCollection *SC)
{
	size_t i;

	if (SC == NULL) {
		return NullPtrError("Add");
	}
    if (SC->Flags & CONTAINER_READONLY)
        return CONTAINER_ERROR_READONLY;
    for (i=0; i<SC->count;i++) {
        FREE(SC,SC->contents[i]);
        SC->contents[i] = NULL;
    }
    SC->count = 0;
    return 1;
}

static int Contains(StringCollection *SC,char *str)
{
    int c;
	size_t i;

	if (SC == NULL) {
		return NullPtrError("Add");
	}
	if (str == NULL)
		return 0;
    c = *str;
    if ((SC->Flags & SC_IGNORECASE) == 0) {
        for (i=0; i<SC->count;i++) {
            if (c == SC->contents[i][0] && !strcmp(SC->contents[i],str))
				return 1;
        }
    }
    else {
        for (i = 0; i<SC->count;i++) {
            if (!stricmp(SC->contents[i],str))
	            return 1;
        }
    }
    return 0;
}

static char **CopyTo(StringCollection *SC)
{
	char **result;
	size_t i;

	if (SC == NULL) {
		NullPtrError("Add");
		return NULL;
	}

    result = MALLOC(SC,(1+SC->count)*sizeof(char *));
    if (result == NULL) {
		iError.RaiseError("iStringCollection.CopyTo",CONTAINER_ERROR_NOMEMORY);
        return NULL;
	}
    for (i=0; i<SC->count;i++) {
        result[i] = DuplicateString(SC,SC->contents[i],"CopyTo");
    }
    result[i] = NULL;
    return result;
}

static size_t IndexOf(StringCollection *SC,char *str)
{
    size_t i;
    int (*cmpfn)(const char *,const char *);

	if (SC == NULL || str == NULL) {
		NullPtrError("Add");
		return 0;
	}
    if (SC->Flags & SC_IGNORECASE) {
        cmpfn = stricmp;
    }
    else cmpfn = strcmp;

    for (i=0; i<SC->count;i++) {
        if (!cmpfn(SC->contents[i],str)) {
            return i+1;
        }
    }
    return 0;
}

static char *IndexAt(StringCollection *SC,size_t idx)
{
    if (idx >=SC->count)
        return NULL;
    return SC->contents[idx];
}

static int InsertAt(StringCollection *SC,size_t idx,char *newval)
{
    char *p;
    if (SC->Flags & CONTAINER_READONLY)
		return CONTAINER_ERROR_READONLY;
    if (SC->count >= (SC->capacity-1)) {
        int r = Resize(SC);
		if (r <= 0)
            return r;
    }
    if ((int)idx < 0)
		return CONTAINER_ERROR_BADARG;
    if (idx >= SC->count)
        return CONTAINER_ERROR_INDEX;
    p = DuplicateString(SC,newval,"InsertAt");
    if (p == NULL) {
        return CONTAINER_ERROR_NOMEMORY;
    }

    if (idx == 0) {
        if (SC->count > 0)
	        memmove(SC->contents+1,SC->contents,SC->count*sizeof(char *));
        SC->contents[0] = p;
    }
    else if (idx == SC->count) {
        SC->contents[idx] = p;
    }
    else if (idx < SC->count) {
        memmove(SC->contents+idx+1,SC->contents+idx,(SC->count-idx+1)*sizeof(char *));
        SC->contents[idx] = p;
    }
	SC->timestamp++;
    ++SC->count;
	return 1;
}

static int Insert(StringCollection *SC,char *newval)
{
    return InsertAt(SC,0,newval);
}

static int RemoveAt(StringCollection *SC,size_t idx)
{
    if (idx >= SC->count )
		return CONTAINER_ERROR_INDEX;
	if (SC->Flags & CONTAINER_READONLY)
        return CONTAINER_ERROR_READONLY;
	/* Test for remove of an empty collection */
    if (SC->count == 0)
        return 0;
    FREE(SC,SC->contents[idx]);
    if (idx < (SC->count-1)) {
        memmove(SC->contents+idx,SC->contents+idx+1,(SC->count-idx)*sizeof(char *));
    }
    SC->contents[SC->count-1]=NULL;
	SC->timestamp++;
    --SC->count;
	return 1;
}

static int Remove(StringCollection *SC,char *str)
{
    size_t idx = IndexOf(SC,str);
    if (idx <= 0)
        return CONTAINER_ERROR_NOTFOUND;
    return RemoveAt(SC,idx-1);
}

static int PushBack(StringCollection *SC,char *str)
{
    char *r;

    if (SC->Flags&CONTAINER_READONLY)
        return 0;
    if (SC->count >= SC->capacity-1) {
        int res = Resize(SC);
		if (res <= 0)
            return res;
    }
    r = DuplicateString(SC,str,"Push");
    if (r == NULL)
        return 0;
    SC->contents[SC->count++] = r;
	SC->timestamp++;
	return 1;
}

static size_t PopBack(StringCollection *SC,char *buffer,size_t buflen)
{
    char *result;
	size_t len,tocopy;

	if (SC == NULL)
		return 0;
    if ((SC->Flags&CONTAINER_READONLY) || SC->count == 0)
        return 0;
	len = 1+strlen(SC->contents[SC->count-1]);
	if (buffer == NULL)
		return len;
    SC->count--;
    result = SC->contents[SC->count];
    SC->contents[SC->count] = NULL;
	SC->timestamp++;
	tocopy = len;
	if (buflen < tocopy)
		tocopy = buflen-1;
	memcpy(buffer,result,tocopy);
	buffer[tocopy-1] = 0;
	FREE(SC,result);
    return len;
}

static int Finalize(StringCollection *SC)
{
	size_t i;
    for (i=0; i<SC->count;i++) {
        FREE(SC,SC->contents[i]);
    }
    FREE(SC,SC->contents);
    CurrentMemoryManager->free(SC);
    return 1;
}

static size_t GetCapacity(StringCollection *SC)
{
    return SC->capacity;
}

static bool SetCapacity(StringCollection *SC,int newCapacity)
{
    if (SC->count != 0)
        return 0;
    FREE(SC,SC->contents);
    SC->contents = MALLOC(SC,newCapacity*sizeof(char *));
    memset(SC->contents,0,sizeof(char *)*newCapacity);
    SC->capacity = newCapacity;
    return 1;
}

static void Apply(StringCollection *SC,int (*Applyfn)(char *,void *),void *arg)
{
    size_t i;
    for (i=0; i<SC->count;i++) {
        Applyfn(SC->contents[i],arg);
    }
}

static int *Map(StringCollection *SC,int (*Applyfn)(char *)){
    int *result = MALLOC(SC,SC->count*sizeof(int));
	size_t i;

    if (result == NULL)
      return NULL;
    for (i=0; i<SC->count;i++) {
        result[i] = Applyfn(SC->contents[i]);
    }
    return result;
}

/* gedd123@free.fr (gerome) proposed calling DuplicateString. Good
suggestion */
static char *ReplaceAt(StringCollection *SC,size_t idx,char *newval)
{
    char *r;
    if (SC->Flags & CONTAINER_READONLY)      return NULL;
    if (idx > SC->count)      return NULL;
    FREE(SC,SC->contents[idx]);
    r = DuplicateString(SC,newval,"ReplaceAt");
    if (r == NULL)	return 0;
    SC->contents[idx] = r;
	SC->timestamp++;
    return newval;
}

static bool Equal(StringCollection *SC1,StringCollection *SC2)
{
    size_t i;
    if (SC1->count != SC2->count)
		return 0;
	for (i=0; i<SC1->count;i++) {
		if (strcmp(SC1->contents[i],SC2->contents[i]))
			return 0;
	}
    return 1;
}

static StringCollection *Copy(StringCollection *SC){
    size_t i;
    StringCollection *result = iStringCollection.Create(SC->count);

	if (result) {
		result->VTable = SC->VTable;
		for (i=0; i<SC->count;i++) {
			if (SC->VTable->Add(result,SC->contents[i]) <= 0) {
				Finalize(result);
				result = NULL;
				break;
			}
		}
    }
    return result;
}


static ErrorFunction SetErrorFunction(StringCollection *SC,ErrorFunction fn)
{
	ErrorFunction old;
	old = SC->RaiseError;
	SC->RaiseError = (fn) ? fn : iError.EmptyErrorFunction;
	return old;
}

static bool Sort(StringCollection *SC)
{
	if (SC->Flags & SC_IGNORECASE)
		qsort(SC->contents,SC->count,sizeof(char *),CaseCompareStrings);
	else
		qsort(SC->contents,SC->count,sizeof(char *),CompareStrings);
	SC->timestamp++;
	return 1;
}

static size_t Sizeof(StringCollection *SC)
{
	size_t result= sizeof(StringCollection);
	size_t i;

	if (SC == NULL)
		return 0;
	for (i=0; i<SC->count;i++) {
		result += strlen(SC->contents[i]) + 1 + sizeof(char *);
	}
	result += (SC->capacity - SC->count) * sizeof(char *);
	return result;
}

/* ------------------------------------------------------------------------------ */
/*                                Iterators                                       */
/* ------------------------------------------------------------------------------ */

struct StringCollectionIterator {
	Iterator it;
	StringCollection *SC;
	size_t index;
	size_t timestamp;
	unsigned long Flags;
};

static void *GetNext(Iterator *it)
{
	struct StringCollectionIterator *sci = (struct StringCollectionIterator *)it;
	StringCollection *SC = sci->SC;

	if (sci->timestamp != SC->timestamp) {
		SC->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	if (sci->index >= SC->count)
		return NULL;
	return IndexAt(SC,sci->index++);
}

static void *GetPrevious(Iterator *it)
{
	struct StringCollectionIterator *ali = (struct StringCollectionIterator *)it;
	StringCollection *SC = ali->SC;

	if (ali->index >= SC->count || ali->index == 0)
		return NULL;
	if (ali->timestamp != SC->timestamp) {
		SC->RaiseError("GetNext",CONTAINER_ERROR_OBJECT_CHANGED);
		return NULL;
	}
	return IndexAt(SC,ali->index--);
}

static void *GetFirst(Iterator *it)
{
	struct StringCollectionIterator *ali = (struct StringCollectionIterator *)it;
	if (ali->SC->count == 0)
		return NULL;
	ali->index = 1;
	return ali->SC->contents;
}

static Iterator *newIterator(StringCollection *SC)
{
	struct StringCollectionIterator *result = MALLOC(SC,sizeof(struct StringCollectionIterator));
	if (result == NULL)
		return NULL;
	result->it.GetNext = GetNext;
	result->it.GetPrevious = GetPrevious;
	result->it.GetFirst = GetFirst;
	result->SC = SC;
	result->timestamp = SC->timestamp;
	return &result->it;
}

static void deleteIterator(Iterator *it)
{
	struct StringCollectionIterator *sci = (struct StringCollectionIterator *)it;
	StringCollection *SC = sci->SC;
	FREE(SC,it);
}

static int SaveHeader(StringCollection *SC,FILE *stream)
{
	return (int)fwrite(SC,1,sizeof(StringCollection),stream);
}

static size_t DefaultSaveFunction(const void *element,void *arg, FILE *Outfile)
{
	const unsigned char *str = element;
	size_t len = strlen(element);

	if (encode_ule128(Outfile, len) <= 0)
		return (size_t)EOF;
	return fwrite(str,1,len+1,Outfile);
}

static size_t DefaultLoadFunction(void *element,void *arg, FILE *Infile)
{
	size_t len = *(size_t *)arg;

	return fread(element,1,len,Infile);
}

static int Save(StringCollection *SC,FILE *stream, SaveFunction saveFn,void *arg)
{
	size_t i;

	if (SC == NULL || stream == NULL) {
		NullPtrError("Save");
		return CONTAINER_ERROR_BADARG;
	}
	if (saveFn == NULL)
		saveFn = DefaultSaveFunction;
	if (SaveHeader(SC,stream) <= 0)
		return EOF;
	for (i=0; i< SC->count; i++) {
		if (saveFn(SC->contents[i],arg,stream) <= 0)
			return EOF;
	}
	return 0;
}

static StringCollection *Load(FILE *stream, ReadFunction readFn,void *arg)
{
	size_t i,len;
	StringCollection *result,SC;

	if (readFn == NULL) {
		readFn = DefaultLoadFunction;
		arg = &len;
	}
	if (fread(&SC,1,sizeof(StringCollection),stream) <= 0) {
		iError.RaiseError("StringCollection.Load",CONTAINER_ERROR_FILE_READ);
		return NULL;
	}
	result = iStringCollection.Create(SC.count);
	result->Flags = SC.Flags;
	for (i=0; i< SC.count; i++) {
		if (decode_ule128(stream, &len) <= 0) {
			goto err;
		}
		len++;
		result->contents[i] = MALLOC(result,len);
		if (readFn(result->contents[i],arg,stream) <= 0) {
		err:
			iError.RaiseError("StringCollection.Load",CONTAINER_ERROR_FILE_READ);
			break;
		}
		result->count++;
	}
	return result;
}
static ArrayList *CastToArray(StringCollection *SC)
{
	ArrayList *AL = iArrayList.Create(sizeof(void *),SC->count);
	size_t i;

	for (i=0; i<SC->count;i++) {
		iArrayList.Add(AL,SC->contents[i]);
	}
	return AL;
}

static StringCollection *CreateFromFile(unsigned char *fileName)
{
	StringCollection *result;
	char *line=NULL;
	int llen=0,r;
	ContainerMemoryManager *mm = CurrentMemoryManager;
	FILE *f = fopen((char *)fileName,"r");
	if (f == NULL) {
		iError.RaiseError("StringCollection.CreateFromFile",CONTAINER_ERROR_NOENT);
		return NULL;
	}
	result = iStringCollection.Create(10);
	if (result == NULL)
		return NULL;
	r = getline(&line,&llen,f,mm);
	while (r >= 0) {
		if (iStringCollection.Add(result,line) <= 0)
			break;
		r = getline(&line,&llen,f,mm);
	}
	free(line);
	fclose(f);
	return result;
}

static int WriteToFile(StringCollection *SC,unsigned char *fileName)
{
	FILE *f = fopen((char *)fileName,"w");
	size_t i;

	if (SC == NULL || fileName == NULL) {
		NullPtrError("WriteToFile");
		return CONTAINER_ERROR_BADARG;
	}
	if (f == NULL) {
		iError.RaiseError("StringCollection.WriteToFile",CONTAINER_ERROR_FILEOPEN);
		return CONTAINER_ERROR_FILEOPEN;
	}
	for (i=0; i<SC->count;i++) {
		if (fwrite(SC->contents[i],1,strlen(SC->contents[i]),f) <= 0) {
			iError.RaiseError("StringCollection.WriteToFile",CONTAINER_ERROR_FILE_WRITE);
			fclose(f);
			return CONTAINER_ERROR_FILE_WRITE;
		}
	}
	fclose(f);
	return SC->count ? 1:0;
}

static size_t FindFirstText(StringCollection *SC,char *text)
{
	size_t i;

	if (SC == NULL || text == NULL)
		return 0;
	for (i=0; i<SC->count;i++) {
		if (strstr(SC->contents[i],text)) {
			return i+1;
		}
	}
	return 0;
}

static size_t FindNextText(StringCollection *SC, char *text, size_t start)
{
	size_t i;

	if (SC == NULL || text == NULL)
		return 0;
	for (i=start; i<SC->count;i++) {
		if (strstr(SC->contents[i],text)) {
			return i+1;
		}
	}
	return 0;
}

static ArrayList *FindTextPositions(StringCollection *SC,char *text)
{
	ArrayList *result = NULL;
	size_t i;

	for (i=0; i<SC->count;i++) {
		if (strstr(SC->contents[i],text)) {
			if (result == NULL) {
				result = iArrayList.Create(sizeof(size_t),10);
				if (result == NULL)
					return NULL;
			}
			if (iArrayList.Add(result,&i) <= 0)
				break;
		}
	}
	return result;
}


/*------------------------------------------------------------------------
 Procedure:     Create ID:1
 Purpose:       Creates a new string collection
 Input:         The initial start size of the collection
 Output:        A pointer to the new collection or NULL
 Errors:        If no more memory is available it returns NULL
 after calling the error function.
 ------------------------------------------------------------------------*/
static StringCollection  *Create(size_t startsize)
{
    StringCollection *result;

    result = CurrentMemoryManager->malloc(sizeof(*result));
    if (result == NULL) {
        iError.RaiseError("newStringCollection",CONTAINER_ERROR_NOMEMORY);
        return NULL;
    }
    memset(result,0,sizeof(StringCollection));
    result->VTable = &iStringCollection;
    if (startsize > 0) {
        result->contents = CurrentMemoryManager->malloc(startsize*sizeof(char *));
        if (result->contents == NULL) {
            iError.RaiseError("newStringCollection",CONTAINER_ERROR_NOMEMORY);
            CurrentMemoryManager->free(result);
            result=NULL;
        }
        else {
            memset(result->contents,0,sizeof(char *)*startsize);
            result->capacity = startsize;
			result->RaiseError = iError.RaiseError;
        }
    }
    return result;
}
StringCollectionInterface iStringCollection = {
	Create,
	CreateFromFile,
    GetCount, GetFlags, SetFlags, Add, AddRange, Clear, Contains,
    CopyTo,
	IndexOf, Insert,
	InsertAt,
	IndexAt,
	Remove, RemoveAt,
    Finalize, GetCapacity, SetCapacity,
	Apply, Map,
	PushBack,
	PopBack,
	ReplaceAt,
    Equal,Copy,
	SetErrorFunction,
	Sort,
	Sizeof,
	NULL,
	newIterator,
	deleteIterator,
	Save,
	Load,
	CastToArray,
	FindFirstText,
	FindNextText,
	FindTextPositions,
	WriteToFile,
};
