#include <string.h>
#include "containererror.h"
// Forward declaration of the string collection type
typedef struct _StringCollection StringCollection;
struct _StringEnumerator;

// Definition of the functions associated with this type.
typedef struct {
	// Returns the number of elements stored
	int (*GetCount)(StringCollection &SC);

	// Is this collection read only?
	int (*IsReadOnly)(StringCollection &SC);

	// Sets this collection read-only or unsets the read-only flag
	int (*SetReadOnly)(StringCollection &SC,int flag);

	// Adds one element at the end. Given string is copied
	int (*Add)(StringCollection &SC,char *newval);

	// Adds a NULL terminated table of strings
	int (*AddRange)(StringCollection &SC,char **newvalues);

	// Clears all data and frees the memory
	int (*Clear)(StringCollection &SC);

	//Case sensitive search of a character string in the data
	int (*Contains)(StringCollection &SC,char *str);

	// Copies all strings into a NULL terminated vector
	char **(*CopyTo)(StringCollection &SC);

	//Returns the index of the given string or -1 if not found
	int (*IndexOf)(StringCollection &SC,char *SearchedString);

	// Inserts a string at the position zero.
	int (*Insert)(StringCollection &SC,char *);

	// Inserts a string at the given position
	int (*InsertAt)(StringCollection &SC,int idx,char *newval);

	// Returns the string at the given position
	char *(*IndexAt)(StringCollection &SC,int idx);

	// Removes the given string if found
	int (*Remove)(StringCollection &SC,char *);

	//Removes the string at the indicated position
	int (*RemoveAt)(StringCollection &SC,int idx);

	// Frees the memory used by the collection
	int (*Finalize)(StringCollection &SC);

	// Returns the current capacity of the collection
	int (*GetCapacity)(StringCollection &SC);

	// Sets the capacity if there are no items in the collection
	int (*SetCapacity)(StringCollection &SC,int newCapacity);

	// Calls the given function for all strings. "Arg" is a used supplied argument
	// that can be NULL that is passed to the function to call
	void (*Apply)(StringCollection &SC,int (*Applyfn)(char *,void * arg),void *arg);

	// Calls the given function for each string and saves all results
	// in an integer vector
	int *(*Map)(StringCollection &SC,int (*Applyfn)(char *));

	// Pushes a string, using the collection as a stack
	int (*Push)(StringCollection &SC,char *str);

	// Pops the last string off the collection
	char * (*Pop)(StringCollection &SC);

	// Replaces the character string at the given position with a new one
	char *(*ReplaceAt)(StringCollection &SC,int idx,char *newval);

	// Returns whether the collection makes case sensitive comparisons or not
	int (*IsCaseSensitive)(StringCollection &SC);

	// Sets case sensitivity by comparisons
	int (*SetCaseSensitive)(StringCollection &SC,int newval);

	// Compares two string collections
	int (*Equal)(StringCollection &SC1,StringCollection &SC2);

	// Copy a string collection
	StringCollection (*Copy)(StringCollection &SC);
} StringCollectionInterface;

// Definition of the String Collection type
struct _StringCollection {
	StringCollectionInterface *lpVtbl; // The table of functions
        size_t count;                  /* in element size units */
        char **contents;               /* The contents of the collection */
        size_t capacity;                /* in element_size units */
	unsigned int flags;             // Read-only or other flags
};

struct _StringEnumerator;
typedef struct _StringEnumeratorFunctions {
	int (*Next)(struct _StringEnumerator &);
	int (*Previous)(struct _StringEnumerator &);
	char *(*GetCurrent)(struct _StringEnumerator &);
	int (*SetCurrent)(struct _StringEnumerator &,int);
} StringEnumeratorFunctions;

typedef struct _StringEnumerator {
	StringEnumeratorFunctions *lpVtbl;
	StringCollection *SC;
	int CurrentIndex;
} StringEnumerator;


#ifndef NO_GC
#include <gc.h>
#define MALLOC GC_malloc
#define FREE(a)
#else
#include <stdlib.h>
#define MALLOC malloc
#define FREE free
#endif

#ifndef DEFAULT_START_SIZE
#define DEFAULT_START_SIZE 20
#endif

// This are the only exported functions from this module
StringCollection newStringCollection(int startsize=0);
char *operator[](StringCollection *,int);
char *operator[](StringCollection SC,int idx);
char *operator[]=(StringCollection *SC,int idx,char *newval);
char *operator[]=(StringCollection SC,int idx,char *newval);
