/* -------------------------------------------------------------------
   Lcc-win32 runtime library
   -----------------

   Vector routines
   ---------------
   This routines handle the vector class. This is a very general implementation
   and efficiency considerations aren't yet primordial. Vectors can have elements
   of any size, and are stretchable.
----------------------------------------------------------------------
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <gc.h>
#include "str.h"
#include "vector.h"
#include "str_private.h"
#include "list.h"
/* local routines to this module */
static bool stretch_vector(vector &v,size_t position);

/*
 Definition of the opaque structure vector. Note that each vector is a *single*
 memory block. Any resizing operations force the reallocation and copy of the
 whole structure. The advantage of this design however, is that we spare us
 an indirection and 4 bytes per extra pointer to contents. Normally vectors are of
 fixed size, and inserting/deleting operations are not common.
 We could change to another allocation strategy just by changing the contents[1]
 declaration below to char *contents.
*/
struct _vector {
        size_t count;                  /* in element size units */
        unsigned char *contents;      /* The contents of the vector follow */
        size_t capacity;                /* in element_size units */
        size_t element_size;            /* in bytes */
};


#define MAX_TMP_SPACE 1024

/* ------------------------------------------------------------------
Allocation (make calls this)
---------------------------------------------------------------------
*/
vector *new_vector(size_t size,size_t element_size,void *contents=0,size_t capacity=0)
{
	vector *result;

	if (size == 0)
		size = DEFAULT_VECTOR_SIZE;
	if (capacity < size)
		capacity = size +1+size/5;
	result = allocate_proper_space(sizeof(vector),1);
	if (result == NULL)
		return NULL;
	result->contents = allocate_proper_space(element_size,capacity);
	if (result->contents == NULL) {
		return NULL;
	}
	result->capacity = capacity;
	result->count = size;
	result->element_size = element_size;
	size *= element_size;
	if (contents)
		memcpy(result->contents,contents,size);
	return result;
}

vector *copy_vector(vector &v)
{
	return new_vector(v.count,v.element_size,v.contents,v.capacity);
}

void dispose_vector(vector *v)
{
}
/*-------------------------------------------------------------------------
  GET routines to read the properties of the vector. Note that not all fields
  are accessible from the outside.
---------------------------------------------------------------------------*/
size_t get_size_vector(vector *v)
{
	return (v) ? v->count : 0;
}

size_t get_capacity_vector(vector *v)
{
	return v->capacity;
}

size_t get_element_size_vector(vector *v)
{
	return (v) ? v->element_size : 0;
}

/*
Get an element. This will be reduced to one machine instruction (BOUNDS)
in later versions.
*/
void * __declspec(naked) get_element_vector(vector *v,size_t position)
{
}
void * operator[](vector *v,size_t position)
{
	return (position < v->count) ?
		v->contents + v->element_size *position : NULL;
}

void *get_contents_vector(vector *v)
{
	return (v) ? v->contents : NULL;
}

/*-------------------------------------------------------------------------
  SET routines to read the properties of the vector. Note that the client code
  can only set an element or set the allocator pair. No other manipulations
  are defined, but maybe a reserve operation could be added. I think this is
  not necessary because we can reserve at creation time.
---------------------------------------------------------------------------*/
void * __declspec(naked) set_element_vector(vector *v,size_t position,void *data)
{
}
void *operator[]=(vector *v,size_t position,void *data)
{
	if (position < v->count) {
		if (v->element_size == sizeof(void *)) {
			((int *)v->contents)[position] = *((int *)data);
		}
		else if (v->element_size == 1)
			v->contents[position] = *(char *)data;
		else
		memcpy(v->contents + v->element_size * position,data,v->element_size);
		return data;
	}
	return NULL;
}

bool reverse_vector(vector &v)
{
	int i,stop;
	char *pdestination,*psource;
	char tmpbuf[v.element_size];

	pdestination = v.contents;
	psource = v.contents+(v.count-1)*v.element_size;
	stop = v.count/2;
	for (i=0; i <stop;i++) {
		memcpy(tmpbuf,pdestination,v.element_size);
		memcpy(pdestination,psource,v.element_size);
		memcpy(psource,tmpbuf,v.element_size);
		pdestination += v.element_size;
		psource -= v.element_size;
	}
	return true;
}
/*-------------------------------------------------------------------------
  METHODS.
---------------------------------------------------------------------------*/
/* Insert into the given vector, an element before the given position */
bool insert_vector(vector *v,size_t position,void *element)
{
	char *pcontents,*pend;
	int tocopy;

	if (v == NULL || position >= v->count)
		return false;
	else if (v->count+1 >= v->capacity) {
		if (!stretch_vector(*v,v->count+1))
			return false;
	}
	pcontents = get_element_vector(v,position);
	pend = v->contents + v->count*v->element_size;
	tocopy = pend - pcontents;
	memmove(pcontents+v->element_size,pcontents,tocopy);
	memcpy(pcontents,element,v->element_size);
	v->count++;
	return true;
}
/* local routine to resize a vector */
static bool stretch_vector(vector &v,size_t newsize)
{
	int newcapacity = calculate_new_capacity(v.capacity,newsize);
	char *data = allocate_proper_space(newcapacity,v.element_size);
	if (data == NULL)
		return false;
	v.capacity = newcapacity;
	memcpy(data,v.contents,v.element_size*v.count);
	v.contents = data;
	return true;
}
/* erase an element from a vector. No resizing occurs */
bool remove_vector(vector *v,size_t position)
{
	char *pcontents;

	if (v == NULL || position >= v->count)
		return false;
	pcontents = get_element_vector(v,position);
	if (pcontents == NULL)
		return false;
	memmove(pcontents,pcontents+v->element_size,v->element_size);
	v->count--;
	return true;
}

size_t clear_vector(vector *v)
{
	size_t oldsize;
	if (v == NULL)
		return 0;
	oldsize = v->count;
	memset(v->contents,0,v->count*v->element_size);
	v->count = 0;
	return oldsize;
}

void *pop_back_vector(vector *v)
{
	if (v == NULL)
		return NULL;
	if (v->count == 0)
		return NULL;
	v->count--;
	return (void *)(v->contents + v->count * v->element_size);

}
int __declspec(naked) equal_vector(vector *v1,vector *v2)
{
}
int operator==(vector v1,vector v2)
{
	if (v1.count != v2.count)
		return 0;
	if (v1.element_size != v2.element_size)
		return 0;
	if (memcmp(v1.contents,v2.contents,v1.count*v1.element_size))
		return 0;
	return 1;
}
/* Builds a new vector of the same shape with the results of "fn" applied
   to each element
*/
vector *map_vector(vector *v,unary_operator fn)
{
	vector *result;
	int i;
	char *pcontents,*pdata,*pdestination;
	char *data;
	char tmpspace[MAX_TMP_SPACE];

	if (v == NULL) {
		return NULL;
	}
	result = new_vector(v->count,v->element_size);
	if (result == NULL) {
		return NULL;
	}
	pcontents = v->contents;
	pdestination = result->contents;
	if (v->element_size*v->count < MAX_TMP_SPACE)
		data = tmpspace;
	else
		data = GC_malloc(v->element_size*v->count);
	for (i=0; i<v->count;i++) {
		pdata = (char *)(fn)(pcontents,data);
		if (pdata)
			memcpy(pdestination,pdata,v->element_size);
		pcontents += v->element_size;
		pdestination += v->element_size;
	}
	return result;
}
#if 0
/* Builds a new vector of the same shape with the results of the binary
   function "fn" applied to each element of both vectors
*/
vector *binary_map_vector(vector *v1,vector *v2,binary_operator fn)
{
	int top,i;
	char *contents1,*contents2,*pdata,*destination;
	vector *result;
	size_t element_size;
	char *data;
	char tmpspace[MAX_TMP_SPACE];

	if (v2 == NULL || v1 == NULL)
		return NULL;
	if (v1->element_size != v2->element_size)
		return NULL;
	element_size = v1->element_size;
	if (v2->count != v1->count) {
		top = v2->count > v1->count ? v1->count : v2->count;
	}
	else top = v1->count;
	contents1 = v1->contents;
	contents2 = v2->contents;
	// Avoid an allocation if possible
	if (element_size < MAX_TMP_SPACE)
		data = tmpspace;
	else
		data = GC_malloc(element_size);
	result = new_vector(top,element_size);
	destination = result->contents;
	for (i=0; i<top;i++) {
		pdata = (char *)(*fn)(contents1,contents2,data);
		if (pdata == NULL)
			break;
		memcpy(destination,pdata,element_size);
		destination += element_size;
		contents1 += element_size;
		contents2 += element_size;
	}
	return result;
}
#endif
/* Iterates a function until the end of the vector is reached or this function
   returns 0.
   Returns a pointer to the last element analyzed or NULL if the end is reached
*/
void *iterate_vector(vector *v,int (*fn)(void *))
{
	int i;
	char *pcontents;

	if (v == NULL)
		return NULL;
	pcontents = v->contents;
	for (i=0; i<v->count;i++) {
		if ((fn)(pcontents) == 0)
			return pcontents;
		pcontents += v->element_size;
	}
	return NULL;
}

void *scan_vector(vector *v,unary_operator fn,void *pstart)
{
	int i;
	char *pcontents;

	if (v == NULL)
		return NULL;
	pcontents = v->contents;
	for (i=0; i<v->count;i++) {
		if (!(fn)(pcontents,pstart))
			return pcontents;
		pcontents += v->element_size;
	}
	return NULL;
}

bool concat_vector(vector &v1,vector &v2)
{
	vector * result;
	char *pcontents;
	size_t size;

	if (v1.element_size != v2.element_size)
		return false;
	if (!stretch_vector(v1,v1.count+v2.count))
		return false;
	size = v1.count*v1.element_size;
	pcontents = v1.contents+size;
	memcpy(pcontents,v2.contents,v2.count*v2.element_size);
	return true;
}

void sort_vector(vector *v,int (*compar)(const void *,const void *))
{
	qsort(v->contents,v->count,v->element_size,compar);
}


vector *nsort_vector(vector *v,int (*compar)(const void *,const void *))
{
	qsort(v->contents,v->count,v->element_size,compar);
	return v;
}

vector *get_subrange_vector(vector *v,size_t start,size_t end)
{
	char *pcontents;
	vector *result;

	pcontents = get_element_vector(v,start);
	if (pcontents == NULL)  {
		return NULL;
	}
	return new_vector(end-start,v->element_size,pcontents,end-start);
}

vector *cast_vector(vector *v1,vector *v2,unary_operator fn)
{
	int top,i;
	char *data,*contents1,*contents2,*pdata;
	char tmpspace[MAX_TMP_SPACE];

	if (v1 == NULL || v2 == NULL)
		return NULL;

	if (v2->count != v1->count) {
		top = v2->count > v1->count ? v1->count : v2->count;
	}
	else top = v1->count;
	contents1 = v1->contents;
	contents2 = v2->contents;
	// Avoid an allocation if possible
	if (v1->element_size < MAX_TMP_SPACE)
		data = tmpspace;
	else
		data = GC_malloc(v1->element_size);
	for (i=0; i<top;i++) {
		pdata = (*fn)(contents2,data);
		if (pdata == NULL)
			break;
		memcpy(contents1,pdata,v1->element_size);
		contents1 += v1->element_size;
		contents2 += v2->element_size;
	}
	return v1;
}
#if 0
vector * overloaded ToVector(list *l)
{
	vector *result = new_vector(l->count,l->ElementSize);

	if (l->count == 0)
		return result;

}
#endif
#ifdef TEST
#include <math.h>
#include <time.h>
int printInt(void *pint)
{
	return printf(" %d",*(int *)pint);
}
void print_vector(vector *v)
{
	iterate_vector(v,printInt);
	printf("\n");
}

int printDouble(void *pdouble)
{
	return printf(" %g",*(double *)pdouble);
}

int comparInts(const void *p1,const void *p2)
{
	return (*(int *)p1) - (*(int *)p2);
}
void *addInts(void *pi1,void *pi2,void *presult)
{
	*(int *)presult = (*(int *)pi1) + (*(int *)pi2);
	return presult;
}
void *intSqrt(void *pint,void *presult)
{
	int *pi = (int *)pint;
	double *pd = (double *)presult;

	*pd = sqrt((double)*pi);
	return presult;
}

void *accumInts(void *pi1,void *presult)
{
	int s;
	s = *(int *)pi1;
	s += *(int *)presult;
	*((int *)presult) = s;
	return presult;
}
void timeit(int times,int size)
{
	vector *v = new_default_int_vector(size);
	long t1,t2;
	int i,j;

	time(&t1);
	printf("Start %s\n",ctime(&t1));

	for (j=0;j<times;j++)
		for (i=0; i<size;i++) {
			set_element_int_vector(v,i,i);
		}
	time(&t2);
	printf("End %s\n",ctime(&t2));
	dispose_vector(v);
}
int main(void)
{
	vector *v,*s,*a,*ac,*dv;
	int i,siz;

	timeit(4000,20000);
	v = new_default_vector(20,sizeof(int));
	siz = get_size_vector(v);
	for (i=0; i<siz;i++) {
		set_element_vector(v,i,&i);
	}
	printf("Initial vector:\n");
	print_vector(v);

	i = 5987;
	v = insert_vector(v,3,&i);
	printf("after insertion position 3\n");
	print_vector(v);

	i = -5987;
	v = insert_vector(v,3,&i);
	printf("after insertion position 3\n");
	print_vector(v);

	sort_vector(v,comparInts);
	printf("after sorting\n");
	print_vector(v);

	s = subrange_vector(v,1,6);
	printf("sub range 1,6\n");
	print_vector(s);

	a = binary_map_vector(v,s,addInts);
	printf("addition of subrange with sorted vector\n");
	print_vector(a);

	i = 0;
	ac = scan_vector(v,accumInts,&i);
	printf("Sum of v is %d\n",i);

	i = 1654;
	set_element_vector(v,0,&i);
	dv = new_default_vector(10,sizeof(double));
	cast_vector(dv,v,intSqrt);
	iterate_vector(dv,printDouble);
	return 0;
}
#endif
