#ifndef __vector_h__
#define __vector_h__
#include <stddef.h>
#include <stdbool.h>
/* ---------------------------------------------------------
   VECTOR Interfaces header file.
   The vector type is an opaque data type. Clients aren't
   allowed to use the internal structure of the vectors, only
   the interface to them.
*/
typedef struct _vector {
        size_t count;                    /* in element size units */
        unsigned char *contents;      /* The contents of the vector follow */
        size_t capacity;                /* in element_size units */
        size_t element_size;            /* in bytes */
} vector;

/*
Define the types of the allocator members
*/
#ifndef _GENERIC_TYPES_DEFINED
#define _GENERIC_TYPES_DEFINED
typedef void *(unary_operator)(void *arg,void *result);
typedef void *(*binary_operator)(void *arg1,void *arg2,void *result);
#endif
/*
Naming rules:
new_<type> Allocate  <type> and return a pointer to it
	For instance: new_vector
new_default_<type> Allocate <type> calling new_<typ> with the default arguments.

Read Access to properties:
-------------------------
get_<property name>_<type>
For instance
get_capacity_vector
Write Access to properties:
--------------------------
set_<property name>_<type>
For instance
set_element_vector

Features:
<feature name>_<type>
For instance:
map_vector

*/
/* --------------------------------------------------------- */
/* Allocation                                                */
/* --------------------------------------------------------- */
/* Allocate a new vector.
   Arguments:
              size_t size. Initial size of the vector.
              size_t element_size: size of each element in bytes. Required.
              void *contents. Can be NULL.
              size_t capacity. Initial capacity. Defaults to VECTOR_CHUNK_SIZE if
                               size < VECTOR_CHUNK_SIZE, to a multiple of
                               VECTOR_CHUNK_SIZE otherwise.
              int (*pmalloc)(unsigned int) Defaults to malloc
              void (*pfree)(void *) Defaults to free.
*/
#define DEFAULT_VECTOR_SIZE 16
vector *new_vector(size_t size,size_t element_size,void *contents=0,size_t capacity=0);
/* Allocates a new vector using default values */
vector *new_default_vector(size_t size,size_t element_size);
/* Copy an existing vector into a freshly allocated one */
vector *copy_vector(vector &);
/* destroy a vector releasing all space associated with it */
void dispose_vector(vector *);
/* --------------------------------------------------------- */
/* Accessing the vector properties                           */
/* --------------------------------------------------------- */
size_t get_size_vector(vector *);
size_t get_element_size_vector(vector *);
size_t get_capacity_vector(vector *);
/* get a pointer to an element */
void *get_element_vector(vector *v,size_t position);
void *operator[]=(vector *v,size_t position,void *data);
void * operator[](vector *v,size_t position);

/* get a sub range from the given vector. Returns a copy */
vector *get_subrange_vector(vector *v,size_t start,size_t end);
/* --------------------------------------------------------- */
/* Setting the vector properties                             */
/* --------------------------------------------------------- */
void *set_element_vector(vector *v,size_t position,void *);

/* --------------------------------------------------------- */
/* Using a vector                                            */
/* --------------------------------------------------------- */

/*
return a pointer to the last element of a vector and remove it
*/
void *pop_back_vector(vector *v);


/* Insert into the given vector, an element before the given position
   The result could be a new vector, and the old vector could have been
   freed.
   An example of usage is
	v = insert_vector(v,5,data);
*/
bool insert_vector(vector *v,size_t position,void *element);

/* Eliminate element at the given position
   Returns: a pointer to the removed element data if successfull, NULL
   otherwise.
*/
bool remove_vector(vector *v,size_t position);


/* Set the given vector to empty.
   Returns the old size of the vector;
*/
size_t clear_vector(vector *v);


/* If the given vectors are equal returns 1, otherwise returns zero. Note that
   neither the capacity, nor the allocator/free members will be included in
   the comparison. Only size, element_size, and contents are tested.
   Two NULL vectors are considered equal.
*/
int equal_vector(vector *v1,vector *v2);


/* Apply the given function to each of the elements of the vector, storing
   the results in a new vector. The resulting vector will have the same shape,
   size, and malloc/free members as the given vector.
   The called function receives a pointer to each element, and a pointer to
   memory where it can store the its results. Normally it should return the
   pointer to the result, or NULL to stop the iteration.
*/
vector *map_vector(vector *v,unary_operator fn);


/* Call the given function for each element of the vector until either the end
   of the vector is reached, or the function returns zero.
   Returns a pointer to the last element analyzed or NULL if the end is
   reached.
*/
void *iterate_vector(vector *v,int (*fn)(void *));

/* Builds a new vector of the same shape with the results of the binary
   function "fn" applied to each element of both vectors. The called function
   receives a pointer where it stores the result of the operation.
*/
vector *binary_map_vector(vector *v1,vector *v2,binary_operator fn);

/* Call the given unary operator fn, passing it the pstart data at each call.
   This can be used to accumulate the results of an operation in the vector,
   or to search for an element in a vector linearly.
*/
void *scan_vector(vector *v,unary_operator fn,void *pstart);
/*
Cast vector v2 into vector v1 using the cast unary operator fn. The contents of
v1 are destroyed.
*/
vector *cast_vector(vector *v1,vector *v2,unary_operator fn);

/* Get a pointer to the actual contents of the vector
*/
void *get_contents_vector(vector *v);
void sort_vector(vector *v,int (*compar)(const void *,const void *));
vector *nsort_vector(vector *v,int (*compar)(const void *,const void *));
bool reverse_vector(vector &v);
bool concat_vector(vector &v1,vector &v2);
#endif // __list_h__
